/*
 * Name:    i_mardtb.c
 *
 * Purpose: MX interface driver for the MarUSA Desktop Beamline
 *          goniostat controller.  This driver uses the 'tcp232'
 *          RS-232 driver to connect to the controller via ethernet.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2002-2003 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXI_MARDTB_DEBUG	FALSE

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_types.h"
#include "mx_ascii.h"
#include "mx_rs232.h"

#include "i_mardtb.h"

MX_RECORD_FUNCTION_LIST mxi_mardtb_record_function_list = {
	NULL,
	mxi_mardtb_create_record_structures,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	mxi_mardtb_open,
	NULL,
	NULL,
	NULL
};

MX_RECORD_FIELD_DEFAULTS mxi_mardtb_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_MARDTB_STANDARD_FIELDS
};

long mxi_mardtb_num_record_fields
		= sizeof( mxi_mardtb_record_field_defaults )
			/ sizeof( mxi_mardtb_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_mardtb_rfield_def_ptr
			= &mxi_mardtb_record_field_defaults[0];

/* --- */

static mx_status_type
mxi_mardtb_get_pointers( MX_RECORD *mardtb_record,
			MX_MARDTB **mardtb,
			const char *calling_fname )
{
	static const char fname[] = "mxi_mardtb_get_pointers()";

	if ( mardtb_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_RECORD pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( mardtb == (MX_MARDTB **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MARDTB pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*mardtb = (MX_MARDTB *) mardtb_record->record_type_struct;

	if ( *mardtb == (MX_MARDTB *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MARDTB pointer for record '%s' is NULL.",
			mardtb_record->name );
	}
	return MX_SUCCESSFUL_RESULT;
}

/* ====== */

MX_EXPORT mx_status_type
mxi_mardtb_create_record_structures( MX_RECORD *record )
{
	static const char fname[] = "mxi_mardtb_create_record_structures()";

	MX_MARDTB *mardtb;

	mardtb = (MX_MARDTB *) malloc( sizeof(MX_MARDTB) );

	if ( mardtb == (MX_MARDTB *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Can't allocate memory for MX_MARDTB structure." );
	}

	record->record_class_struct = NULL;
	record->class_specific_function_list = NULL;

	record->record_type_struct = mardtb;

	mardtb->record = record;

	mardtb->currently_active_record = NULL;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_mardtb_open( MX_RECORD *record )
{
	static const char fname[] = "mxi_mardtb_open()";

	MX_MARDTB *mardtb;
	mx_status_type mx_status;

	mx_status = mxi_mardtb_get_pointers( record, &mardtb, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, record->name));

	/* Discard the signon banner from the remote host. */

	mx_status = mx_rs232_discard_unread_input( mardtb->rs232_record,
							MXI_MARDTB_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* We are presumably at the login prompt here, so send the username
	 * to the remote host.
	 */

	mx_status = mx_rs232_putline( mardtb->rs232_record, mardtb->username,
						NULL, MXI_MARDTB_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Discard login messages. */

	mx_status = mx_rs232_discard_unread_input( mardtb->rs232_record,
							MXI_MARDTB_DEBUG );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/* === Extra functions for the use of this driver. === */

MX_EXPORT mx_status_type
mxi_mardtb_command( MX_MARDTB *mardtb,
			char *command,
			char *response,
			size_t response_buffer_length,
			int debug_flag )
{
	const char fname[] = "mxi_mardtb_command()";

	MX_RECORD *currently_active_record;
	char c;
	int at_start_of_line, carriage_return_seen, non_blank_line_seen;
	int i, line, num_chars, exit_loop;
	mx_status_type mx_status;

	if ( mardtb == (MX_MARDTB *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"NULL MX_MARDTB pointer passed." );
	}
	if ( command == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"NULL command buffer pointer passed." );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2, ("%s: sending '%s'", fname, command));
	}

	/* See if some Mar DTB record is currently using the connection.
	 * Normally, this will mean a move is in progress.
	 */

	currently_active_record = mardtb->currently_active_record;

	if ( currently_active_record != NULL ) {
		return mx_error( MXE_TRY_AGAIN, fname,
		"The command '%s' could not be sent to Mar DTB record '%s', "
		"since it is currently in use by record '%s'.",
			command, mardtb->record->name,
			currently_active_record->name );
	}

	/* Discard any existing characters in the input buffer. */

	mx_status = mx_rs232_discard_unread_input( mardtb->rs232_record,
							debug_flag );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Send the command string. */

	mx_status = mx_rs232_putline( mardtb->rs232_record, command, NULL, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* The Mar Desktop Beamline controller always (?) returns at least
	 * two lines of text.  The first line contains an echoed copy of
	 * the command line we just sent, while the second line is a blank
	 * spacer line.  These two lines are unconditionally thrown away
	 * here.  The line delimiter is assumed to be a <CR> character.
	 *
	 * If we have discarded MX_MARDTB_NUM_CHARS_LIMIT characters
	 * without seeing a <CR>, we assume that something is wrong
	 * and return with an error message.
	 */

	for ( line = 1; line <= 2; line++ ) {
		carriage_return_seen = FALSE;
		num_chars = 0;

		while ( carriage_return_seen == FALSE ) {

			mx_status = mx_rs232_getchar( mardtb->rs232_record,
							&c, MXF_232_WAIT );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			if ( c == MX_CR ) {
				carriage_return_seen = TRUE;
			}

			if ( ++num_chars > MX_MARDTB_NUM_CHARS_LIMIT ) {
				return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"The record '%s' has received over %d characters without "
		"seeing a carriage return character.  Something is wrong "
		"with the connection to the Mar Desktop Beamline controller.",
					mardtb->rs232_record->name,
					MX_MARDTB_NUM_CHARS_LIMIT );
			}
		}
	}

	/* There should be one more character in the input buffer, namely,
	 * a line feed <LF> character.
	 */

	mx_status = mx_rs232_getchar( mardtb->rs232_record,
					&c, MXF_232_WAIT );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( c != MX_LF ) {
		mx_warning(
"%s: A '%c' (%#x) character was seen when a line feed was expected.  "
"The character will be ignored in the hope that things will work anyway.",
			fname, c, c );
	}

	/* Return now if no further response is expected. */

	if ( response == NULL ) {
#if 0
		/* If no further response is expected, throw any remaining
		 * returned characters away.
		 */

		mx_status = mx_rs232_discard_unread_input(
					mardtb->rs232_record, debug_flag );

		return mx_status;
#else
		return MX_SUCCESSFUL_RESULT;
#endif
	}

	/* Read the expected response. */

	/* Discard characters until we find a non-blank
	 * response line.  Null characters and linefeed
	 * characters are discarded as well.
	 */

	at_start_of_line = TRUE;
	non_blank_line_seen = FALSE;
	num_chars = 0;

	while( non_blank_line_seen == FALSE ) {

		mx_status = mx_rs232_getchar( mardtb->rs232_record,
							&c, MXF_232_WAIT );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		switch( c ) {
		case MX_NUL:
		case MX_LF:
		case MX_CR:
			/* Ignore null, <CR> and <LF> characters. */

			MX_DEBUG( 2,("%s: ignoring (%0#x) character.",
				fname, c));

			break;
		default:
			/* This must be the first character
			 * of the response we are looking for,
			 * so store the character at the
			 * start of the response buffer and
			 * break out of the while() loop.
			 */

			response[0] = c;

			MX_DEBUG( 2,("%s: Start of the response line seen.",
				fname));

			MX_DEBUG( 2,("%s: response[0] = '%c' (%0#x)",
					fname, c, c));

			non_blank_line_seen = TRUE;

			break;
		}

		if ( ++num_chars > MX_MARDTB_NUM_CHARS_LIMIT ) {
			return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"The record '%s' has received over %d characters without "
		"seeing a response line.  Something is wrong with the "
		"connection to the Mar Desktop Beamline controller.",
				mardtb->rs232_record->name,
				MX_MARDTB_NUM_CHARS_LIMIT );
		}
	}

	/* Read the rest of the response characters on this line. */

	num_chars = 0;
	exit_loop = FALSE;

	for ( i = 1; i < response_buffer_length; ) {

		mx_status = mx_rs232_getchar( mardtb->rs232_record,
							&c, MXF_232_WAIT );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		if ( ++num_chars > MX_MARDTB_NUM_CHARS_LIMIT ) {
			return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"The record '%s' has received over %d characters without "
		"seeing the end of the response line.  Something is wrong "
		"with the connection to the Mar Desktop Beamline controller.",
				mardtb->rs232_record->name,
				MX_MARDTB_NUM_CHARS_LIMIT );
		}

		switch( c ) {
		case MX_NUL:
			/* We still discard null characters. */

			MX_DEBUG( 2,("%s: ignoring (%0#x) character.",
				fname, c));

			break;
		case MX_CR:
		case MX_LF:
			/* This is the end of the response line.  Make sure
			 * the line is null terminated and then exit the loop.
			 */

			MX_DEBUG( 2,("%s: end of response seen.  char = (%0#x)",
				fname, c));

			response[i] = '\0';

			exit_loop = TRUE;
			break;
		default:
			/* Copy this character to the response buffer. */

			response[i] = c;

			MX_DEBUG( 2,("%s: response[%d] = '%c' (%0#x)",
					fname, i, c, c));

			i++;
			break;
		}

		if ( exit_loop ) {
			break;
		}
	}

	if ( ( exit_loop == FALSE ) && ( i >= response_buffer_length ) ) {
		response[response_buffer_length - 1] = '\0';

		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"The response to the MarDTB command '%s' exceeded the "
		"length of the response buffer (%ld).  Partial response = '%s'",
			command, (long) response_buffer_length, response );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: response = '%s'", fname, response));
	}

	MX_DEBUG( 2,("%s complete.", fname));

	return mx_status;
}

MX_EXPORT mx_status_type
mxi_mardtb_check_for_move_in_progress( MX_MARDTB *mardtb,
					int *move_in_progress )
{
	static const char fname[] = "mxi_mardtb_check_for_move_in_progress()";

	int input_available;
	mx_status_type mx_status;

	if ( mardtb == (MX_MARDTB *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_MARDTB pointer passed is NULL." );
	}
	if ( move_in_progress == (int *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The move_in_progress pointer passed is NULL." );
	}

	if ( mardtb->currently_active_record == NULL ) {
		MX_DEBUG( 2,("%s: No move was already in progress.", fname));
		MX_DEBUG( 2,("%s: *move_in_progress = %d",
					fname, *move_in_progress));

		*move_in_progress = FALSE;

		return MX_SUCCESSFUL_RESULT;
	}

	/* At the end of a move, the Mar Desktop Beamline controller sends out
	 * a new command prompt.  This is the only indication it gives that
	 * the move is complete.  If no new characters are available from the
	 * controller, then the move is still in progress.
	 */

	mx_status = mx_rs232_input_is_available( mardtb->rs232_record,
						&input_available );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s: input_available = %d", fname, input_available));

	if ( input_available ) {
		MX_DEBUG( 2,("%s: The move by motor '%s' is complete.",
			fname, mardtb->currently_active_record->name));

		*move_in_progress = FALSE;

		mardtb->currently_active_record = NULL;
	} else {
		MX_DEBUG( 2,("%s: The move by motor '%s' is still in progress.",
			fname, mardtb->currently_active_record->name));

		*move_in_progress = TRUE;
	}

	MX_DEBUG( 2,("%s: *move_in_progress = %d", fname, *move_in_progress));

	return MX_SUCCESSFUL_RESULT;
}

