/*
 * Name:    i_xia_network.c
 *
 * Purpose: MX network device for communicating with an XIA DXP multichannel
 *          analyzer system that is controlled by a remote MX server using
 *          the 'xia_xerxes' driver.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2002-2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXI_XIA_NETWORK_DEBUG		FALSE

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_types.h"
#include "mx_net.h"
#include "mx_mca.h"
#include "i_xia_network.h"
#include "d_xia_dxp_mca.h"

MX_RECORD_FUNCTION_LIST mxi_xia_network_record_function_list = {
	NULL,
	mxi_xia_network_create_record_structures,
	mxi_xia_network_finish_record_initialization,
	NULL,
	NULL,
	NULL,
	NULL,
	mxi_xia_network_open,
	NULL,
	NULL,
	mxi_xia_network_resynchronize
};

MX_RECORD_FIELD_DEFAULTS mxi_xia_network_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_XIA_NETWORK_STANDARD_FIELDS
};

long mxi_xia_network_num_record_fields
		= sizeof( mxi_xia_network_record_field_defaults )
			/ sizeof( mxi_xia_network_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_xia_network_rfield_def_ptr
			= &mxi_xia_network_record_field_defaults[0];

static void mxi_xia_network_set_data_available_flags(
					MX_XIA_NETWORK *xia_network,
					int flag_value );

static mx_status_type
mxi_xia_network_get_pointers( MX_MCA *mca,
			MX_XIA_DXP_MCA **xia_dxp_mca,
			MX_XIA_NETWORK **xia_network,
			const char *calling_fname )
{
	const char fname[] = "mxi_xia_network_get_pointers()";

	MX_RECORD *xia_dxp_record;

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( xia_dxp_mca == (MX_XIA_DXP_MCA **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_XIA_DXP_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( xia_network == (MX_XIA_NETWORK **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_XIA_NETWORK pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( mca->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_RECORD pointer for the MX_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*xia_dxp_mca = (MX_XIA_DXP_MCA *) mca->record->record_type_struct;

	xia_dxp_record = (*xia_dxp_mca)->xia_dxp_record;

	if ( xia_dxp_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The xia_dxp_record pointer for MCA record '%s' is NULL.",
			mca->record->name );
	}

	*xia_network = (MX_XIA_NETWORK *) xia_dxp_record->record_type_struct;

	if ( (*xia_network) == (MX_XIA_NETWORK *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_NETWORK pointer for XIA DXP record '%s' is NULL.",
			xia_dxp_record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*==========================*/

MX_EXPORT mx_status_type
mxi_xia_network_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_network_create_record_structures()";

	MX_XIA_NETWORK *xia_network;

	/* Allocate memory for the necessary structures. */

	xia_network = (MX_XIA_NETWORK *) malloc( sizeof(MX_XIA_NETWORK) );

	if ( xia_network == (MX_XIA_NETWORK *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Can't allocate memory for MX_XIA_NETWORK structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = NULL;;
	record->record_type_struct = xia_network;

	record->class_specific_function_list = NULL;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_network_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_network_finish_record_initialization()";

	MX_XIA_NETWORK *xia_network;

	xia_network = (MX_XIA_NETWORK *) record->record_type_struct;

	if ( xia_network == (MX_XIA_NETWORK *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_NETWORK pointer for record '%s' is NULL.",
			record->name );
	}

	/* We cannot allocate the mca_record_array structures until
	 * mxi_xia_network_open() is invoked, so for now we initialize
	 * these data structures to all zeros.
	 */

	xia_network->num_mcas = 0;
	xia_network->mca_record_array = NULL;

	return MX_SUCCESSFUL_RESULT;
}

/* mxi_xia_network_set_data_available_flags() sets mca->new_data_available
 * and xia_dxp_mca->new_statistics_available for all of the MCAs controlled
 * by this XIA remote network interface.
 */

static void
mxi_xia_network_set_data_available_flags( MX_XIA_NETWORK *xia_network,
					int flag_value )
{
	MX_RECORD *mca_record;
	MX_MCA *mca;
	MX_XIA_DXP_MCA *xia_dxp_mca;
	unsigned long i;

	for ( i = 0; i < xia_network->num_mcas; i++ ) {

		mca_record = (xia_network->mca_record_array)[i];

		mca = (MX_MCA *) mca_record->record_class_struct;

		xia_dxp_mca = (MX_XIA_DXP_MCA *) mca_record->record_type_struct;

		mca->new_data_available = flag_value;
		xia_dxp_mca->new_statistics_available = flag_value;
	}
}

MX_EXPORT mx_status_type
mxi_xia_network_open( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_network_open()";

	MX_XIA_NETWORK *xia_network;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int i, num_mcas;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	xia_network = (MX_XIA_NETWORK *) (record->record_type_struct);

	if ( xia_network == (MX_XIA_NETWORK *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_XIA_NETWORK pointer for record '%s' is NULL.",
			record->name);
	}

	/* Find out how many detector channels (MCAs) there are. */

	sprintf( rfname, "%s.num_mcas", xia_network->remote_record_name );

	mx_status = mx_get_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, &num_mcas );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Allocate and initialize an array to store pointers to all of
	 * the MCA records used by this interface.
	 */

	xia_network->mca_record_array = ( MX_RECORD ** ) malloc(
					num_mcas * sizeof( MX_RECORD * ) );

	if ( xia_network->mca_record_array == ( MX_RECORD ** ) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %d element array "
		"of MX_RECORD pointers for the mca_record_array data "
		"structure of record '%s'.", num_mcas, record->name );
	}

	for ( i = 0; i < num_mcas; i++ ) {
		xia_network->mca_record_array[i] = NULL;
	}

	xia_network->num_mcas = num_mcas;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_network_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_network_resynchronize()";

	MX_XIA_NETWORK *xia_network;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int resynchronize;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	xia_network = (MX_XIA_NETWORK *) (record->record_type_struct);

	if ( xia_network == (MX_XIA_NETWORK *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_XIA_NETWORK pointer for record '%s' is NULL.",
			record->name );
	}

	sprintf( rfname, "%s.resynchronize", xia_network->remote_record_name );

	resynchronize = 1;

	mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_INT, &resynchronize );

	return mx_status;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxi_xia_network_is_busy( MX_MCA *mca,
			int *busy_flag,
			int debug_flag )
{
	const char fname[] = "mxi_xia_network_is_busy()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_NETWORK *xia_network;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int busy;
	mx_status_type mx_status;

	mx_status = mxi_xia_network_get_pointers( mca,
					&xia_dxp_mca, &xia_network, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.busy", xia_dxp_mca->mca_label );

	mx_status = mx_get_by_name( xia_network->server_record,
				rfname, MXFT_INT, &busy );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( busy ) {
		mca->busy = TRUE;
	} else {
		mca->busy = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_network_read_parameter( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type *value_ptr,
			int debug_flag )
{
	const char fname[] = "mxi_xia_network_read_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_NETWORK *xia_network;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long dimension_array[1];
	unsigned long parameter_value;
	mx_status_type mx_status;

	mx_status = mxi_xia_network_get_pointers( mca,
					&xia_dxp_mca, &xia_network, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Record '%s', channel %d, parameter '%s'.",
			fname, mca->record->name,
			xia_dxp_mca->detector_channel, parameter_name));
	}

	sprintf( rfname, "%s.parameter_name", xia_dxp_mca->mca_label );

	dimension_array[0] = (long) strlen( parameter_name );

	mx_status = mx_put_array_by_name( xia_network->server_record,
				rfname, MXFT_STRING, 1, dimension_array,
					parameter_name );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.parameter_value", xia_dxp_mca->mca_label );

	mx_status = mx_get_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, &parameter_value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	*value_ptr = (mx_uint32_type) parameter_value;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: value = %lu", fname, *value_ptr));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_network_write_parameter( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type value,
			int debug_flag )
{
	const char fname[] = "mxi_xia_network_write_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_NETWORK *xia_network;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long dimension_array[1];
	unsigned long parameter_value;
	mx_status_type mx_status;

	mx_status = mxi_xia_network_get_pointers( mca,
					&xia_dxp_mca, &xia_network, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	parameter_value = (unsigned long) value;

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Record '%s', channel %d, parameter '%s', value = %lu.",
			fname, mca->record->name, xia_dxp_mca->detector_channel,
			parameter_name, parameter_value));
	}

	sprintf( rfname, "%s.parameter_name", xia_dxp_mca->mca_label );

	dimension_array[0] = (long) strlen( parameter_name );

	mx_status = mx_put_array_by_name( xia_network->server_record,
				rfname, MXFT_STRING, 1, dimension_array,
					parameter_name );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.parameter_value", xia_dxp_mca->mca_label );

	mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, &parameter_value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Write to parameter '%s' succeeded.",
			fname, parameter_name));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_network_write_param_to_all_channels( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type value,
			int debug_flag )
{
	const char fname[] = "mxi_xia_network_write_param_to_all_channels()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_NETWORK *xia_network;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long dimension_array[1];
	unsigned long parameter_value;
	mx_status_type mx_status;

	mx_status = mxi_xia_network_get_pointers( mca,
					&xia_dxp_mca, &xia_network, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	parameter_value = (unsigned long) value;

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Record '%s', channel %d, parameter '%s', value = %lu.",
			fname, mca->record->name, xia_dxp_mca->detector_channel,
			parameter_name, parameter_value));
	}

	sprintf( rfname, "%s.parameter_name", xia_dxp_mca->mca_label );

	dimension_array[0] = (long) strlen( parameter_name );

	mx_status = mx_put_array_by_name( xia_network->server_record,
				rfname, MXFT_STRING, 1, dimension_array,
					parameter_name );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.param_value_to_all_channels",
					xia_dxp_mca->mca_label );

	mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, &parameter_value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Write to parameter '%s' succeeded.",
			fname, parameter_name));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_network_start_run( MX_MCA *mca,
			int clear_flag,
			int debug_flag )
{
	const char fname[] = "mxi_xia_network_start_run()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_NETWORK *xia_network;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long dimension_array[1];
	double parameter_array[2];
	mx_status_type mx_status;

	mx_status = mxi_xia_network_get_pointers( mca,
					&xia_dxp_mca, &xia_network, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: Record '%s', clear_flag = %d.",
				fname, mca->record->name, clear_flag));
	}

	dimension_array[0] = 2;

	parameter_array[0] = (double) mca->preset_type;

	switch( mca->preset_type ) {
	case MXF_MCA_PRESET_LIVE_TIME:
		parameter_array[1] = mca->preset_live_time;
		break;
	case MXF_MCA_PRESET_REAL_TIME:
		parameter_array[1] = mca->preset_real_time;
		break;
	case MXF_MCA_PRESET_COUNT:
		parameter_array[1] = (double) mca->preset_count;
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown preset type %d for MCA '%s'",
			mca->preset_type, mca->record->name );
		break;
	}

	mxi_xia_network_set_data_available_flags( xia_network, TRUE );

	sprintf( rfname, "%s.start_with_preset", xia_dxp_mca->mca_label );

	mx_status = mx_put_array_by_name( xia_network->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				&parameter_array );

	return mx_status;
}

MX_EXPORT mx_status_type
mxi_xia_network_stop_run( MX_MCA *mca,
			int debug_flag )
{
	const char fname[] = "mxi_xia_network_stop_run()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_NETWORK *xia_network;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int stop;
	mx_status_type mx_status;

	mx_status = mxi_xia_network_get_pointers( mca,
					&xia_dxp_mca, &xia_network, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: Record '%s'.", fname, mca->record->name));
	}

	sprintf( rfname, "%s.stop", xia_dxp_mca->mca_label );

	stop = 1;

	mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_INT, &stop );

	return mx_status;
}

MX_EXPORT mx_status_type
mxi_xia_network_read_spectrum( MX_MCA *mca,
			int debug_flag )
{
	const char fname[] = "mxi_xia_network_read_spectrum()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_NETWORK *xia_network;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long dimension_array[1];
	mx_status_type mx_status;

	mx_status = mxi_xia_network_get_pointers( mca,
					&xia_dxp_mca, &xia_network, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: reading out %ld channels from MCA '%s'.",
			fname, mca->current_num_channels, mca->record->name));
	}

	sprintf( rfname, "%s.channel_array", xia_dxp_mca->mca_label );

	dimension_array[0] = mca->current_num_channels;

	mx_status = mx_get_array_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
					mca->channel_array );

	if ( debug_flag ) {
		MX_DEBUG(-2,(
		"%s: readout from MCA '%s' complete.  mx_status = %ld",
			fname, mca->record->name, mx_status.code));
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mxi_xia_network_read_statistics( MX_MCA *mca,
			int debug_flag )
{
	const char fname[] = "mxi_xia_network_read_statistics()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_NETWORK *xia_network;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long dimension_array[1];
	mx_status_type mx_status;

	mx_status = mxi_xia_network_get_pointers( mca,
					&xia_dxp_mca, &xia_network, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( rfname, "%s.statistics", xia_dxp_mca->mca_label );

	dimension_array[0] = MX_XIA_DXP_NUM_STATISTICS;

	mx_status = mx_get_array_by_name( xia_network->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
					xia_dxp_mca->statistics );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mca->real_time = xia_dxp_mca->statistics[ MX_XIA_DXP_REAL_TIME ];
	mca->live_time = xia_dxp_mca->statistics[ MX_XIA_DXP_LIVE_TIME ];
	xia_dxp_mca->input_count_rate = 
		xia_dxp_mca->statistics[ MX_XIA_DXP_INPUT_COUNT_RATE ];
	xia_dxp_mca->output_count_rate =
		xia_dxp_mca->statistics[ MX_XIA_DXP_OUTPUT_COUNT_RATE ];
	xia_dxp_mca->num_fast_peaks = mx_round(
		xia_dxp_mca->statistics[ MX_XIA_DXP_NUM_FAST_PEAKS ] );
	xia_dxp_mca->num_events = mx_round(
		xia_dxp_mca->statistics[ MX_XIA_DXP_NUM_EVENTS ] );
	xia_dxp_mca->num_underflows = mx_round(
		xia_dxp_mca->statistics[ MX_XIA_DXP_NUM_UNDERFLOWS ] );
	xia_dxp_mca->num_overflows = mx_round(
		xia_dxp_mca->statistics[ MX_XIA_DXP_NUM_OVERFLOWS ] );

	return mx_status;
}

MX_EXPORT mx_status_type
mxi_xia_network_get_mx_parameter( MX_MCA *mca )
{
	const char fname[] = "mxi_xia_network_get_mx_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_NETWORK *xia_network;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long dimension_array[2];
	mx_status_type mx_status;

	mx_status = mxi_xia_network_get_pointers( mca,
					&xia_dxp_mca, &xia_network, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for MCA '%s', parameter type '%s' (%d).",
		fname, mca->record->name,
		mx_get_field_label_string(mca->record,mca->parameter_type),
		mca->parameter_type));

	switch( mca->parameter_type ) {
	case MXLV_MCA_CURRENT_NUM_CHANNELS:
		sprintf( rfname, "%s.current_num_channels",
				xia_dxp_mca->mca_label );

		mx_status = mx_get_by_name( xia_network->server_record,
				rfname, MXFT_LONG,
				&(mca->current_num_channels) );

		if ( mca->current_num_channels > mca->maximum_num_channels ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"The MCA '%s' controlled by server '%s' is reported to have %lu channels, "
"but the record '%s' is only configured to support up to %lu channels.",
				xia_dxp_mca->mca_label,
				xia_network->server_record->name,
				mca->current_num_channels,
				mca->record->name,
				mca->maximum_num_channels );
		}
		break;

	case MXLV_MCA_PRESET_TYPE:
		sprintf( rfname, "%s.preset_type",
					xia_dxp_mca->mca_label );

		mx_status = mx_get_by_name( xia_network->server_record,
				rfname, MXFT_INT, &(mca->preset_type) );
		break;

	case MXLV_MCA_ROI_ARRAY:
		sprintf( rfname, "%s.current_num_rois",
					xia_dxp_mca->mca_label );

		mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_LONG, &(mca->current_num_rois) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.roi_array",
					xia_dxp_mca->mca_label );

		dimension_array[0] = mca->current_num_rois;
		dimension_array[1] = 2;

		mx_status = mx_get_array_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, 2, dimension_array,
				&(mca->roi_array) );
		break;

	case MXLV_MCA_ROI_INTEGRAL_ARRAY:
		sprintf( rfname, "%s.current_num_rois",
					xia_dxp_mca->mca_label );

		mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_LONG, &(mca->current_num_rois) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.roi_integral_array",
					xia_dxp_mca->mca_label );

		dimension_array[0] = mca->current_num_rois;

		mx_status = mx_get_array_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
				&(mca->roi_integral_array) );
		break;

	case MXLV_MCA_ROI:
		sprintf( rfname, "%s.roi_number",
					xia_dxp_mca->mca_label );

		mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, &(mca->roi_number) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.roi", xia_dxp_mca->mca_label );

		dimension_array[0] = 2;

		mx_status = mx_get_array_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
				&(mca->roi) );
		break;
	
	case MXLV_MCA_ROI_INTEGRAL:
		sprintf( rfname, "%s.roi_number",
					xia_dxp_mca->mca_label );

		mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, &(mca->roi_number) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.roi_integral",
				xia_dxp_mca->mca_label );

		mx_status = mx_get_by_name( xia_network->server_record,
				rfname, MXFT_ULONG,
				&(mca->roi_integral) );
		break;

	case MXLV_MCA_SOFT_ROI_ARRAY:
		sprintf( rfname, "%s.soft_roi_array",
					xia_dxp_mca->mca_label );

		dimension_array[0] = mca->num_soft_rois;
		dimension_array[1] = 2;

		mx_status = mx_get_array_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, 2, dimension_array,
				&(mca->soft_roi_array) );
		break;

	case MXLV_MCA_SOFT_ROI_INTEGRAL_ARRAY:
		sprintf( rfname, "%s.soft_roi_integral_array",
					xia_dxp_mca->mca_label );

		dimension_array[0] = mca->num_soft_rois;

		mx_status = mx_get_array_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
				&(mca->soft_roi_integral_array) );
		break;

	case MXLV_MCA_SOFT_ROI:
		sprintf( rfname, "%s.soft_roi_number",
					xia_dxp_mca->mca_label );

		mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, &(mca->soft_roi_number) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.soft_roi",
					xia_dxp_mca->mca_label );

		dimension_array[0] = 2;

		mx_status = mx_get_array_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
				&(mca->soft_roi) );
		break;

	case MXLV_MCA_SOFT_ROI_INTEGRAL:
		sprintf( rfname, "%s.soft_roi_number",
					xia_dxp_mca->mca_label );

		mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, &(mca->soft_roi_number) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.soft_roi_integral",
				xia_dxp_mca->mca_label );

		mx_status = mx_get_by_name( xia_network->server_record,
				rfname, MXFT_ULONG,
				&(mca->soft_roi_integral) );
		break;

	case MXLV_MCA_CHANNEL_NUMBER:
		sprintf( rfname, "%s.channel_number",
				xia_dxp_mca->mca_label );

		mx_status = mx_get_by_name( xia_network->server_record,
				rfname, MXFT_ULONG,
				&(mca->channel_number) );
		break;

	case MXLV_MCA_CHANNEL_VALUE:
		sprintf( rfname, "%s.channel_value",
				xia_dxp_mca->mca_label );

		mx_status = mx_get_by_name( xia_network->server_record,
				rfname, MXFT_ULONG,
				&(mca->channel_value) );

		break;

	case MXLV_MCA_REAL_TIME:
		sprintf( rfname, "%s.real_time",
				xia_dxp_mca->mca_label );

		mx_status = mx_get_by_name( xia_network->server_record,
				rfname, MXFT_DOUBLE,
				&(mca->real_time) );

		break;

	case MXLV_MCA_LIVE_TIME:
		sprintf( rfname, "%s.live_time",
				xia_dxp_mca->mca_label );

		mx_status = mx_get_by_name( xia_network->server_record,
				rfname, MXFT_DOUBLE,
				&(mca->live_time) );
		break;

	default:
		return mx_mca_default_get_parameter_handler( mca );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_network_set_mx_parameter( MX_MCA *mca )
{
	const char fname[] = "mxi_xia_network_set_mx_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_NETWORK *xia_network;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH+1 ];
	long dimension_array[2];
	mx_status_type mx_status;

	mx_status = mxi_xia_network_get_pointers( mca,
					&xia_dxp_mca, &xia_network, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for MCA '%s', parameter type '%s' (%d).",
		fname, mca->record->name,
		mx_get_field_label_string(mca->record,mca->parameter_type),
		mca->parameter_type));

	switch( mca->parameter_type ) {
	case MXLV_MCA_CURRENT_NUM_CHANNELS:
		sprintf( rfname, "%s.current_num_channels",
					xia_dxp_mca->mca_label );

		mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_LONG,
				&(mca->current_num_channels) );
		break;

	case MXLV_MCA_PRESET_TYPE:
		sprintf( rfname, "%s.preset_type",
					xia_dxp_mca->mca_label );

		mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_INT, &(mca->preset_type) );
		break;

	case MXLV_MCA_ROI_ARRAY:
		sprintf( rfname, "%s.current_num_rois",
					xia_dxp_mca->mca_label );

		mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_LONG, &(mca->current_num_rois) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.roi_array",
					xia_dxp_mca->mca_label );

		dimension_array[0] = mca->current_num_rois;
		dimension_array[1] = 2;

		mx_status = mx_put_array_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, 2, dimension_array,
				&(mca->roi_array) );
		break;

	case MXLV_MCA_ROI:
		sprintf( rfname, "%s.roi_number",
					xia_dxp_mca->mca_label );

		mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, &(mca->roi_number) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.roi", xia_dxp_mca->mca_label );

		dimension_array[0] = 2;

		mx_status = mx_put_array_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
				&(mca->roi) );
		break;

	case MXLV_MCA_SOFT_ROI_ARRAY:
		sprintf( rfname, "%s.soft_roi_array",
					xia_dxp_mca->mca_label );

		dimension_array[0] = mca->num_soft_rois;
		dimension_array[1] = 2;

		mx_status = mx_put_array_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, 2, dimension_array,
				&(mca->soft_roi_array) );
		break;

	case MXLV_MCA_SOFT_ROI:
		sprintf( rfname, "%s.soft_roi_number",
					xia_dxp_mca->mca_label );

		mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, &(mca->soft_roi_number) );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		sprintf( rfname, "%s.soft_roi",
					xia_dxp_mca->mca_label );

		dimension_array[0] = 2;

		mx_status = mx_put_array_by_name( xia_network->server_record,
				rfname, MXFT_ULONG, 1, dimension_array,
				&(mca->soft_roi) );
		break;

	case MXLV_MCA_CHANNEL_NUMBER:
		sprintf( rfname, "%s.channel_number",
				xia_dxp_mca->mca_label );

		mx_status = mx_put_by_name( xia_network->server_record,
				rfname, MXFT_ULONG,
				&(mca->channel_number) );
		break;

	default:
		return mx_mca_default_set_parameter_handler( mca );
		break;
	}

	return mx_status;
}

