/*
 * Name:    mx_epics.c
 *
 * Purpose: Some functions for using EPICS with MX.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999-2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define EPICS_DEBUG_TIMING	FALSE

#include <stdio.h>

#include "mxconfig.h"

#if HAVE_EPICS

/* The following three header files come from the EPICS base source code
 * distribution.
 */

#include "epicsVersion.h"
#include "tsDefs.h"
#include "cadef.h"

#include "mx_record.h"
#include "mx_epics.h"

#if EPICS_DEBUG_TIMING
#include "mx_hrt_debug.h"
#endif

/* Forward declaration of some needed functions. */

#define MXF_EPICS_INTERNAL_CAGET	1
#define MXF_EPICS_INTERNAL_CAPUT	2
#define MXF_EPICS_INTERNAL_CAPUT_NOWAIT	3

static mx_status_type
mx_epics_internal_handle_channel_disconnection( const char *calling_fname,
						int function_type,
						MX_EPICS_PV *pv,
						chid channel_id,
						long epics_type,
						unsigned long num_elements,
						void *data_buffer,
						double timeout,
						int max_retries );

/*---*/

/* mx_epics_connect_timeout_interval contains the timeout interval for
 * the initial connection or reconnection to EPICS process variables.
 */

static double mx_epics_connect_timeout_interval = 1.0;	/* in seconds */

/* The variable mx_epics_timeout_interval contains the timeout interval
 * for all other EPICS Channel Access calls.
 */

static double mx_epics_timeout_interval = 30.0;    	/* in seconds */

/*---*/

/* The variable mx_epics_debug_flag is used to control whether or not
 * EPICS debugging messages are generated.
 */

static int mx_epics_debug_flag = FALSE;

MX_EXPORT int
mx_epics_get_debug_flag( void )
{
	if ( mx_epics_debug_flag == FALSE ) {
		return FALSE;
	} else {
		return TRUE;
	}
}

MX_EXPORT void
mx_epics_set_debug_flag( int flag )
{
	if ( flag == FALSE ) {
		mx_epics_debug_flag = FALSE;
	} else {
		mx_epics_debug_flag = TRUE;
	}
}

/*---*/

MX_EXPORT mx_status_type
mx_epics_poll_for_events( const char *calling_fname )
{
	static const char fname[] = "mx_epics_poll_for_events()";

	int epics_status;

	epics_status = ca_poll();

	switch( epics_status ) {
	case ECA_NORMAL:
	case ECA_TIMEOUT:
		break;		/* no error */
	case ECA_EVDISALLOW:
		return mx_error( MXE_UNSUPPORTED, fname,
	    "Cannot execute ca_poll() from within an EPICS event handler.");
		break;
	default:
		return mx_error( MXE_FUNCTION_FAILED, fname,
		"EPICS background event processing for '%s' failed.  "
		"EPICS status = %d, EPICS error = '%s'",
			calling_fname, epics_status,
			ca_message( epics_status ) );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_epics_flush_io( void )
{
	static const char fname[] = "mx_epics_flush_io()";

	int epics_status;

	epics_status = ca_flush_io();

	if ( epics_status != ECA_NORMAL ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
	"An attempt to force execution of EPICS Channel Access I/O failed.  "
	"EPICS status = %d, EPICS error = '%s'.",
			epics_status, ca_message( epics_status ) );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*---*/

MX_EXPORT void
mx_epics_pvname_init( MX_EPICS_PV *pv, char *name_format, ... )
{
	va_list args;
	char buffer[1000];

	va_start(args, name_format);
	vsprintf(buffer, name_format, args);
	va_end(args);

	mx_strncpy( pv->pvname, buffer, MXU_EPICS_PVNAME_LENGTH );

	pv->channel_id = NULL;

	return;
}

/*---*/

static mx_status_type
mx_epics_internal_connect( char *pvname, chid *channel_id, int quiet )
{
	static const char fname[] = "mx_epics_internal_connect()";

	int epics_status;
	mx_status_type mx_status;

#if EPICS_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	if ( strlen(pvname) == 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Cannot connect to a zero-length EPICS PV name." );
	}

	/* Make the search request. */

#if EPICS_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	epics_status = ca_search( pvname, channel_id );

	if ( epics_status != ECA_NORMAL ) {
		if ( quiet ) {
			return mx_error_quiet( MXE_NOT_FOUND, fname,
			"Unable to connect to EPICS PV '%s'.  "
			"EPICS status = %d, EPICS error = '%s'",
			pvname, epics_status, ca_message( epics_status ) );
		} else {
			return mx_error( MXE_NOT_FOUND, fname,
			"Unable to connect to EPICS PV '%s'.  "
			"EPICS status = %d, EPICS error = '%s'",
			pvname, epics_status, ca_message( epics_status ) );
		}
	}

	/* Wait for the search request to complete. */

	epics_status = ca_pend_io( mx_epics_connect_timeout_interval );

#if EPICS_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, pvname );
#endif

	switch( epics_status ) {
	case ECA_NORMAL:	/* Success */
		MX_DEBUG( 2,("%s: successfully connected to PV '%s'.",
			fname, pvname ));
		break;
	case ECA_TIMEOUT:
		if ( quiet ) {
			return mx_error_quiet( MXE_TIMED_OUT, fname,
	"Initial connection to EPICS PV '%s' timed out after %g seconds.",
				pvname, mx_epics_connect_timeout_interval );
		} else {
			return mx_error( MXE_TIMED_OUT, fname,
	"Initial connection to EPICS PV '%s' timed out after %g seconds.",
				pvname, mx_epics_connect_timeout_interval );
		}
		break;
	default:
		if ( quiet ) {
			return mx_error_quiet( MXE_NOT_FOUND, fname,
		"The connection to EPICS PV '%s' did not complete.  "
		"EPICS status = %d, EPICS error = '%s'",
			    pvname, epics_status, ca_message( epics_status ) );
		} else {
			return mx_error( MXE_NOT_FOUND, fname,
		"The connection to EPICS PV '%s' did not complete.  "
		"EPICS status = %d, EPICS error = '%s'",
			    pvname, epics_status, ca_message( epics_status ) );
		}
	}

	/* Allow background events to execute. */

	mx_status = mx_epics_poll_for_events( fname );
		
	return mx_status;
}

static mx_status_type
mx_epics_internal_disconnect( char *pvname,
				chid channel_id )
{
	static const char fname[] = "mx_epics_internal_disconnect()";

	int epics_status;

#if EPICS_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

#if EPICS_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	if ( channel_id == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	epics_status = ca_clear_channel( channel_id );

#if EPICS_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, pvname );
#endif

	if ( epics_status != ECA_NORMAL ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
		"Attempt to disconnect from EPICS PV '%s' failed.  "
		"EPICS status = %d, EPICS error = '%s'.", pvname,
		epics_status, ca_message( epics_status ) );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*---*/

MX_EXPORT mx_status_type
mx_epics_pv_connect( MX_EPICS_PV *pv )
{
	static const char fname[] = "mx_epics_pv_connect()";

	chid channel_id;
	mx_status_type mx_status;

	if ( pv == (MX_EPICS_PV *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_PV pointer passed was NULL." );
	}

	if ( pv->channel_id != NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The EPICS PV '%s' is already connected.", pv->pvname );
	}

	mx_status = mx_epics_internal_connect( pv->pvname, &channel_id, FALSE );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	pv->channel_id = channel_id;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_epics_pv_disconnect( MX_EPICS_PV *pv )
{
	static const char fname[] = "mx_epics_pv_disconnect()";

	mx_status_type mx_status;

	if ( pv == (MX_EPICS_PV *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_PV pointer passed was NULL." );
	}

	if ( pv->channel_id == NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The EPICS PV '%s' was already disconnected.", pv->pvname );
	}

	mx_status = mx_epics_internal_disconnect( pv->pvname, pv->channel_id );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	pv->channel_id = NULL;

	/* Allow background events to execute. */

	mx_status = mx_epics_poll_for_events( fname );
		
	return mx_status;
}

/*---*/

static mx_status_type
mx_epics_internal_caget( MX_EPICS_PV *pv,
			chid channel_id,
			long epics_type,
			unsigned long num_elements,
			void *data_buffer,
			double timeout,
			int max_retries )
{
	static const char fname[] = "mx_epics_internal_caget()";

	READONLY char *pvname;
	int epics_status;
	mx_status_type mx_status;

#if EPICS_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	if ( channel_id == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The channel_id passed was NULL." );
	}
	if ( data_buffer == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The data_buffer passed was NULL." );
	}

	pvname = ca_name( channel_id );

#if EPICS_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	epics_status = ca_array_get( epics_type, num_elements,
					channel_id, data_buffer );

	MX_DEBUG( 2,("%s: ca_array_get() epics_status = %d",
			fname, epics_status));

	switch( epics_status ) {
	case ECA_NORMAL:
		break;
	case ECA_BADCHID:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"ca_array_get() was called with an "
			"illegal EPICS channel id" );
		break;
	case ECA_BADCOUNT:
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"The requested number of elements %lu for "
			"EPICS PV '%s' is larger than the maximum allowed "
			"(%lu) for this process variable.",
				num_elements, ca_name( channel_id ),
			(unsigned long) ca_element_count( channel_id ) );
		break;
	case ECA_BADTYPE:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"An illegal EPICS datatype was requested for "
			"EPICS PV '%s'.", ca_name( channel_id ) );
		break;
	case ECA_GETFAIL:
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"An attempt to read the value of local EPICS "
			"database variable '%s' failed.",
				ca_name( channel_id ) );
		break;
	case ECA_NORDACCESS:
		return mx_error( MXE_PERMISSION_DENIED, fname,
			"An attempt to read from read-protected EPICS process "
			"variable '%s' failed.",
			ca_name( channel_id ) );
		break;
	case ECA_DISCONN:
	case ECA_DISCONNCHID:
		return mx_epics_internal_handle_channel_disconnection(
						fname,
						MXF_EPICS_INTERNAL_CAGET,
						pv,
						channel_id,
						epics_type,
						num_elements,
						data_buffer,
						timeout,
						max_retries );
		break;
	default:
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"A ca_array_get() call failed for EPICS PV '%s'.  "
			"EPICS status = %d, EPICS error = '%s'.",
				ca_name( channel_id ),
				epics_status,
				ca_message( epics_status ) );
		break;
	}

	epics_status = ca_pend_io( timeout );

#if EPICS_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, pvname );
#endif

	switch( epics_status ) {
	case ECA_NORMAL:
		break;
	case ECA_TIMEOUT:
		return mx_error( MXE_TIMED_OUT, fname,
		"caget from EPICS PV '%s' timed out after %g seconds.",
			ca_name( channel_id ), timeout );
		break;
	default:
		return mx_error( MXE_FUNCTION_FAILED, fname,
		"ca_pend_io() failed for EPICS PV '%s'.  "
		"EPICS status = %d, EPICS error = '%s'.",
			ca_name( channel_id ),
			epics_status,
			ca_message( epics_status ) );
		break;
	}


	if ( mx_epics_debug_flag ) {
		if ( num_elements > 1 ) {
			if ( epics_type == MX_CA_STRING ) {
				MX_DEBUG(-2,("%s: '%s' value read = '%s'.",
					fname, pvname, (char *) data_buffer));
			} else {
				MX_DEBUG(-2,
				("%s: multielement array read from '%s'.",
					fname, pvname));
			}
		} else {
			char byte_value;
			short short_value;
			long long_value;
			float float_value;
			double double_value;

			switch( epics_type ) {
			case MX_CA_CHAR:
				byte_value = *((char *) data_buffer);

				MX_DEBUG(-2,("%s: '%s' value read = %d.",
					fname, pvname, byte_value));
				break;
			case MX_CA_STRING:
				MX_DEBUG(-2,("%s: '%s' value read = '%s'",
					fname, pvname, (char *) data_buffer));
				break;
			case MX_CA_SHORT:
				short_value = *((short *) data_buffer);
	
				MX_DEBUG(-2,("%s: '%s' value read = %hd.",
					fname, pvname, short_value));
				break;
			case MX_CA_LONG:
				long_value = *((long *) data_buffer);
	
				MX_DEBUG(-2,("%s: '%s' value read = %ld.",
					fname, pvname, long_value));
				break;
			case MX_CA_FLOAT:
				float_value = *((float *) data_buffer);

				MX_DEBUG(-2,("%s: '%s' value read = %g.",
					fname, pvname, float_value));
				break;
			case MX_CA_DOUBLE:
				double_value = *((double *) data_buffer);

				MX_DEBUG(-2,("%s: '%s' value read = %g.",
					fname, pvname, double_value));
				break;
			default:
				MX_DEBUG(-2,
				("%s: '%s' has unsupported epics_type %ld",
					fname, pvname, epics_type));
				break;
			}
		}
	}

	/* Allow background events to execute. */

	mx_status = mx_epics_poll_for_events( fname );
		
	return mx_status;
}

MX_EXPORT mx_status_type
mx_caget( MX_EPICS_PV *pv,
		long epics_type,
		unsigned long num_elements,
		void *data_buffer )
{
	static const char fname[] = "mx_caget()";

	mx_status_type mx_status;

	if ( pv == (MX_EPICS_PV *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_PV pointer passed was NULL." );
	}

	if ( pv->channel_id == NULL ) {
		mx_status = mx_epics_pv_connect( pv );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	mx_status = mx_epics_internal_caget( pv, pv->channel_id, epics_type,
						num_elements, data_buffer,
						mx_epics_timeout_interval, 1 );
	return mx_status;
}

MX_EXPORT mx_status_type
mx_caget_with_timeout( MX_EPICS_PV *pv,
		long epics_type,
		unsigned long num_elements,
		void *data_buffer,
		double timeout )
{
	static const char fname[] = "mx_caget_with_timeout()";

	mx_status_type mx_status;

	if ( pv == (MX_EPICS_PV *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_PV pointer passed was NULL." );
	}

	if ( pv->channel_id == NULL ) {
		mx_status = mx_epics_pv_connect( pv );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	mx_status = mx_epics_internal_caget( pv, pv->channel_id, epics_type,
						num_elements, data_buffer,
						timeout, 1 );
	return mx_status;
}

MX_EXPORT mx_status_type
mx_caget_by_name( char *pvname,
		long epics_type,
		unsigned long num_elements,
		void *data_buffer )
{
	static const char fname[] = "mx_caget_by_name()";

	chid channel_id;
	mx_status_type mx_status;

	if ( pvname == (char *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The pvname pointer passed was NULL." );
	}

	mx_status = mx_epics_internal_connect( pvname, &channel_id, FALSE );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_epics_internal_caget( NULL, channel_id, epics_type,
						num_elements, data_buffer,
						mx_epics_timeout_interval, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_epics_internal_disconnect( pvname, channel_id );

	return mx_status;
}

/*---*/

/* The variable mx_epics_put_callback_magic_cookie is used by the functions
 * mx_epics_internal_caput() and mx_epics_ca_array_put_callback_handler()
 * as a safe non-NULL value to set the 'puser' pointer in a 'chid' to.
 * The convention used is that if ca_puser(chid) is not NULL, then a
 * put callback is in progress for the specified channel id, while if
 * it _is_ NULL, then no callback is in progress.
 *
 * On some platforms, we could just do something like ca_set_puser(chid,1)
 * and get away with it.  However, other platforms will give you a
 * segmentation fault for your trouble if you try to use an invalid
 * pointer value like 0x1 as a placeholder.  Thus, the variable
 * mx_epics_put_callback_magic_cookie provides a safe address to set the
 * 'puser' pointer to.
 *
 * Warning: Note that the EPICS ca_puser() macro triggers a warning message
 * with GCC due to the -Wcast-qual flag.  Thus, we must specify the
 * -Wno-cast-qual flag.
 */

static int mx_epics_put_callback_magic_cookie = 123456;

static void
mx_epics_ca_array_put_callback_handler( struct event_handler_args args )
{
	static const char fname[] = "mx_epics_ca_array_put_callback_handler()";

	if ( args.chid == NULL ) {
		(void) mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Somehow this callback handler was invoked with a NULL "
		"channel id!  This should not happen and is a BUG!" );

		return;
	}

	if ( ca_puser(args.chid) == NULL ) {
		mx_warning( "%s was invoked for EPICS PV '%s', "
			"but no callback was supposed to be in progress!  "
			"Perhaps a previous call to mx_caput() failed?",
			fname, ca_name(args.chid) );
		return;
	}

	MX_DEBUG( 2,("*** %s invoked for EPICS PV '%s' ***",
		fname, ca_name(args.chid) ));

	/* Record that the callback has now occurred. */

	ca_set_puser(args.chid, NULL);

	return;
}

static mx_status_type
mx_epics_internal_caput( MX_EPICS_PV *pv,
			chid channel_id,
			long epics_type,
			unsigned long num_elements,
			void *data_buffer,
			double timeout,
			int max_retries )
{
	static const char fname[] = "mx_epics_internal_caput()";

	READONLY char *pvname;
	int epics_status;
	unsigned long i, milliseconds_to_wait, milliseconds_between_polls;
	mx_status_type mx_status;

#if EPICS_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	if ( channel_id == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The channel_id passed was NULL." );
	}
	if ( data_buffer == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The data_buffer passed was NULL." );
	}

	pvname = ca_name( channel_id );

	if ( mx_epics_debug_flag ) {
		if ( num_elements > 1 ) {
			if ( epics_type == MX_CA_STRING ) {
				MX_DEBUG(-2,("%s: sending '%s' to '%s'.",
					fname, (char *) data_buffer, pvname));
			} else {
				MX_DEBUG(-2,
				("%s: sending multielement array to '%s'.",
					fname, pvname));
			}
		} else {
			char byte_value;
			short short_value;
			long long_value;
			float float_value;
			double double_value;
	
			switch( epics_type ) {
			case MX_CA_CHAR:
				byte_value = *((char *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %d to '%s'.",
					fname, byte_value, pvname));
				break;
			case MX_CA_STRING:
				MX_DEBUG(-2,("%s: sending '%s' to '%s'",
					fname, (char *) data_buffer, pvname));
				break;
			case MX_CA_SHORT:
				short_value = *((short *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %hd to '%s'.",
					fname, short_value, pvname));
				break;
			case MX_CA_LONG:
				long_value = *((long *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %ld to '%s'.",
					fname, long_value, pvname));
				break;
			case MX_CA_FLOAT:
				float_value = *((float *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %g to '%s'.",
					fname, float_value, pvname));
				break;
			case MX_CA_DOUBLE:
				double_value = *((double *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %g to '%s'.",
					fname, double_value, pvname));
				break;
			default:
				MX_DEBUG(-2,
				("%s: '%s' has unsupported epics_type %ld",
					fname, pvname, epics_type));
				break;
			}
		}
	}

#if EPICS_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	/* Mark this channel ID as having a put callback in progress. */

	ca_set_puser( channel_id, &mx_epics_put_callback_magic_cookie );

	/* Send the request to Channel Access. */

	epics_status = ca_array_put_callback( epics_type, num_elements,
					channel_id, data_buffer,
					mx_epics_ca_array_put_callback_handler,
					NULL );

	switch( epics_status ) {
	case ECA_NORMAL:
		break;
	case ECA_BADCHID:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"ca_array_put_callback() was called with an "
			"illegal EPICS channel id" );
		break;
	case ECA_BADCOUNT:
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"The requested number of elements %lu for "
			"EPICS PV '%s' is larger than the maximum allowed "
			"(%lu) for this process variable.",
				num_elements, ca_name( channel_id ),
			(unsigned long) ca_element_count( channel_id ) );
		break;
	case ECA_BADTYPE:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"An illegal EPICS datatype was requested for "
			"EPICS PV '%s'.", ca_name( channel_id ) );
		break;
	case ECA_ALLOCMEM:
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory while trying to perform a "
			"ca_array_put_callback() call for EPICS PV '%s'.",
				ca_name( channel_id ) );
		break;
	case ECA_NOWTACCESS:
		return mx_error( MXE_PERMISSION_DENIED, fname,
			"An attempt to write to read-only EPICS process "
			"variable '%s' failed.",
			ca_name( channel_id ) );
		break;
	case ECA_DISCONN:
	case ECA_DISCONNCHID:
		return mx_epics_internal_handle_channel_disconnection(
						fname,
						MXF_EPICS_INTERNAL_CAPUT,
						pv,
						channel_id,
						epics_type,
						num_elements,
						data_buffer,
						timeout,
						max_retries );
		break;
	default:
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"A ca_array_put_callback() call failed for EPICS "
			"PV '%s'.  EPICS status = %d, EPICS error = '%s'.",
				ca_name( channel_id ),
				epics_status,
				ca_message( epics_status ) );
		break;
	}

	/* Wait for the put to complete. */

	milliseconds_to_wait = mx_round( 1000.0 * timeout );

	milliseconds_between_polls = 10;

	for (i = 0; i < milliseconds_to_wait; i += milliseconds_between_polls)
	{
		/* See if the callback has completed and allow other 
		 * background events to execute too.
		 */

		mx_status = mx_epics_poll_for_events( fname );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		if ( ca_puser( channel_id ) == NULL ) {
			MX_DEBUG( 2,(
			"%s: Callback for EPICS PV '%s' has completed.",
				fname, ca_name( channel_id ) ));

			break;		/* Exit the for() loop. */
		}

		mx_msleep( milliseconds_between_polls );
	}

#if EPICS_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, pvname );
#endif

	if ( i >= milliseconds_to_wait ) {
		return mx_error( MXE_TIMED_OUT, fname,
		"caput to EPICS PV '%s' timed out after %g seconds.",
			ca_name( channel_id ), timeout );
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mx_caput( MX_EPICS_PV *pv,
		long epics_type,
		unsigned long num_elements,
		void *data_buffer )
{
	static const char fname[] = "mx_caput()";

	mx_status_type mx_status;

	if ( pv == (MX_EPICS_PV *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_PV pointer passed was NULL." );
	}

	if ( pv->channel_id == NULL ) {
		mx_status = mx_epics_pv_connect( pv );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	mx_status = mx_epics_internal_caput( pv, pv->channel_id, epics_type,
						num_elements, data_buffer,
						mx_epics_timeout_interval, 1 );
	return mx_status;
}

MX_EXPORT mx_status_type
mx_caput_with_timeout( MX_EPICS_PV *pv,
		long epics_type,
		unsigned long num_elements,
		void *data_buffer,
		double timeout )
{
	static const char fname[] = "mx_caput_with_timeout()";

	mx_status_type mx_status;

	if ( pv == (MX_EPICS_PV *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_PV pointer passed was NULL." );
	}

	if ( pv->channel_id == NULL ) {
		mx_status = mx_epics_pv_connect( pv );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	mx_status = mx_epics_internal_caput( pv, pv->channel_id, epics_type,
						num_elements, data_buffer,
						timeout, 1 );
	return mx_status;
}

MX_EXPORT mx_status_type
mx_caput_by_name( char *pvname,
		long epics_type,
		unsigned long num_elements,
		void *data_buffer )
{
	static const char fname[] = "mx_caput_by_name()";

	chid channel_id;
	mx_status_type mx_status;

	if ( pvname == (char *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The pvname pointer passed was NULL." );
	}

	mx_status = mx_epics_internal_connect( pvname, &channel_id, FALSE );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_epics_internal_caput( NULL, channel_id, epics_type,
						num_elements, data_buffer,
						mx_epics_timeout_interval, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_epics_internal_disconnect( pvname, channel_id );

	return mx_status;
}

/*---*/

static mx_status_type
mx_epics_internal_caput_nowait( MX_EPICS_PV *pv,
				chid channel_id,
				long epics_type,
				unsigned long num_elements,
				void *data_buffer,
				int max_retries )
{
	static const char fname[] = "mx_epics_internal_caput_nowait()";

	READONLY char *pvname;
	int epics_status;
	mx_status_type mx_status;

#if EPICS_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	if ( channel_id == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The channel_id passed was NULL." );
	}
	if ( data_buffer == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The data_buffer passed was NULL." );
	}

	pvname = ca_name( channel_id );

	if ( mx_epics_debug_flag ) {
		if ( num_elements > 1 ) {
			if ( epics_type == MX_CA_STRING ) {
				MX_DEBUG(-2,("%s: sending '%s' to '%s'.",
					fname, (char *) data_buffer, pvname));
			} else {
				MX_DEBUG(-2,
				("%s: sending multielement array to '%s'.",
					fname, pvname));
			}
		} else {
			char byte_value;
			short short_value;
			long long_value;
			float float_value;
			double double_value;
	
			switch( epics_type ) {
			case MX_CA_CHAR:
				byte_value = *((char *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %d to '%s'.",
					fname, byte_value, pvname));
				break;
			case MX_CA_STRING:
				MX_DEBUG(-2,("%s: sending '%s' to '%s'",
					fname, (char *) data_buffer, pvname));
				break;
			case MX_CA_SHORT:
				short_value = *((short *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %hd to '%s'.",
					fname, short_value, pvname));
				break;
			case MX_CA_LONG:
				long_value = *((long *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %ld to '%s'.",
					fname, long_value, pvname));
				break;
			case MX_CA_FLOAT:
				float_value = *((float *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %g to '%s'.",
					fname, float_value, pvname));
				break;
			case MX_CA_DOUBLE:
				double_value = *((double *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %g to '%s'.",
					fname, double_value, pvname));
				break;
			default:
				MX_DEBUG(-2,
				("%s: '%s' has unsupported epics_type %ld",
					fname, pvname, epics_type));
				break;
			}
		}
	}

#if EPICS_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	epics_status = ca_array_put( epics_type, num_elements,
					channel_id, data_buffer );

#if EPICS_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, pvname );
#endif

	switch( epics_status ) {
	case ECA_NORMAL:
		break;
	case ECA_BADCHID:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"ca_array_put() was called with an "
			"illegal EPICS channel id" );
		break;
	case ECA_BADCOUNT:
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"The requested number of elements %lu for "
			"EPICS PV '%s' is larger than the maximum allowed "
			"(%lu) for this process variable.",
				num_elements, ca_name( channel_id ),
			(unsigned long) ca_element_count( channel_id ) );
		break;
	case ECA_BADTYPE:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"An illegal EPICS datatype was requested for "
			"EPICS PV '%s'.", ca_name( channel_id ) );
		break;
	case ECA_STRTOBIG:
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"An unusually long string was supplied for a call "
			"to ca_array_put() for EPICS PV '%s'.",
				ca_name( channel_id ) );
		break;
	case ECA_NOWTACCESS:
		return mx_error( MXE_PERMISSION_DENIED, fname,
			"An attempt to write to read-only EPICS process "
			"variable '%s' failed.",
			ca_name( channel_id ) );
		break;
	case ECA_DISCONN:
	case ECA_DISCONNCHID:
		return mx_epics_internal_handle_channel_disconnection( 
						fname,
						MXF_EPICS_INTERNAL_CAPUT_NOWAIT,
						pv,
						channel_id,
						epics_type,
						num_elements,
						data_buffer,
						0.0,
						max_retries );
		break;
	default:
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"A ca_array_put() call failed for EPICS "
			"PV '%s'.  EPICS status = %d, EPICS error = '%s'.",
				ca_name( channel_id ),
				epics_status,
				ca_message( epics_status ) );
		break;
	}

	/* Allow background events to execute. */

	mx_status = mx_epics_poll_for_events( fname );
		
	return mx_status;
}

MX_EXPORT mx_status_type
mx_caput_nowait( MX_EPICS_PV *pv,
		long epics_type,
		unsigned long num_elements,
		void *data_buffer )
{
	static const char fname[] = "mx_caput_nowait()";

	mx_status_type mx_status;

	if ( pv == (MX_EPICS_PV *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_PV pointer passed was NULL." );
	}

	if ( pv->channel_id == NULL ) {
		mx_status = mx_epics_pv_connect( pv );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	mx_status = mx_epics_internal_caput_nowait( pv, pv->channel_id,
				epics_type, num_elements, data_buffer, 1 );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_caput_nowait_by_name( char *pvname,
			long epics_type,
			unsigned long num_elements,
			void *data_buffer )
{
	static const char fname[] = "mx_caput_nowait_by_name()";

	chid channel_id;
	mx_status_type mx_status;

	if ( pvname == (char *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The pvname pointer passed was NULL." );
	}

	mx_status = mx_epics_internal_connect( pvname, &channel_id, FALSE );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_epics_internal_caput_nowait( NULL, channel_id,
				epics_type, num_elements, data_buffer, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_epics_internal_disconnect( pvname, channel_id );

	return mx_status;
}

/*---*/

static mx_status_type
mx_epics_internal_handle_channel_disconnection( const char *calling_fname,
						int function_type,
						MX_EPICS_PV *pv,
						chid channel_id,
						long epics_type,
						unsigned long num_elements,
						void *data_buffer,
						double timeout,
						int max_retries )
{
	static const char fname[] =
		"mx_epics_internal_handle_channel_disconnection()";

	char pvname[ MXU_EPICS_PVNAME_LENGTH + 1 ];
	chid new_channel_id;
	mx_status_type mx_status;

	mx_strncpy( pvname, ca_name( channel_id ), MXU_EPICS_PVNAME_LENGTH );

	MX_DEBUG( 2,("%s invoked for EPICS PV '%s', function type %d",
		fname, pvname, function_type ));

	/* Discard the old channel id. */

	(void) ca_clear_channel( channel_id );

	if ( pv != (MX_EPICS_PV *) NULL ) {
		if ( strcmp( pv->pvname, pvname ) != 0 ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The PV name '%s' from the passed MX_EPICS_PV structure "
		"does not match the PV name '%s' from the passed channel_id.",
				pv->pvname, pvname );
		}

		pv->channel_id = NULL;
	}

	if ( max_retries <= 0 ) {
		return mx_error( MXE_NETWORK_CONNECTION_LOST, calling_fname,
			"The connection to EPICS PV '%s' "
			"has been disconnected.", pvname );
	}

	/* Try to reconnect to the PV. */

	mx_status = mx_epics_internal_connect( pvname, &new_channel_id, TRUE );

	if ( mx_status.code != MXE_SUCCESS ) {
		return mx_error( MXE_NETWORK_CONNECTION_LOST, calling_fname,
			"The connection to EPICS PV '%s' "
			"has been disconnected and "
			"an attempt to reconnect failed.", pvname );
	}

	/* Try the EPICS command again. */

	switch( function_type ) {
	case MXF_EPICS_INTERNAL_CAGET:
		mx_status = mx_epics_internal_caget( pv,
						new_channel_id,
						epics_type,
						num_elements,
						data_buffer,
						timeout,
						max_retries - 1 );
		break;
	case MXF_EPICS_INTERNAL_CAPUT:
		mx_status = mx_epics_internal_caput( pv,
						new_channel_id,
						epics_type,
						num_elements,
						data_buffer,
						timeout,
						max_retries - 1 );
		break;
	case MXF_EPICS_INTERNAL_CAPUT_NOWAIT:
		mx_status = mx_epics_internal_caput_nowait( pv,
						new_channel_id,
						epics_type,
						num_elements,
						data_buffer,
						max_retries - 1 );
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"An illegal function_type %d was passed for EPICS PV '%s'.",
			function_type, pvname );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

/*---*/

MX_EXPORT mx_status_type
mx_epics_get_num_elements( MX_EPICS_PV *pv,
			unsigned long *num_elements )
{
	static const char fname[] = "mx_epics_get_num_elements()";

	mx_status_type mx_status;

	if ( pv == (MX_EPICS_PV *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_PV pointer passed was NULL." );
	}
	if ( num_elements == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The num_elements pointer passed was NULL." );
	}

	if ( pv->channel_id == NULL ) {
		mx_status = mx_epics_pv_connect( pv );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	*num_elements = ca_element_count( pv->channel_id );

	return MX_SUCCESSFUL_RESULT;
}

/*---*/

MX_EXPORT mx_status_type
mx_epics_start_group( MX_EPICS_GROUP *epics_group )
{
	static const char fname[] = "mx_epics_start_group()";

	CA_SYNC_GID group_id;
	int epics_status;

#if EPICS_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	if ( epics_group == (MX_EPICS_GROUP *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_GROUP pointer passed was NULL." );
	}

#if EPICS_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	epics_status = ca_sg_create( &group_id );

#if EPICS_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, " " );
#endif

	switch( epics_status ) {
	case ECA_NORMAL:
		break;
	case ECA_ALLOCMEM:
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	    "Ran out of memory trying to create an EPICS synchronous group." );
		break;
	default:
		return mx_error( MXE_FUNCTION_FAILED, fname,
		"Creation of an EPICS synchronous group failed.  "
		"EPICS status = %d, EPICS error = '%s'.",
		epics_status, ca_message( epics_status ) );
	}

	epics_group->group_id = group_id;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_epics_end_group( MX_EPICS_GROUP *epics_group )
{
	static const char fname[] = "mx_epics_end_group()";

	CA_SYNC_GID group_id;
	int epics_status;
	mx_status_type mx_status;

#if EPICS_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	if ( epics_group == (MX_EPICS_GROUP *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_GROUP pointer passed was NULL." );
	}

	group_id = epics_group->group_id;

	/* Wait for the synchronous group to finish running. */

#if EPICS_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	epics_status = ca_sg_block( group_id, mx_epics_timeout_interval );

	switch( epics_status ) {
	case ECA_NORMAL:
		mx_status = MX_SUCCESSFUL_RESULT;
		break;
	case ECA_TIMEOUT:
		mx_status = mx_error( MXE_TIMED_OUT, fname,
			"Timed out after waiting %g seconds for "
			"synchronous group %u to finish running.",
				mx_epics_timeout_interval,
				(unsigned int) group_id );
		break;
	case ECA_EVDISALLOW:
		mx_status = mx_error( MXE_UNSUPPORTED, fname,
			"Cannot use synchronous group requests inside "
			"an EPICS event handler." );
		break;
	case ECA_BADSYNCGRP:
		mx_status = mx_error( MXE_NOT_FOUND, fname,
			"An attempt to use EPICS synchronous group %u failed "
			"since no such synchronous group currently exists.",
				(unsigned int) group_id );
		break;
	default:
		mx_status = mx_error( MXE_FUNCTION_FAILED, fname,
			"A failure occurred while waiting for EPICS "
			"synchronous group %u to complete execution.  "
			"EPICS status = %d, EPICS error = '%s'.",
				group_id, epics_status,
				ca_message( epics_status ) );
		break;
	}

	/* Delete the synchronous group, unless we got a ECA_BADSYNCGRP
	 * error code from ca_sg_block() above.
	 */

	if ( epics_status != ECA_BADSYNCGRP ) {

		epics_status = ca_sg_delete( group_id );

		if ( epics_status != ECA_NORMAL ) {
			mx_status = mx_error( MXE_FUNCTION_FAILED, fname,
			"A failure occurred while trying to delete EPICS "
			"synchronous group %u.  "
			"EPICS status = %d, EPICS error = '%s'.",
				group_id, epics_status,
				ca_message( epics_status ) );
		}
	}

#if EPICS_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, " " );
#endif

	return mx_status;
}

MX_EXPORT mx_status_type
mx_epics_delete_group( MX_EPICS_GROUP *epics_group )
{
	static const char fname[] = "mx_epics_delete_group()";

	CA_SYNC_GID group_id;
	int epics_status;
	mx_status_type mx_status;

#if EPICS_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	if ( epics_group == (MX_EPICS_GROUP *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_GROUP pointer passed was NULL." );
	}

	group_id = epics_group->group_id;

#if EPICS_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	epics_status = ca_sg_delete( group_id );

#if EPICS_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, " " );
#endif

	switch( epics_status ) {
	case ECA_NORMAL:
		mx_status = MX_SUCCESSFUL_RESULT;
		break;
	case ECA_BADSYNCGRP:
		mx_status = mx_error( MXE_NOT_FOUND, fname,
		"An attempt to delete EPICS synchronous group %u failed "
		"since no such synchronous group currently exists.",
			(unsigned int) group_id );
		break;
	default:
		mx_status = mx_error( MXE_FUNCTION_FAILED, fname,
		"A failure occurred while trying to delete EPICS "
		"synchronous group %u.  EPICS status = %d, EPICS error = '%s'.",
			group_id, epics_status, ca_message( epics_status ) );
		break;
	}

	return mx_status;
}

/*---*/

static mx_status_type
mx_epics_internal_group_caget( CA_SYNC_GID group_id,
				chid channel_id,
				long epics_type,
				unsigned long num_elements,
				void *data_buffer )
{
	static const char fname[] = "mx_epics_internal_group_caget()";

	READONLY char *pvname;
	int epics_status;
	mx_status_type mx_status;

#if EPICS_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	if ( channel_id == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The channel_id passed was NULL." );
	}
	if ( data_buffer == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The data_buffer passed was NULL." );
	}

	pvname = ca_name( channel_id );

#if EPICS_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	epics_status = ca_sg_array_get( group_id, epics_type,
					num_elements, channel_id,
					data_buffer );

#if EPICS_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, pvname );
#endif

	MX_DEBUG( 2,("%s: ca_sg_array_get() epics_status = %d",
			fname, epics_status));

	switch( epics_status ) {
	case ECA_NORMAL:
		break;
	case ECA_BADCHID:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"ca_sg_array_get() was called with an "
			"illegal EPICS channel id" );
		break;
	case ECA_BADCOUNT:
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"The requested number of elements %lu for "
			"EPICS PV '%s' is larger than the maximum allowed "
			"(%lu) for this process variable.",
				num_elements, ca_name( channel_id ),
			(unsigned long) ca_element_count( channel_id ) );
		break;
	case ECA_BADTYPE:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"An illegal EPICS datatype was requested for "
			"EPICS PV '%s'.", ca_name( channel_id ) );
		break;
	case ECA_BADSYNCGRP:
		return mx_error( MXE_NOT_FOUND, fname,
			"An attempt to use EPICS synchronous group %u failed "
			"since no such synchronous group currently exists.",
				(unsigned int) group_id );
		break;
	case ECA_GETFAIL:
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"An attempt to read the value of local EPICS "
			"database variable '%s' failed.",
				ca_name( channel_id ) );
		break;
	case ECA_DISCONN:
	case ECA_DISCONNCHID:
		return mx_error( MXE_NETWORK_CONNECTION_LOST, fname,
			"The connection to EPICS PV '%s' "
			"has been disconnected.",
			ca_name( channel_id ) );
		break;
	default:
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"A ca_sg_array_get() call failed for EPICS PV '%s'.  "
			"EPICS status = %d, EPICS error = '%s'.",
				ca_name( channel_id ),
				epics_status,
				ca_message( epics_status ) );
		break;
	}

	/* Allow background events to execute. */

	mx_status = mx_epics_poll_for_events( fname );
		
	return mx_status;
}

MX_EXPORT mx_status_type
mx_group_caget( MX_EPICS_GROUP *epics_group,
		MX_EPICS_PV *pv,
		long epics_type,
		unsigned long num_elements,
		void *data_buffer )
{
	static const char fname[] = "mx_group_caget()";

	mx_status_type mx_status;

	if ( epics_group == (MX_EPICS_GROUP *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_GROUP pointer passed was NULL." );
	}

	if ( pv == (MX_EPICS_PV *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_PV pointer passed was NULL." );
	}

	if ( pv->channel_id == NULL ) {
		mx_status = mx_epics_pv_connect( pv );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	mx_status = mx_epics_internal_group_caget( epics_group->group_id,
						pv->channel_id, epics_type,
						num_elements, data_buffer );

	if ( mx_status.code == MXE_NETWORK_CONNECTION_LOST ) {
		(void) ca_clear_channel( pv->channel_id );

		pv->channel_id = NULL;
	}

	return mx_status;
}

/*---*/

static mx_status_type
mx_epics_internal_group_caput( CA_SYNC_GID group_id,
				chid channel_id,
				long epics_type,
				unsigned long num_elements,
				void *data_buffer )
{
	static const char fname[] = "mx_epics_internal_group_caput()";

	READONLY char *pvname;
	int epics_status;
	mx_status_type mx_status;

#if EPICS_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	if ( channel_id == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The channel_id passed was NULL." );
	}
	if ( data_buffer == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The data_buffer passed was NULL." );
	}

	pvname = ca_name( channel_id );

	if ( mx_epics_debug_flag ) {
		if ( num_elements > 1 ) {
			if ( epics_type == MX_CA_STRING ) {
				MX_DEBUG(-2,("%s: sending '%s' to '%s'.",
					fname, (char *) data_buffer, pvname));
			} else {
				MX_DEBUG(-2,
				("%s: sending multielement array to '%s'.",
					fname, pvname));
			}
		} else {
			char byte_value;
			short short_value;
			long long_value;
			float float_value;
			double double_value;
	
			switch( epics_type ) {
			case MX_CA_CHAR:
				byte_value = *((char *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %d to '%s'.",
					fname, byte_value, pvname));
				break;
			case MX_CA_STRING:
				MX_DEBUG(-2,("%s: sending '%s' to '%s'",
					fname, (char *) data_buffer, pvname));
				break;
			case MX_CA_SHORT:
				short_value = *((short *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %hd to '%s'.",
					fname, short_value, pvname));
				break;
			case MX_CA_LONG:
				long_value = *((long *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %ld to '%s'.",
					fname, long_value, pvname));
				break;
			case MX_CA_FLOAT:
				float_value = *((float *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %g to '%s'.",
					fname, float_value, pvname));
				break;
			case MX_CA_DOUBLE:
				double_value = *((double *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %g to '%s'.",
					fname, double_value, pvname));
				break;
			default:
				MX_DEBUG(-2,
				("%s: '%s' has unsupported epics_type %ld",
					fname, pvname, epics_type));
				break;
			}
		}
	}

#if EPICS_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	epics_status = ca_sg_array_put( group_id, epics_type, num_elements,
					channel_id, data_buffer );

#if EPICS_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, pvname );
#endif

	switch( epics_status ) {
	case ECA_NORMAL:
		break;
	case ECA_BADCHID:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"ca_sg_array_put() was called with an "
			"illegal EPICS channel id" );
		break;
	case ECA_BADCOUNT:
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"The requested number of elements %lu for "
			"EPICS PV '%s' is larger than the maximum allowed "
			"(%lu) for this process variable.",
				num_elements, ca_name( channel_id ),
			(unsigned long) ca_element_count( channel_id ) );
		break;
	case ECA_BADTYPE:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"An illegal EPICS datatype was requested for "
			"EPICS PV '%s'.", ca_name( channel_id ) );
		break;
	case ECA_BADSYNCGRP:
		return mx_error( MXE_NOT_FOUND, fname,
			"An attempt to use EPICS synchronous group %u failed "
			"since no such synchronous group currently exists.",
				(unsigned int) group_id );
		break;
	case ECA_STRTOBIG:
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"An unusually long string was supplied for a call "
			"to ca_sg_array_put() for EPICS PV '%s'.",
				ca_name( channel_id ) );
		break;
	case ECA_PUTFAIL:
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"An attempt to write to the value of local EPICS "
			"database variable '%s' failed.",
				ca_name( channel_id ) );
		break;
	case ECA_DISCONN:
	case ECA_DISCONNCHID:
		return mx_error( MXE_NETWORK_CONNECTION_LOST, fname,
			"The connection to EPICS PV '%s' "
			"has been disconnected.",
			ca_name( channel_id ) );
		break;
	default:
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"A ca_sg_array_put() call failed for EPICS PV '%s'.  "
			"EPICS status = %d, EPICS error = '%s'.",
				ca_name( channel_id ),
				epics_status,
				ca_message( epics_status ) );
		break;
	}

	/* Allow background events to execute. */

	mx_status = mx_epics_poll_for_events( fname );
		
	return mx_status;
}

MX_EXPORT mx_status_type
mx_group_caput( MX_EPICS_GROUP *epics_group,
		MX_EPICS_PV *pv,
		long epics_type,
		unsigned long num_elements,
		void *data_buffer )
{
	static const char fname[] = "mx_group_caput()";

	mx_status_type mx_status;

	if ( pv == (MX_EPICS_PV *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_PV pointer passed was NULL." );
	}

	if ( pv->channel_id == NULL ) {
		mx_status = mx_epics_pv_connect( pv );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	mx_status = mx_epics_internal_group_caput( epics_group->group_id,
						pv->channel_id, epics_type,
						num_elements, data_buffer );

	if ( mx_status.code == MXE_NETWORK_CONNECTION_LOST ) {
		(void) ca_clear_channel( pv->channel_id );

		pv->channel_id = NULL;
	}

	return mx_status;
}

/*---*/

MX_EXPORT mx_status_type
mx_epics_convert_mx_type_to_epics_type( long mx_type,
					chtype *epics_type )
{
	static const char fname[] = "mx_epics_convert_mx_type_to_epics_type()";

	switch( mx_type ) {
	case MXFT_STRING:
		*epics_type = DBR_STRING;
		break;
	case MXFT_CHAR:
	case MXFT_UCHAR:
		*epics_type = DBR_CHAR;
		break;
	case MXFT_SHORT:
	case MXFT_USHORT:
		*epics_type = DBR_SHORT;
		break;
	case MXFT_INT:
	case MXFT_UINT:
	case MXFT_LONG:
	case MXFT_ULONG:
		*epics_type = DBR_LONG;
		break;
	case MXFT_FLOAT:
		*epics_type = DBR_FLOAT;
		break;
	case MXFT_DOUBLE:
		*epics_type = DBR_DOUBLE;
		break;
	default:
		*epics_type = -1;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"MX record field type of %ld cannot be converted to an EPICS type.",
			mx_type );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_epics_convert_epics_type_to_mx_type( chtype epics_type,
					long *mx_type )
{
	static const char fname[] = "mx_epics_convert_epics_type_to_mx_type()";

	switch( epics_type ) {
	case DBR_STRING:
		*mx_type = MXFT_STRING;
		break;
	case DBR_CHAR:
		*mx_type = MXFT_CHAR;
		break;
	case DBR_SHORT:
		*mx_type = MXFT_SHORT;
		break;
	case DBR_LONG:
		*mx_type = MXFT_LONG;
		break;
	case DBR_FLOAT:
		*mx_type = MXFT_FLOAT;
		break;
	case DBR_DOUBLE:
		*mx_type = MXFT_DOUBLE;
		break;
	default:
		*mx_type = -1;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"EPICS type of %ld cannot be converted to an MX record field type.",
			*mx_type );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */
