/*
 * Name:    mx_mc_analog_input.c
 *
 * Purpose: MX function library for multichannel analog input devices.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_mc_analog_input.h"

static mx_status_type
mx_multichannel_analog_input_get_pointers( MX_RECORD *record,
		MX_MULTICHANNEL_ANALOG_INPUT **multichannel_analog_input,
		MX_MULTICHANNEL_ANALOG_INPUT_FUNCTION_LIST **flist_ptr,
		const char *calling_fname )
{
	static const char fname[] =
		"mx_multichannel_analog_input_get_pointers()";

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed by '%s' is NULL.",
			calling_fname );
	}

	if ( record->mx_class != MXC_MULTICHANNEL_ANALOG_INPUT ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
"The record '%s' passed by '%s' is not a multichannel analog input record.  "
"(superclass = %ld, class = %ld, type = %ld)",
			record->name, calling_fname,
			record->mx_superclass,
			record->mx_class,
			record->mx_type );
	}

	if (multichannel_analog_input != (MX_MULTICHANNEL_ANALOG_INPUT **) NULL)
	{
		*multichannel_analog_input = (MX_MULTICHANNEL_ANALOG_INPUT *)
						record->record_class_struct;

		if ( *multichannel_analog_input ==
				(MX_MULTICHANNEL_ANALOG_INPUT *) NULL )
		{
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_MULTICHANNEL_ANALOG_INPUT pointer for record '%s' passed by '%s' is NULL.",
				record->name, calling_fname );
		}
	}

	if (flist_ptr != (MX_MULTICHANNEL_ANALOG_INPUT_FUNCTION_LIST **) NULL)
	{
		*flist_ptr = (MX_MULTICHANNEL_ANALOG_INPUT_FUNCTION_LIST *)
				(record->class_specific_function_list);

		if ( *flist_ptr ==
			(MX_MULTICHANNEL_ANALOG_INPUT_FUNCTION_LIST *) NULL )
		{
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MULTICHANNEL_ANALOG_INPUT_FUNCTION_LIST ptr for "
			"record '%s' passed by '%s' is NULL.",
				record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_multichannel_analog_input_initialize_type( long record_type,
			long *num_record_fields,
			MX_RECORD_FIELD_DEFAULTS **record_field_defaults,
			long *maximum_num_channels_varargs_cookie )
{
	static const char fname[] = "mx_multichannel_analog_input_initialize_type()";

	MX_DRIVER *driver;
	MX_RECORD_FIELD_DEFAULTS **record_field_defaults_ptr;
	MX_RECORD_FIELD_DEFAULTS *field;
	long referenced_field_index;
	mx_status_type status;

	if ( num_record_fields == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"num_record_fields pointer passed was NULL." );
	}
	if ( record_field_defaults == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"record_field_defaults pointer passed was NULL." );
	}
	if ( maximum_num_channels_varargs_cookie == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
	"maximum_num_channels_varargs_cookie pointer passed was NULL." );
	}

	driver = mx_get_driver_by_type( mx_type_list, record_type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record type %ld not found in mx_type_list.",
			record_type );
	}

	record_field_defaults_ptr
			= driver->record_field_defaults_ptr;

	if (record_field_defaults_ptr == (MX_RECORD_FIELD_DEFAULTS **) NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	*record_field_defaults = *record_field_defaults_ptr;

	if ( *record_field_defaults == (MX_RECORD_FIELD_DEFAULTS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	if ( driver->num_record_fields == (long *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'num_record_fields' pointer for record type '%s' is NULL.",
			driver->name );
	}

	*num_record_fields = *(driver->num_record_fields);

	/*** Construct a varargs cookie for 'maximum_num_channels'. ***/

	status = mx_find_record_field_defaults_index(
			*record_field_defaults, *num_record_fields,
			"maximum_num_channels", &referenced_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie( referenced_field_index, 0,
					maximum_num_channels_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	/*** 'channel_array' depends on 'maximum_num_channels'. ***/

	status = mx_find_record_field_defaults( *record_field_defaults,
			*num_record_fields, "channel_array", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	field->dimension[0] = *maximum_num_channels_varargs_cookie;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_multichannel_analog_input_finish_record_initialization( MX_RECORD *multichannel_analog_input_record )
{
	static const char fname[] =
		"mx_multichannel_analog_input_finish_record_initialization()";

	MX_MULTICHANNEL_ANALOG_INPUT *multichannel_analog_input;
	MX_RECORD_FIELD *channel_array_field;
	mx_status_type mx_status;

	if ( multichannel_analog_input_record == ( MX_RECORD * ) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	multichannel_analog_input = (MX_MULTICHANNEL_ANALOG_INPUT *)
		multichannel_analog_input_record->record_class_struct;

	if ( multichannel_analog_input ==
			(MX_MULTICHANNEL_ANALOG_INPUT *) NULL )
	{
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MULTICHANNEL_ANALOG_INPUT pointer for record '%s' is NULL.",
			multichannel_analog_input_record->name );
	}

	/* Since 'channel_array' is a varying length field, but is not
	 * specified in the record description, we must fix up the
	 * dimension[0] value for the field by hand.
	 */

	mx_status = mx_find_record_field( multichannel_analog_input_record,
						"channel_array",
						&channel_array_field );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	channel_array_field->dimension[0] =
		(long) multichannel_analog_input->maximum_num_channels;

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mx_multichannel_analog_input_read( MX_RECORD *record,
					unsigned long *num_channels,
					double **channel_array )
{
	static const char fname[] = "mx_multichannel_analog_input_read()";

	MX_MULTICHANNEL_ANALOG_INPUT *multichannel_analog_input;
	MX_MULTICHANNEL_ANALOG_INPUT_FUNCTION_LIST *function_list;
	mx_status_type ( *read_fn ) ( MX_MULTICHANNEL_ANALOG_INPUT * );
	mx_status_type mx_status;

	mx_status = mx_multichannel_analog_input_get_pointers( record,
			&multichannel_analog_input, &function_list, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	read_fn = function_list->read;

	if ( read_fn == NULL ){
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"read function ptr for MX_MULTICHANNEL_ANALOG_INPUT ptr 0x%p is NULL.",
			multichannel_analog_input);
	}

	mx_status = (*read_fn)( multichannel_analog_input );

	if ( num_channels != NULL ) {
		*num_channels = multichannel_analog_input->current_num_channels;
	}

	if ( channel_array != NULL ) {
		*channel_array = multichannel_analog_input->channel_array;
	}

	return mx_status;
}

