/*
 * Name:    mx_net.c
 *
 * Purpose: Client side part of MX network protocol.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 1999-2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>

#include "mx_osdef.h"
#include "mx_util.h"
#include "mx_types.h"
#include "mx_array.h"
#include "mx_bit.h"
#include "mx_record.h"
#include "mx_socket.h"
#include "mx_net.h"

#if ( HAVE_TCPIP == 0 )
#     define htonl(x) (x)
#     define ntohl(x) (x)
#endif

#define NETWORK_DEBUG		FALSE

#define NETWORK_DEBUG_TIMING	FALSE

#if NETWORK_DEBUG_TIMING
#include "mx_hrt_debug.h"
#endif

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_network_receive_message( MX_RECORD *server_record,
				unsigned long buffer_length, void *buffer )
{
	static const char fname[] = "mx_network_receive_message()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_SERVER_FUNCTION_LIST *function_list;
	mx_status_type ( *fptr ) ( MX_NETWORK_SERVER *, unsigned long, void * );
	mx_status_type mx_status;

	if ( server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Record pointer passed was NULL." );
	}
	if ( buffer == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Message buffer pointer passed for record '%s' was NULL.",
			server_record->name );
	}

	server = (MX_NETWORK_SERVER *) (server_record->record_class_struct);

	if ( server == (MX_NETWORK_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_SERVER pointer for server record '%s' is NULL.",
			server_record->name );
	}

	function_list = (MX_NETWORK_SERVER_FUNCTION_LIST *)
			(server_record->class_specific_function_list);

	if ( function_list == (MX_NETWORK_SERVER_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_SERVER_FUNCTION_LIST pointer for server record '%s' is NULL.",
			server_record->name );
	}

	fptr = function_list->receive_message;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"receive_message function pointer for server record '%s' is NULL.",
			server_record->name );
	}

	mx_status = ( *fptr ) ( server, buffer_length, buffer );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_network_send_message( MX_RECORD *server_record, void *buffer )
{
	static const char fname[] = "mx_network_send_message()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_SERVER_FUNCTION_LIST *function_list;
	mx_status_type ( *fptr ) ( MX_NETWORK_SERVER *, void * );
	mx_status_type mx_status;

	if ( server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Record pointer passed was NULL." );
	}
	if ( buffer == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Message buffer pointer passed for record '%s' was NULL.",
			server_record->name );
	}

	server = (MX_NETWORK_SERVER *) (server_record->record_class_struct);

	if ( server == (MX_NETWORK_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_SERVER pointer for server record '%s' is NULL.",
			server_record->name );
	}

	function_list = (MX_NETWORK_SERVER_FUNCTION_LIST *)
			(server_record->class_specific_function_list);

	if ( function_list == (MX_NETWORK_SERVER_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_SERVER_FUNCTION_LIST pointer for server record '%s' is NULL.",
			server_record->name );
	}

	fptr = function_list->send_message;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"send_message function pointer for server record '%s' is NULL.",
			server_record->name );
	}

	mx_status = ( *fptr ) ( server, buffer );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_network_reconnect_if_down( MX_RECORD *server_record )
{
	static const char fname[] = "mx_network_reconnect_if_down()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_SERVER_FUNCTION_LIST *function_list;
	mx_status_type ( *fptr ) ( MX_NETWORK_SERVER * );
	mx_status_type mx_status;

	if ( server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Record pointer passed was NULL." );
	}

	server = (MX_NETWORK_SERVER *) (server_record->record_class_struct);

	if ( server == (MX_NETWORK_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_SERVER pointer for server record '%s' is NULL.",
			server_record->name );
	}

	function_list = (MX_NETWORK_SERVER_FUNCTION_LIST *)
			(server_record->class_specific_function_list);

	if ( function_list == (MX_NETWORK_SERVER_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_SERVER_FUNCTION_LIST pointer for server record '%s' is NULL.",
			server_record->name );
	}

	fptr = function_list->reconnect_if_down;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"reconnect_if_down function pointer for server record '%s' is NULL.",
			server_record->name );
	}

	mx_status = ( *fptr ) ( server );

	return mx_status;
}

MX_EXPORT void
mx_network_display_message_buffer( void *buffer_ptr )
{
	static const char fname[] = "mx_network_display_message_buffer()";

	mx_uint32_type *header;
	unsigned char *buffer, *message;
	mx_uint32_type magic_number, header_length, message_length;
	mx_uint32_type message_type, status_code, i;
	unsigned char c;

	if ( buffer_ptr == NULL ) {
		(void) mx_error( MXE_NULL_ARGUMENT, fname,
			"Buffer pointer passed was NULL." );
		return;
	}

	buffer = buffer_ptr;
	header = buffer_ptr;

	magic_number   = ntohl( header[ MX_NETWORK_MAGIC ] );
	header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
	message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
	message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );
	status_code    = ntohl( header[ MX_NETWORK_STATUS_CODE ] );

	message = buffer + header_length;

	/**** Print out the header. ****/

	/* Header magic number */

	mx_info( "Magic number:     0x%8lx   (%2x %2x %2x %2x)",
	  (long) magic_number, buffer[0], buffer[1], buffer[2], buffer[3] );

	if ( magic_number != MX_NETWORK_MAGIC_VALUE ) {
		(void) mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Invalid magic number 0x%lx in network buffer header.",
			(long) magic_number );
		return;
	}

	/* Header length */

	i = MX_NETWORK_HEADER_LENGTH * sizeof( mx_uint32_type );

	mx_info( "Header length:  %12lu   (%2x %2x %2x %2x)",
		header_length, buffer[i+0], buffer[i+1],
			buffer[i+2], buffer[i+3] );

	if ( header_length < MX_NETWORK_HEADER_LENGTH_VALUE ) {
		mx_info( "*** Header length %lu was unexpectedly short. ***",
			header_length );
		return;
	}

	/* Message length */

	i = MX_NETWORK_MESSAGE_LENGTH * sizeof( mx_uint32_type );

	mx_info( "Message length: %12lu   (%2x %2x %2x %2x)",
		message_length, buffer[i+0], buffer[i+1],
			buffer[i+2], buffer[i+3] );

	/* Message type */

	i = MX_NETWORK_MESSAGE_TYPE * sizeof( mx_uint32_type );

	mx_info( "Message type:   %12lu   (%2x %2x %2x %2x)",
		message_type, buffer[i+0], buffer[i+1],
			buffer[i+2], buffer[i+3] );

	/* Status code */

	i = MX_NETWORK_STATUS_CODE * sizeof( mx_uint32_type );

	mx_info( "Status code:    %12lu   (%2x %2x %2x %2x)",
		status_code, buffer[i+0], buffer[i+1],
			buffer[i+2], buffer[i+3] );

	/* Remaining part of header. */

	if ( header_length > MX_NETWORK_HEADER_LENGTH_VALUE ) {
		mx_info( "Remaining header bytes:" );

		for ( i += sizeof( mx_uint32_type ); i < header_length; i++ ) {
			c = buffer[i];

			if ( isprint(c) ) {
				mx_info("  buffer[%lu]  = %02x (%c)",
						i, c, c);
			} else {
				mx_info("  buffer[%lu]  = %02x", i, c);
			}
		}
	}

	/* Print out the message body. */

	mx_info( "Message contents:" );

	for ( i = 0; i < message_length; i++ ) {
		c = message[i];

		if ( isprint(c) ) {
			mx_info("  message[%lu] = %02x (%c)", i, c, c);
		} else {
			mx_info("  message[%lu] = %02x", i, c);
		}
	}
	return;
}

/* ====================================================================== */

static mx_status_type
mx_network_field_get_parameters( MX_RECORD *server_record,
			MX_RECORD_FIELD *local_field,
			long *datatype,
			long *num_dimensions,
			long **dimension_array,
			size_t **data_element_size_array,
			MX_NETWORK_SERVER **server,
			MX_NETWORK_SERVER_FUNCTION_LIST **function_list,
			const char *calling_fname )
{
	static const char fname[] = "mx_network_field_get_parameters()";

	if ( server_record == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"server_record pointer passed to '%s' is NULL.",
			calling_fname );
	}
	if ( local_field == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"local_field pointer passed to '%s' is NULL.",
			calling_fname );
	}
	*datatype = local_field->datatype;

	switch( *datatype ) {
	case MXFT_STRING:
	case MXFT_CHAR:
	case MXFT_UCHAR:
	case MXFT_SHORT:
	case MXFT_USHORT:
	case MXFT_INT:
	case MXFT_UINT:
	case MXFT_LONG:
	case MXFT_ULONG:
	case MXFT_FLOAT:
	case MXFT_DOUBLE:
	case MXFT_HEX:
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Data type %ld passed to '%s' is not a legal network datatype.",
			*datatype, calling_fname );
		break;
	}

	*num_dimensions = local_field->num_dimensions;
	*dimension_array = local_field->dimension;
	*data_element_size_array = local_field->data_element_size;

	*server = (MX_NETWORK_SERVER *) (server_record->record_class_struct);
	*function_list = (MX_NETWORK_SERVER_FUNCTION_LIST *)
				(server_record->class_specific_function_list);

	return MX_SUCCESSFUL_RESULT;
}

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_get_array_main( MX_RECORD *server_record,
		char *remote_record_field_name,
		MX_NETWORK_HANDLE *handle,
		long datatype,
		long num_dimensions,
		long *dimension,
		void *value_ptr )
{
	static const char fname[] = "mx_get_array_main()";

	MX_RECORD_FIELD local_temp_record_field;
	long *dimension_array;
	long local_dimension_array[MXU_FIELD_MAX_DIMENSIONS];
	size_t data_element_size[MXU_FIELD_MAX_DIMENSIONS];
	mx_status_type mx_status;

	if ( dimension != NULL ) {
		dimension_array = dimension;

	} else if ( num_dimensions == 0 ) {
		dimension_array = local_dimension_array;
		dimension_array[0] = 0;
	} else {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"dimension array pointer is NULL, but num_dimensions (%ld) is greater than 0",
			num_dimensions );
	}

	/* Setting the first element of data_element_size to be zero causes
	 * mx_construct_temp_record_field() to use a default data element
	 * size array.
	 */

	data_element_size[0] = 0L;

	/* Setup a temporary record field to be used by
	 * mx_get_field_array().
	 */

	mx_status = mx_construct_temp_record_field( &local_temp_record_field,
			datatype, num_dimensions, dimension_array,
			data_element_size, value_ptr );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Send the request to the server. */

	mx_status = mx_get_field_array( server_record,
			remote_record_field_name,
			handle,
			&local_temp_record_field,
			value_ptr );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_put_array_main( MX_RECORD *server_record,
		char *remote_record_field_name,
		MX_NETWORK_HANDLE *handle,
		long datatype,
		long num_dimensions,
		long *dimension,
		void *value_ptr )
{
	static const char fname[] = "mx_put_array_main()";

	MX_RECORD_FIELD local_temp_record_field;
	long *dimension_array;
	long local_dimension_array[MXU_FIELD_MAX_DIMENSIONS];
	size_t data_element_size[MXU_FIELD_MAX_DIMENSIONS];
	mx_status_type mx_status;

	if ( dimension != NULL ) {
		dimension_array = dimension;

	} else if ( num_dimensions == 0 ) {
		dimension_array = local_dimension_array;
		dimension_array[0] = 0;
	} else {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"dimension array pointer is NULL, but num_dimensions (%ld) is greater than 0",
			num_dimensions );
	}

	/* Setting the first element of data_element_size to be zero causes
	 * mx_construct_temp_record_field() to use a default data element
	 * size array.
	 */

	data_element_size[0] = 0L;

	/* Setup a temporary record field to be used by
	 * mx_put_field_array().
	 */

	mx_status = mx_construct_temp_record_field( &local_temp_record_field,
			datatype, num_dimensions, dimension_array,
			data_element_size, value_ptr );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Send the request to the server. */

	mx_status = mx_put_field_array( server_record,
			remote_record_field_name,
			handle,
			&local_temp_record_field,
			value_ptr );

	return mx_status;
}

/* ====================================================================== */

#define MXU_GET_PUT_ARRAY_ASCII_LOCATION_LENGTH \
	MXU_HOSTNAME_LENGTH + MXU_RECORD_FIELD_NAME_LENGTH + 80

static mx_status_type
mx_get_array_ascii_error_message( mx_uint32_type status_code,
			char *server_name,
			char *record_field_name,
			char *error_message )
{
	char location[ MXU_GET_PUT_ARRAY_ASCII_LOCATION_LENGTH + 1 ];

	sprintf( location,
		"MX GET from server '%s', record field '%s'",
		server_name, record_field_name );

	return mx_error( (long) status_code, location, error_message );
}

static mx_status_type
mx_put_array_ascii_error_message( mx_uint32_type status_code,
			char *server_name,
			char *record_field_name,
			char *error_message )
{
	char location[ MXU_GET_PUT_ARRAY_ASCII_LOCATION_LENGTH + 1 ];

	sprintf( location,
		"MX PUT to server '%s', record field '%s'",
		server_name, record_field_name );

	return mx_error( (long) status_code, location, error_message );
}

MX_EXPORT mx_status_type
mx_get_field_array( MX_RECORD *server_record,
			char *remote_record_field_name,
			MX_NETWORK_HANDLE *handle,
			MX_RECORD_FIELD *local_field,
			void *value_ptr )
{
	static const char fname[] = "mx_get_field_array()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_SERVER_FUNCTION_LIST *function_list;
	MX_RECORD_FIELD_PARSE_STATUS temp_parse_status;
	mx_status_type ( *token_parser )
		(void *, char *, MX_RECORD *, MX_RECORD_FIELD *,
		MX_RECORD_FIELD_PARSE_STATUS *);
	long datatype, num_dimensions, *dimension_array;
	size_t *data_element_size_array;
	int array_is_dynamically_allocated;

	MX_NETWORK_MESSAGE_BUFFER *aligned_buffer;
	mx_uint32_type *header, *uint32_message;
	char *buffer;
	char *message;
	mx_uint32_type header_length, message_length;
	mx_uint32_type message_type, status_code;
	mx_status_type mx_status;
	char token_buffer[500];

#if NETWORK_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	static char separators[] = MX_RECORD_FIELD_SEPARATORS;

	mx_status = mx_network_field_get_parameters(
			server_record, local_field,
			&datatype, &num_dimensions,
			&dimension_array, &data_element_size_array,
			&server, &function_list, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,
	("%s: server_record = '%s', remote_record_field_name = '%s'",
		fname, server_record->name, remote_record_field_name));

	if ( local_field->flags & MXFF_VARARGS ) {
		array_is_dynamically_allocated = TRUE;
	} else {
		array_is_dynamically_allocated = FALSE;
	}

	mx_status = mx_network_reconnect_if_down( server_record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/*********** Send a 'get array' command. *************/

	aligned_buffer = server->message_buffer;

	header = &(aligned_buffer->uint32_buffer[0]);
	buffer = &(aligned_buffer->char_buffer[0]);

	header[MX_NETWORK_MAGIC] = htonl( MX_NETWORK_MAGIC_VALUE );
	header[MX_NETWORK_HEADER_LENGTH]
				= htonl( MX_NETWORK_HEADER_LENGTH_VALUE );
	header[MX_NETWORK_STATUS_CODE] = htonl( MXE_SUCCESS );

	message = buffer + MX_NETWORK_HEADER_LENGTH_VALUE;

	uint32_message = header + MX_NETWORK_NUM_HEADER_VALUES;

	if ( handle == (MX_NETWORK_HANDLE *) NULL ) {

		/* Use ASCII record field name */

		header[MX_NETWORK_MESSAGE_TYPE]
				= htonl( MX_NETMSG_GET_ARRAY_BY_NAME );

		mx_strncpy( message, remote_record_field_name,
				MXU_RECORD_FIELD_NAME_LENGTH );

		message_length = strlen( message ) + 1;
	} else {

		/* Use binary network handle */

		header[MX_NETWORK_MESSAGE_TYPE]
				= htonl( MX_NETMSG_GET_ARRAY_BY_HANDLE );

		uint32_message[0] = htonl( handle->record_handle );
		uint32_message[1] = htonl( handle->field_handle );

		message_length = 2 * sizeof( mx_uint32_type );
	}

	header[MX_NETWORK_MESSAGE_LENGTH] = htonl( message_length );

#if NETWORK_DEBUG
	mx_network_display_message_buffer( buffer );
#endif

#if NETWORK_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	mx_status = mx_network_send_message( server_record, buffer );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/************* Wait for the response. **************/

	mx_status = mx_network_receive_message( server_record,
				MX_NETWORK_MAXIMUM_MESSAGE_SIZE, buffer );

#if NETWORK_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, remote_record_field_name );
#endif

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if NETWORK_DEBUG
	mx_network_display_message_buffer( buffer );
#endif

	header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
	message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
	message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );
	status_code    = ntohl( header[ MX_NETWORK_STATUS_CODE ] );

	if ( message_type != mx_server_response(MX_NETMSG_GET_ARRAY_BY_NAME) ) {

		if ( message_type == MX_NETMSG_UNEXPECTED_ERROR ) {
			return mx_get_array_ascii_error_message(
					status_code,
					server_record->name,
					remote_record_field_name,
					message );
		}

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Message type for response was not MX_NETMSG_GET_ARRAY_ASCII_RESPONSE.  "
"Instead it was of type = %#lx", message_type );
	}

	message = buffer + header_length;

	/* If the remote command failed, the message field will include
	 * the text of the error message rather than the array data
	 * we wanted.
	 */

	if ( status_code != MXE_SUCCESS ) {
		return mx_get_array_ascii_error_message(
				status_code,
				server_record->name,
				remote_record_field_name,
				message );
	}

	/************ Parse the data that was returned. ***************/

	switch( server->data_format ) {
	case MX_NETWORK_DATAFMT_ASCII:

		/* The data was returned using the ASCII MX database format. */

		mx_status = mx_get_token_parser( datatype, &token_parser );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mx_initialize_parse_status( &temp_parse_status,
						message, separators );

		/* If this is a string field, figure out what the maximum
		 * string token length is.
		 */

		if ( datatype == MXFT_STRING ) {
			temp_parse_status.max_string_token_length =
				mx_get_max_string_token_length( local_field );
		} else {
			temp_parse_status.max_string_token_length = 0L;
		}

		/* Now we are ready to parse the tokens. */

		if ( (num_dimensions == 0) ||
			((datatype == MXFT_STRING) && (num_dimensions == 1)) )
		{
			mx_status = mx_get_next_record_token(
				&temp_parse_status,
				token_buffer, sizeof(token_buffer) );

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			mx_status = ( *token_parser ) ( value_ptr,
					token_buffer, NULL, local_field,
					&temp_parse_status );
		} else {
			mx_status = mx_parse_array_description( value_ptr,
				num_dimensions - 1, NULL, local_field,
				&temp_parse_status, token_parser );
		}
		break;

	case MX_NETWORK_DATAFMT_RAW:
		mx_status = mx_copy_buffer_to_array( message, message_length,
				value_ptr, array_is_dynamically_allocated,
				datatype, num_dimensions,
				dimension_array, data_element_size_array,
				NULL );

		if ( mx_status.code != MXE_SUCCESS ) {
			(void) mx_error( mx_status.code, fname,
	"Buffer copy for MX_GET of parameter '%s' in server '%s' failed.",
				remote_record_field_name, server_record->name );
		}
		break;

	case MX_NETWORK_DATAFMT_XDR:
#if HAVE_XDR
		mx_status = mx_xdr_data_transfer( MX_XDR_DECODE,
				value_ptr, array_is_dynamically_allocated,
				datatype, num_dimensions,
				dimension_array, data_element_size_array,
				message, message_length, NULL );

		if ( mx_status.code != MXE_SUCCESS ) {
			(void) mx_error( mx_status.code, fname,
	"Buffer copy for MX_GET of parameter '%s' in server '%s' failed.",
				remote_record_field_name, server_record->name );
		}
#else
		return mx_error( MXE_UNSUPPORTED, fname,
			"XDR network data format is not supported "
			"on this system." );
#endif
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unrecognized network data format type %lu was requested.",
			server->data_format );
		break;
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mx_put_field_array( MX_RECORD *server_record,
			char *remote_record_field_name,
			MX_NETWORK_HANDLE *handle,
			MX_RECORD_FIELD *local_field,
			void *value_ptr )
{
	static const char fname[] = "mx_put_field_array()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_SERVER_FUNCTION_LIST *function_list;
	mx_status_type ( *token_constructor )
		(void *, char *, size_t, MX_RECORD *, MX_RECORD_FIELD *);
	long datatype, num_dimensions, *dimension_array;
	size_t *data_element_size_array;
	int array_is_dynamically_allocated;

	MX_NETWORK_MESSAGE_BUFFER *aligned_buffer;
	mx_uint32_type *header, *uint32_message;
	char *buffer;
	char *message, *ptr;
	unsigned long i, ptr_address, remainder, gap_size;
	mx_uint32_type header_length, message_length, message_type, status_code;
	size_t buffer_left, num_bytes_copied;
	mx_status_type mx_status;

#if NETWORK_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	mx_status = mx_network_field_get_parameters(
			server_record, local_field,
			&datatype, &num_dimensions,
			&dimension_array, &data_element_size_array,
			&server, &function_list, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( local_field->flags & MXFF_VARARGS ) {
		array_is_dynamically_allocated = TRUE;
	} else {
		array_is_dynamically_allocated = FALSE;
	}

	mx_status = mx_network_reconnect_if_down( server_record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/************ Construct a 'put array' command. *************/

	aligned_buffer = server->message_buffer;

	header = &(aligned_buffer->uint32_buffer[0]);
	buffer = &(aligned_buffer->char_buffer[0]);

	header[MX_NETWORK_MAGIC] = htonl( MX_NETWORK_MAGIC_VALUE );
	header[MX_NETWORK_HEADER_LENGTH]
				= htonl( MX_NETWORK_HEADER_LENGTH_VALUE );
	header[MX_NETWORK_STATUS_CODE] = htonl( MXE_SUCCESS );

	message = buffer + MX_NETWORK_HEADER_LENGTH_VALUE;

	uint32_message = header + MX_NETWORK_NUM_HEADER_VALUES;

	if ( handle == (MX_NETWORK_HANDLE *) NULL ) {

		/* Use ASCII record field name */

		header[MX_NETWORK_MESSAGE_TYPE]
				= htonl( MX_NETMSG_PUT_ARRAY_BY_NAME );

		sprintf( message, "%*s", -MXU_RECORD_FIELD_NAME_LENGTH,
				remote_record_field_name );

		message_length = MXU_RECORD_FIELD_NAME_LENGTH;
	} else {

		/* Use binary network handle */

		header[MX_NETWORK_MESSAGE_TYPE]
				= htonl( MX_NETMSG_PUT_ARRAY_BY_HANDLE );

		uint32_message[0] = htonl( handle->record_handle );
		uint32_message[1] = htonl( handle->field_handle );

		message_length = 2 * sizeof( mx_uint32_type );
	}

	ptr = message + message_length;

	MX_DEBUG( 2,("%s: message = %p, ptr = %p, message_length = %lu",
		fname, message, ptr, message_length));

	buffer_left = MX_NETWORK_MAXIMUM_MESSAGE_SIZE
			- MX_NETWORK_HEADER_LENGTH_VALUE - message_length;

	/* Construct the data to send. */

	switch( server->data_format ) {
	case MX_NETWORK_DATAFMT_ASCII:

		/* Send the data using the ASCII MX database format. */

		mx_status = mx_get_token_constructor( datatype,
						&token_constructor );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		if ( (num_dimensions == 0)
		  || ((datatype == MXFT_STRING) && (num_dimensions == 1)) )
		{
			mx_status = (*token_constructor) ( value_ptr, ptr,
			    MX_NETWORK_MAXIMUM_MESSAGE_SIZE - ( ptr - buffer ),
			    NULL, local_field );
		} else {
			mx_status = mx_create_array_description( value_ptr, 
			    local_field->num_dimensions - 1, ptr,
			    MX_NETWORK_MAXIMUM_MESSAGE_SIZE - ( ptr - buffer ),
			    NULL, local_field, token_constructor );
		}

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* The extra 1 is for the '\0' byte at the end of the string. */

		message_length += 1 + strlen( ptr );

		break;

	case MX_NETWORK_DATAFMT_RAW:
		mx_status = mx_copy_array_to_buffer( value_ptr,
				array_is_dynamically_allocated,
				datatype, num_dimensions,
				dimension_array, data_element_size_array,
				ptr, buffer_left,
				&num_bytes_copied );

		if ( mx_status.code != MXE_SUCCESS ) {
			(void) mx_error( mx_status.code, fname,
	"Buffer copy for MX_PUT of parameter '%s' in server '%s' failed.",
				remote_record_field_name, server_record->name );
		}

		message_length += num_bytes_copied;
		break;

	case MX_NETWORK_DATAFMT_XDR:
#if HAVE_XDR
		/* The XDR data pointer 'ptr' must be aligned on a 4 byte
		 * address boundary for XDR data conversion to work correctly
		 * on all architectures.  If the pointer does not point to
		 * an address that is a multiple of 4 bytes, we move it to
		 * the next address that _is_ and fill the bytes inbetween
		 * with zeros.
		 */

		ptr_address = (unsigned long) ptr;

		remainder = ptr_address % 4;

		if ( remainder != 0 ) {
			gap_size = 4 - remainder;

			for ( i = 0; i < gap_size; i++ ) {
				ptr[i] = '\0';
			}

			ptr += gap_size;

			message_length += gap_size;
		}

		/* Now we are ready to do the XDR data conversion. */

		mx_status = mx_xdr_data_transfer( MX_XDR_ENCODE,
				value_ptr,
				array_is_dynamically_allocated,
				datatype, num_dimensions,
				dimension_array, data_element_size_array,
				ptr, buffer_left,
				&num_bytes_copied );

		if ( mx_status.code != MXE_SUCCESS ) {
			(void) mx_error( mx_status.code, fname,
	"Buffer copy for MX_PUT of parameter '%s' in server '%s' failed.",
				remote_record_field_name, server_record->name );
		}

		message_length += num_bytes_copied;
#else
		return mx_error( MXE_UNSUPPORTED, fname,
			"XDR network data format is not supported "
			"on this system." );
#endif
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unrecognized network data format type %lu was requested.",
			server->data_format );
		break;
	}

	header[MX_NETWORK_MESSAGE_LENGTH] = htonl( message_length );

	message_length += MX_NETWORK_HEADER_LENGTH_VALUE;

	MX_DEBUG( 2,("%s: message = '%s'", fname, message));

#if NETWORK_DEBUG
	mx_network_display_message_buffer( buffer );
#endif

#if NETWORK_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	/*************** Send the message. **************/

	mx_status = mx_network_send_message( server_record, buffer );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/************** Wait for the response. ************/

	mx_status = mx_network_receive_message( server_record,
				MX_NETWORK_MAXIMUM_MESSAGE_SIZE, buffer );

#if NETWORK_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, remote_record_field_name );
#endif

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if NETWORK_DEBUG
	mx_network_display_message_buffer( buffer );
#endif

        header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
        message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
        message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );
        status_code = ntohl( header[ MX_NETWORK_STATUS_CODE ] );

        if ( message_type != mx_server_response(MX_NETMSG_PUT_ARRAY_BY_NAME) ) {

		if ( message_type == MX_NETMSG_UNEXPECTED_ERROR ) {
			return mx_put_array_ascii_error_message(
					status_code,
					server_record->name,
					remote_record_field_name,
					message );
		}

                return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Message type for response was not MX_NETMSG_PUT_ARRAY_ASCII_RESPONSE.  "
"Instead it was of type = %#lx", message_type );
        }

        message = buffer + header_length;

        /* If the remote command failed, the message field will include
         * the text of the error message rather than the array data
         * we wanted.
         */

        if ( status_code != MXE_SUCCESS ) {
		return mx_put_array_ascii_error_message(
				status_code,
				server_record->name,
				remote_record_field_name,
				message );
        } else {
		return MX_SUCCESSFUL_RESULT;
	}
}

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_get_field_type( MX_RECORD *server_record, 
			char *remote_record_field_name,
			long max_dimensions,
			long *datatype,
			long *num_dimensions,
			long *dimension_array )
{
	static const char fname[] = "mx_get_field_type()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_MESSAGE_BUFFER *aligned_buffer;
	mx_uint32_type *header;
	char *buffer, *message;
	mx_uint32_type *message_uint32_array;
	mx_uint32_type header_length, message_length;
	mx_uint32_type message_type, status_code;
	mx_uint32_type i, expected_message_length;
	mx_uint32_type header_length_in_32bit_words;
	mx_status_type mx_status;

#if NETWORK_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	MX_DEBUG( 2,("%s invoked, remote_record_field_name = '%s'",
			fname, remote_record_field_name));

	if ( server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"server_record argument passed was NULL." );
	}

	server = (MX_NETWORK_SERVER *) server_record->record_class_struct;

	if ( server == (MX_NETWORK_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_SERVER pointer for server record '%s' is NULL.",
			server_record->name );
	}

	mx_status = mx_network_reconnect_if_down( server_record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/************ Send the 'get field type' message. *************/

	aligned_buffer = server->message_buffer;

	header = &(aligned_buffer->uint32_buffer[0]);
	buffer = &(aligned_buffer->char_buffer[0]);

	header[MX_NETWORK_MAGIC] = htonl( MX_NETWORK_MAGIC_VALUE );
	header[MX_NETWORK_HEADER_LENGTH]
				= htonl( MX_NETWORK_HEADER_LENGTH_VALUE );
	header[MX_NETWORK_MESSAGE_TYPE] = htonl( MX_NETMSG_GET_FIELD_TYPE );
	header[MX_NETWORK_STATUS_CODE] = htonl( MXE_SUCCESS );

	message = buffer + MX_NETWORK_HEADER_LENGTH_VALUE;

	mx_strncpy( message, remote_record_field_name,
					MXU_RECORD_FIELD_NAME_LENGTH );

	message_length = strlen( message ) + 1;
	header[MX_NETWORK_MESSAGE_LENGTH] = htonl( message_length );

#if NETWORK_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	mx_status = mx_network_send_message( server_record, buffer );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/************** Wait for the response. **************/

	mx_status = mx_network_receive_message( server_record,
				MX_NETWORK_MAXIMUM_MESSAGE_SIZE, buffer );

#if NETWORK_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, remote_record_field_name );
#endif

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
	message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
	message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );
	status_code = ntohl( header[ MX_NETWORK_STATUS_CODE ] );

	if ( message_type != mx_server_response( MX_NETMSG_GET_FIELD_TYPE ) ) {

		if ( message_type == MX_NETMSG_UNEXPECTED_ERROR ) {
			return mx_error( (long)status_code, fname,
				"%s", message );
		}

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Message type for response was not MX_NETMSG_GET_FIELD_TYPE_RESPONSE.  "
"Instead it was of type = %#lx", message_type );
	}

#if 0
	mx_network_display_message_buffer( buffer );
#endif

	/* If the remote command failed, the message field will include
	 * the text of the error message rather than the field type
	 * data we requested.
	 */

	if ( status_code != MXE_SUCCESS ) {
		return mx_error( (long)status_code, fname, message );
	}

	if ( message_length < (2 * sizeof(mx_uint32_type)) ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Incomplete message received.  Message length = %ld",
			message_length );
	}

	/***** Get the type information out of what we were sent. *****/

	header_length_in_32bit_words = header_length / sizeof(mx_uint32_type);

	message_uint32_array = header + header_length_in_32bit_words;

	*datatype = (long) ntohl( (unsigned long) message_uint32_array[0] );
	*num_dimensions
		= (long) ntohl( (unsigned long) message_uint32_array[1] );

	expected_message_length = sizeof(long) * ( (*num_dimensions) + 2 );

	if ( message_length < expected_message_length ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Incomplete dimension array received.  %ld dimension values were received, "
"but %ld dimension values were expected.",
			message_length - 2, *num_dimensions );
	}
	if ( *num_dimensions < max_dimensions ) {
		max_dimensions = *num_dimensions;
	}
	if ( max_dimensions <= 0 ) {
		dimension_array[0] = (long)
			ntohl( (unsigned long) message_uint32_array[2] );
	}
	for ( i = 0; i < max_dimensions; i++ ) {
		dimension_array[i] = (long)
			ntohl( (unsigned long) message_uint32_array[i+2] );
	}
	return MX_SUCCESSFUL_RESULT;
}

/* ====================================================================== */

/* mx_set_client_info() sends information to the server about the client
 * for debugging and status reporting purposes _only_.  No authentication
 * is done, so this mechanism cannot and _must_ _not_ be used for access
 * control.  Someday we may have SSL/Kerberos/whatever-based authentication,
 * but this function does not do any of that kind of thing just now.
 */

MX_EXPORT mx_status_type
mx_set_client_info( MX_RECORD *server_record,
			char *username,
			char *program_name )
{
	static const char fname[] = "mx_set_client_info()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_MESSAGE_BUFFER *aligned_buffer;
	mx_uint32_type *header;
	char *buffer, *message, *ptr;
	mx_uint32_type header_length, message_length;
	mx_uint32_type message_type, status_code;
	mx_status_type mx_status;

#if NETWORK_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	MX_DEBUG( 2,("%s invoked, username = '%s', program_name = '%s'",
			fname, username, program_name));

	if ( server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"server_record argument passed was NULL." );
	}

	server = (MX_NETWORK_SERVER *) server_record->record_class_struct;

	if ( server == (MX_NETWORK_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_SERVER pointer for server record '%s' is NULL.",
			server_record->name );
	}

	/************ Send the 'set client info' message. *************/

	aligned_buffer = server->message_buffer;

	header = &(aligned_buffer->uint32_buffer[0]);
	buffer = &(aligned_buffer->char_buffer[0]);

	header[MX_NETWORK_MAGIC] = htonl( MX_NETWORK_MAGIC_VALUE );
	header[MX_NETWORK_HEADER_LENGTH]
				= htonl( MX_NETWORK_HEADER_LENGTH_VALUE );
	header[MX_NETWORK_MESSAGE_TYPE] = htonl( MX_NETMSG_SET_CLIENT_INFO );
	header[MX_NETWORK_STATUS_CODE] = htonl( MXE_SUCCESS );

	message = buffer + MX_NETWORK_HEADER_LENGTH_VALUE;

	mx_strncpy( message, username, MXU_USERNAME_LENGTH );

	strncat( message, " ", 1 );

	strncat( message, program_name, MXU_PROGRAM_NAME_LENGTH );

	strncat( message, " ", 1 );

	message_length = strlen( message );

	ptr = message + message_length;

	sprintf( ptr, "%lu", mx_process_id() );

	MX_DEBUG( 2,("%s: message = '%s'", fname, message));

	message_length = strlen( message ) + 1;
	header[MX_NETWORK_MESSAGE_LENGTH] = htonl( message_length );

#if NETWORK_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	mx_status = mx_network_send_message( server_record, buffer );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/************** Wait for the response. **************/

	mx_status = mx_network_receive_message( server_record,
				MX_NETWORK_MAXIMUM_MESSAGE_SIZE, buffer );

#if NETWORK_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, " " );
#endif

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
	message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
	message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );
	status_code = ntohl( header[ MX_NETWORK_STATUS_CODE ] );

	if ( message_type != mx_server_response( MX_NETMSG_SET_CLIENT_INFO ) ) {

		if ( message_type == MX_NETMSG_UNEXPECTED_ERROR ) {
			return mx_error( (long)status_code, fname,
				"%s", message );
		}

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Message type for response was not MX_NETMSG_SET_CLIENT_INFO_RESPONSE.  "
"Instead it was of type = %#lx", message_type );
	}

#if 0
	mx_network_display_message_buffer( buffer );
#endif

	/* If the remote command failed, the message field will include
	 * the text of the error message rather than the field type
	 * data we requested.
	 */

	if ( status_code != MXE_SUCCESS ) {
		return mx_error( (long)status_code, fname, message );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_network_get_option( MX_RECORD *server_record,
			unsigned long option_number,
			unsigned long *option_value )
{
	static const char fname[] = "mx_network_get_option()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_MESSAGE_BUFFER *aligned_buffer;
	mx_uint32_type *header, *uint32_message;
	char *buffer, *message;
	mx_uint32_type header_length, message_length;
	mx_uint32_type message_type, status_code;
	mx_uint32_type header_length_in_32bit_words;
	mx_status_type mx_status;

#if NETWORK_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	MX_DEBUG( 2,("%s invoked, option_number = '%#lx'",
			fname, option_number));

	if ( server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"server_record argument passed was NULL." );
	}
	if ( option_value == (unsigned long *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"option_value pointer passed was NULL." );
	}

	server = (MX_NETWORK_SERVER *) server_record->record_class_struct;

	if ( server == (MX_NETWORK_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_SERVER pointer for server record '%s' is NULL.",
			server_record->name );
	}

	/************ Send the 'get option' message. *************/

	aligned_buffer = server->message_buffer;

	header = &(aligned_buffer->uint32_buffer[0]);
	buffer = &(aligned_buffer->char_buffer[0]);

	header[MX_NETWORK_MAGIC] = htonl( MX_NETWORK_MAGIC_VALUE );
	header[MX_NETWORK_HEADER_LENGTH]
				= htonl( MX_NETWORK_HEADER_LENGTH_VALUE );
	header[MX_NETWORK_MESSAGE_TYPE] = htonl( MX_NETMSG_GET_OPTION );
	header[MX_NETWORK_STATUS_CODE] = htonl( MXE_SUCCESS );

	message = buffer + MX_NETWORK_HEADER_LENGTH_VALUE;
	uint32_message = header + MX_NETWORK_NUM_HEADER_VALUES;

	uint32_message[0] = htonl( option_number );

	header[MX_NETWORK_MESSAGE_LENGTH] = htonl( sizeof(mx_uint32_type) );

#if NETWORK_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	mx_status = mx_network_send_message( server_record, buffer );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/************** Wait for the response. **************/

	mx_status = mx_network_receive_message( server_record,
				MX_NETWORK_MAXIMUM_MESSAGE_SIZE, buffer );

#if NETWORK_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, " " );
#endif

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
	message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
	message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );
	status_code = ntohl( header[ MX_NETWORK_STATUS_CODE ] );

	if ( message_type != mx_server_response( MX_NETMSG_GET_OPTION ) ) {

		if ( message_type == MX_NETMSG_UNEXPECTED_ERROR ) {

			if ( status_code == MXE_NOT_YET_IMPLEMENTED ) {

				/* The remote server is an old server that
				 * does not implement the 'get option'
				 * message.
				 */

				return mx_error_quiet((long) status_code, fname,
	"The MX 'get option' message type is not implemented by server '%s'.",
					server_record->name );
			} else {
				/* Some other error occurred. */

				return mx_error( MXE_NETWORK_IO_ERROR, fname,
	"Unexpected server error for MX server '%s'.  "
	"Server error was '%s -> %s'", server_record->name,
				mx_status_code_string((long) status_code ),
					message );
			}
		}

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Message type for response was not MX_NETMSG_GET_OPTION_RESPONSE.  "
"Instead it was of type = %#lx", message_type );
	}

#if 0
	mx_network_display_message_buffer( buffer );
#endif

	/* If the remote command failed, the message field will include
	 * the text of the error message rather than the field type
	 * data we requested.
	 */

	if ( status_code != MXE_SUCCESS ) {
		return mx_error( (long)status_code, fname, message );
	}

	if ( message_length < sizeof(mx_uint32_type) ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Incomplete message received.  Message length = %ld",
			message_length );
	}

	header_length_in_32bit_words = header_length / sizeof(mx_uint32_type);

	uint32_message = header + header_length_in_32bit_words;

	*option_value = ntohl( uint32_message[0] );

	MX_DEBUG( 2,("%s invoked, *option_value = '%#lx'",
			fname, *option_value));

	return MX_SUCCESSFUL_RESULT;
}

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_network_set_option( MX_RECORD *server_record,
			unsigned long option_number,
			unsigned long option_value )
{
	static const char fname[] = "mx_network_set_option()";

	MX_NETWORK_SERVER *server;
	MX_NETWORK_MESSAGE_BUFFER *aligned_buffer;
	mx_uint32_type *header, *uint32_message;
	char *buffer, *message;
	mx_uint32_type header_length, message_length;
	mx_uint32_type message_type, status_code;
	mx_status_type mx_status;

#if NETWORK_DEBUG_TIMING
	MX_HRT_TIMING measurement;
#endif

	MX_DEBUG( 2,
		("%s invoked, option_number = '%#lx', option_value = '%#lx'",
			fname, option_number, option_value));

	if ( server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"server_record argument passed was NULL." );
	}

	server = (MX_NETWORK_SERVER *) server_record->record_class_struct;

	if ( server == (MX_NETWORK_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_SERVER pointer for server record '%s' is NULL.",
			server_record->name );
	}

	/************ Send the 'set option' message. *************/

	aligned_buffer = server->message_buffer;

	header = &(aligned_buffer->uint32_buffer[0]);
	buffer = &(aligned_buffer->char_buffer[0]);

	header[MX_NETWORK_MAGIC] = htonl( MX_NETWORK_MAGIC_VALUE );
	header[MX_NETWORK_HEADER_LENGTH]
				= htonl( MX_NETWORK_HEADER_LENGTH_VALUE );
	header[MX_NETWORK_MESSAGE_TYPE] = htonl( MX_NETMSG_SET_OPTION );
	header[MX_NETWORK_STATUS_CODE] = htonl( MXE_SUCCESS );

	message = buffer + MX_NETWORK_HEADER_LENGTH_VALUE;
	uint32_message = header + MX_NETWORK_NUM_HEADER_VALUES;

	uint32_message[0] = htonl( option_number );
	uint32_message[1] = htonl( option_value );

	header[MX_NETWORK_MESSAGE_LENGTH] = htonl( 2 * sizeof(mx_uint32_type) );

#if NETWORK_DEBUG_TIMING
	MX_HRT_START( measurement );
#endif

	mx_status = mx_network_send_message( server_record, buffer );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/************** Wait for the response. **************/

	mx_status = mx_network_receive_message( server_record,
				MX_NETWORK_MAXIMUM_MESSAGE_SIZE, buffer );

#if NETWORK_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname, " " );
#endif

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
	message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
	message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );
	status_code = ntohl( header[ MX_NETWORK_STATUS_CODE ] );

	if ( message_type != mx_server_response( MX_NETMSG_SET_OPTION ) ) {

		if ( message_type == MX_NETMSG_UNEXPECTED_ERROR ) {

			if ( status_code == MXE_NOT_YET_IMPLEMENTED ) {

				/* The remote server is an old server that
				 * does not implement the 'set option'
				 * message.
				 */

				return mx_error_quiet((long) status_code, fname,
	"The MX 'set option' message type is not implemented by server '%s'.",
					server_record->name );
			} else {
				/* Some other error occurred. */

				return mx_error( MXE_NETWORK_IO_ERROR, fname,
	"Unexpected server error for MX server '%s'.  "
	"Server error was '%s -> %s'", server_record->name,
				mx_status_code_string((long) status_code ),
					message );
			}
		}

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Message type for response was not MX_NETMSG_SET_OPTION_RESPONSE.  "
"Instead it was of type = %#lx", message_type );
	}

#if 0
	mx_network_display_message_buffer( buffer );
#endif

	/* If the remote command failed, the message field will include
	 * the text of the error message rather than the field type
	 * data we requested.
	 */

	if ( status_code != MXE_SUCCESS ) {
		return mx_error( (long)status_code, fname, message );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_network_request_data_format( MX_RECORD *server_record,
				unsigned long requested_format )
{
	static const char fname[] = "mx_network_request_data_format()";

	MX_NETWORK_SERVER *server;
	unsigned long local_native_data_format, remote_native_data_format;
	mx_status_type mx_status;

	if ( server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"server_record argument passed was NULL." );
	}

	server = (MX_NETWORK_SERVER *) server_record->record_class_struct;

	if ( server == (MX_NETWORK_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_SERVER pointer for server record '%s' is NULL.",
			server_record->name );
	}

	/* If the caller has not requested automatic data format negotiation,
	 * just attempt to set the requested format and fail if it is not
	 * available.
	 */

	if ( requested_format != MX_NETWORK_DATAFMT_NEGOTIATE ) {

		mx_status = mx_network_set_option( server_record,
			MX_NETWORK_OPTION_DATAFMT, requested_format );

		if ( mx_status.code == MXE_SUCCESS ) {
			server->data_format = requested_format;

			return MX_SUCCESSFUL_RESULT;
		} else
		if ( ( mx_status.code == MXE_NOT_YET_IMPLEMENTED )
		  && ( requested_format != MX_NETWORK_DATAFMT_ASCII ) )
		{
			/* The server does not support anything other
			 * than ASCII format.
			 */

			server->data_format = MX_NETWORK_DATAFMT_ASCII;

			return mx_error( mx_status.code, fname,
			"MX server '%s' only supports ASCII data format.",
				server_record->name );
		} else {
			return mx_status;
		}
	}

	/* We have been requested to perform automatic data format negotiation.
	 * To do this, we must first get the local native data format.
	 */

	local_native_data_format = mx_native_data_format();

	/* Find out the native data format of the remote server. */

	mx_status = mx_network_get_option( server_record,
			MX_NETWORK_OPTION_NATIVE_DATAFMT,
			&remote_native_data_format );

	switch( mx_status.code ) {
	case MXE_SUCCESS:
		break;
	case MXE_NOT_YET_IMPLEMENTED:

		/* The server is not new enough to have 'get option' and
		 * 'set option' implemented in it, so we have to settle
		 * for ASCII data format.
		 */

		MX_DEBUG( 2,
			("%s: Server '%s' only supports ASCII network format.",
	 		fname, server_record->name ));

		server->data_format = MX_NETWORK_DATAFMT_ASCII;
		
		/* Cannot do any further negotiation, so we just return now. */

		return MX_SUCCESSFUL_RESULT;
		break;
	default:
		return mx_status;
		break;
	}

	MX_DEBUG( 2,
    ("%s: local_native_data_format = %#lx, remote_native_data_format = %#lx",
     		fname, local_native_data_format, remote_native_data_format));

	/* Try to select a binary data format. */

	if ( local_native_data_format == remote_native_data_format ) {

		/* The local and remote data formats are the same, so
		 * we should try raw format.
		 */

		requested_format = MX_NETWORK_DATAFMT_RAW;
	} else {
		/* Otherwise, we should use XDR format. */

		requested_format = MX_NETWORK_DATAFMT_XDR;
	}

	mx_status = mx_network_set_option( server_record,
			MX_NETWORK_OPTION_DATAFMT, requested_format );

	switch ( mx_status.code ) {
	case MXE_SUCCESS:
		/* If the 'set option' command succeeded, we are done. */

		MX_DEBUG( 2,("%s: binary data format successfully selected.",
			fname));

		server->data_format = requested_format;

		return MX_SUCCESSFUL_RESULT;
		break;
	case MXE_NOT_YET_IMPLEMENTED:
		/* The server does not implement 'set option'. */

		MX_DEBUG( 2,("%s: Server '%s' does not support 'set option'.",
			fname, server_record->name));

		server->data_format = MX_NETWORK_DATAFMT_ASCII;

		return MX_SUCCESSFUL_RESULT;
		break;
	}

	/* Selecting a binary data format failed, so try ASCII format. */

	mx_status = mx_network_set_option( server_record,
			MX_NETWORK_OPTION_DATAFMT, MX_NETWORK_DATAFMT_ASCII );

	/* Even if selecting ASCII format failed, we have no alternative
	 * to assuming ASCII format is in use.  So we make this assumption
	 * and hope for the best.
	 */

	server->data_format = MX_NETWORK_DATAFMT_ASCII;

	if ( mx_status.code == MXE_SUCCESS ) {
		MX_DEBUG( 2,("%s: ASCII data format successfully selected.",
				fname));
	} else {
		MX_DEBUG( 2,("%s: Assuming ASCII data format is in use.",
				fname));
	}

	return MX_SUCCESSFUL_RESULT;
}

