/*
 * Name:    mx_socket.h 
 *
 * Purpose: BSD sockets and Winsock interface functions.
 *
 * Author:  William Lavender
 *
 * This header file includes all the defines and prototypes normally needed
 * by a socket-based program.  Thus, to use sockets, it is only necessary to
 * use #include "mx_socket.h" rather than including all the individual headers
 * separately.  This also promotes portability since different operating
 * systems have different sets of header files that they want to be included
 * in order to implement socket functionality.
 *
 *----------------------------------------------------------------------
 *
 * Copyright 1999-2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#ifndef __MX_SOCKET_H__
#define __MX_SOCKET_H__

#include "mx_osdef.h"

/* Include standard socket defines and prototypes. */

#if defined( OS_WIN32 )
#  include <winsock.h>

#elif defined( OS_VMS ) && defined( HAVE_WINTCP )

#  include "vms_twg.h"		/* For WIN/TCP from The Wollogong Group. */

#else

#  if defined( OS_VXWORKS )
#     include <sockLib.h>
#     include <ioLib.h>
#  else
#     include <sys/time.h>
#  endif

#  include <time.h>
#  include <netdb.h>
#  include <sys/types.h>
#  include <sys/socket.h>
#  include <sys/ioctl.h>
#  include <netinet/in.h>
#  include <netinet/tcp.h>
#  include <arpa/inet.h>

#  if defined( OS_HPUX )
      extern int h_errno;
#  endif

#  if defined( OS_DJGPP )
      extern int sock_init(void);
      extern int h_errno;
#  endif

#endif

#if HAVE_UNIX_DOMAIN_SOCKETS
#  include <sys/un.h>
#endif

/* Define primary socket data structures. */

/* ..._FD means 'file descriptor' in case you were wondering. */

/* ===== Unix BSD sockets support ===== */

#if defined( OS_UNIX ) || defined( OS_CYGWIN ) || defined( OS_VMS ) \
 || defined( OS_DJGPP ) || defined( OS_RTEMS ) || defined( OS_VXWORKS )

#  define MX_SOCKET_FD			int
#  define MX_INVALID_SOCKET_FD		(-1)
#  define MX_SOCKET_ERROR		(-1)

#  define mx_socket_get_last_error()	(errno)

/* ===== Microsoft Win32 Winsock support ===== */

#elif defined( OS_WIN32 )

#  define MX_SOCKET_FD			SOCKET
#  define MX_INVALID_SOCKET_FD		INVALID_SOCKET
#  define MX_SOCKET_ERROR		SOCKET_ERROR

#  define mx_socket_get_last_error()	WSAGetLastError()

/* ============================ */

#else
#  error "Socket definitions not configured for this operating system."
#endif

typedef struct {
	MX_SOCKET_FD socket_fd;
	unsigned long socket_flags;
} MX_SOCKET;

/* ***** Various define's ***** */

/* Redefine some error codes for Win32. */

#if defined( OS_WIN32 )

#  define ECONNRESET	WSAECONNRESET
#  define ENOTCONN	WSAENOTCONN
#  define EWOULDBLOCK	WSAEWOULDBLOCK

#endif

#define MX_SOCKET_DEFAULT_BUFFER_SIZE		0

/* The following are flags for the various mx_..._socket_open_as_...()
 * functions below.
 */

#define MXF_SOCKET_DISABLE_NAGLE_ALGORITHM	0x1
#define MXF_SOCKET_QUIET			0x2

/* The following defines are for the second argument to the function
 * mx_socket_check_error_status().
 */

#define MXF_SOCKCHK_INVALID	(-1)
#define MXF_SOCKCHK_ZERO	0
#define MXF_SOCKCHK_NULL	1

/* ***** General function definitions ***** */

MX_API mx_status_type mx_tcp_socket_open_as_client( MX_SOCKET **client_socket,
						char *hostname,
						int port_number,
						unsigned long socket_flags,
						size_t buffer_size );

MX_API mx_status_type mx_tcp_socket_open_as_server( MX_SOCKET **server_socket,
						int port_number,
						unsigned long socket_flags,
						size_t buffer_size );

MX_API mx_status_type mx_unix_socket_open_as_client( MX_SOCKET **client_socket,
						char *pathname,
						unsigned long socket_flags,
						size_t buffer_size );

MX_API mx_status_type mx_unix_socket_open_as_server( MX_SOCKET **server_socket,
						char *pathname,
						unsigned long socket_flags,
						size_t buffer_size );

MX_API mx_status_type mx_socket_close( MX_SOCKET *mx_socket );

MX_API int  mx_socket_is_open( MX_SOCKET *mx_socket );

MX_API void mx_socket_mark_as_closed( MX_SOCKET *mx_socket );

MX_API char *mx_socket_strerror( int saved_errno );

MX_API int  mx_socket_check_error_status( void *value_to_check,
						int type_of_check,
						char **error_string );

MX_API mx_status_type mx_socket_get_inet_address( char *hostname,
						unsigned long *inet_address );

MX_API mx_status_type mx_socket_send( MX_SOCKET *mx_socket,
				void *message_buffer,
				size_t message_length_in_bytes );

MX_API mx_status_type mx_socket_receive( MX_SOCKET *mx_socket,
				void *message_buffer,
				size_t message_length_in_bytes,
				void *input_terminators,
				size_t input_terminators_length_in_bytes );

MX_API mx_status_type mx_socket_putline( MX_SOCKET *mx_socket,
					char *buffer,
					char *line_terminators );

MX_API mx_status_type mx_socket_getline( MX_SOCKET *mx_socket,
					char *buffer,
					size_t buffer_length,
					char *line_terminators );

#if HAVE_FIONREAD_FOR_SOCKETS
#define mx_socket_num_input_bytes_available(s,n) \
		mx_socket_fionread_num_input_bytes_available(s,n)
#else
#define mx_socket_num_input_bytes_available(s,n) \
		mx_socket_select_num_input_bytes_available(s,n)
#endif

MX_API mx_status_type mx_socket_fionread_num_input_bytes_available(
					MX_SOCKET *mx_socket,
					int *num_input_bytes_available );

MX_API mx_status_type mx_socket_select_num_input_bytes_available(
					MX_SOCKET *mx_socket,
					int *num_input_bytes_available );

MX_API mx_status_type mx_socket_discard_unread_input( MX_SOCKET *mx_socket );

#endif /* __MX_SOCKET_H__ */

