/*
 * Name:    mx_usb.c
 *
 * Purpose: MX function library for controlling a USB interface.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_types.h"
#include "mx_usb.h"
#include "mx_driver.h"

MX_EXPORT mx_status_type
mx_usb_get_pointers( MX_USB *usb,
			MX_USB_FUNCTION_LIST **function_list_ptr,
			const char *calling_fname )
{
	static const char fname[] = "mx_usb_get_pointers()";

	MX_RECORD *usb_record;

	if ( usb == (MX_USB *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The usb pointer passed by '%s' was NULL.",
			calling_fname );
	}

	usb_record = usb->record;

	if ( usb_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The record pointer for the MX_USB structure passed is NULL." );
	}

	if ( usb_record->mx_class != MXI_USB ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The record '%s' passed by '%s' is not a USB interface.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			usb_record->name, calling_fname,
			usb_record->mx_superclass,
			usb_record->mx_class,
			usb_record->mx_type );
	}

	if ( function_list_ptr == (MX_USB_FUNCTION_LIST **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The usb pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*function_list_ptr = (MX_USB_FUNCTION_LIST *)
				usb_record->class_specific_function_list;

	if ( *function_list_ptr == (MX_USB_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_USB_FUNCTION_LIST pointer for record '%s' passed by '%s' is NULL.",
			usb_record->name, calling_fname );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_usb_enumerate_devices( MX_USB *usb )
{
	static const char fname[] = "mx_usb_enumerate_devices()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB * );
	mx_status_type mx_status;

	if ( usb == (MX_USB *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_USB pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fptr = fl_ptr->enumerate_devices;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'enumerate_devices' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb->record ),
			usb->record->name );
	}

	mx_status = (*fptr)( usb );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_delete_device( MX_USB_DEVICE *usb_device )
{
	static const char fname[] = "mx_usb_delete_device()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB_DEVICE * );
	mx_status_type mx_status;

	if ( usb_device == (MX_USB_DEVICE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_USB_DEVICE pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb_device->usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fptr = fl_ptr->delete_device;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'delete_device' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb_device->usb->record ),
			usb_device->usb->record->name );
	}

	mx_status = (*fptr)( usb_device );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_reset_device( MX_USB_DEVICE *usb_device )
{
	static const char fname[] = "mx_usb_reset()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB_DEVICE * );
	mx_status_type mx_status;

	if ( usb_device == (MX_USB_DEVICE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_USB_DEVICE pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb_device->usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fptr = fl_ptr->reset_device;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'reset' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb_device->usb->record ),
			usb_device->usb->record->name );
	}

	mx_status = (*fptr)( usb_device );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_find_device_by_order( MX_USB *usb,
				MX_USB_DEVICE **usb_device,
				unsigned long vendor_id,
				unsigned long product_id,
				int order_number )
{
	static const char fname[] = "mx_usb_find_device_by_order()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB *, MX_USB_DEVICE ** );
	mx_status_type mx_status;

	if ( usb == (MX_USB *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_USB pointer passed was NULL." );
	}
	if ( usb_device == (MX_USB_DEVICE **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_USB_DEVICE pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fptr = fl_ptr->find_device;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'find_device' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb->record ),
			usb->record->name );
	}

	usb->find_type = MXF_USB_FIND_TYPE_ORDER;

	usb->find.by_order.vendor_id = vendor_id;
	usb->find.by_order.product_id = product_id;
	usb->find.by_order.order_number = order_number;

	mx_status = (*fptr)( usb, usb_device );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_find_device_by_serial_number( MX_USB *usb,
				MX_USB_DEVICE **usb_device,
				unsigned long vendor_id,
				unsigned long product_id,
				char *serial_number )
{
	static const char fname[] = "mx_usb_find_device_by_serial_number()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB *, MX_USB_DEVICE ** );
	mx_status_type mx_status;

	if ( usb == (MX_USB *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_USB pointer passed was NULL." );
	}
	if ( usb_device == (MX_USB_DEVICE **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_USB_DEVICE pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fptr = fl_ptr->find_device;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'find_device' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb->record ),
			usb->record->name );
	}

	usb->find_type = MXF_USB_FIND_TYPE_SERIAL_NUMBER;

	usb->find.by_serial_number.vendor_id = vendor_id;
	usb->find.by_serial_number.product_id = product_id;
	usb->find.by_serial_number.serial_number = serial_number;

	mx_status = (*fptr)( usb, usb_device );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_find_configurations_interfaces_and_endpoints( MX_USB_DEVICE *usb_device )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_usb_get_mx_configuration( MX_USB_DEVICE *usb_device,
			MX_USB_CONFIGURATION **usb_configuration,
			int configuration_number )
{
	static const char fname[] = "mx_usb_get_mx_configuration()";

	MX_USB_CONFIGURATION *test_configuration;
	int i;

	if ( usb_device == (MX_USB_DEVICE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_DEVICE pointer passed was NULL." );
	}
	if ( usb_configuration == (MX_USB_CONFIGURATION **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_CONFIGURATION pointer passed was NULL." );
	}

	for ( i = 0; i < usb_device->num_configurations; i++ ) {
		test_configuration = &(usb_device->configuration_array[i]);

		if ( configuration_number ==
				test_configuration->configuration_number )
		{
			*usb_configuration = test_configuration;

			return MX_SUCCESSFUL_RESULT;
		}
	}
	return mx_error( MXE_NOT_FOUND, fname,
		"Configuration %d was not found for the specified USB device "
		"(vendor %#lx, product %#lx, order %d).",
			configuration_number, usb_device->vendor_id,
			usb_device->product_id, usb_device->order_number );
}

MX_EXPORT mx_status_type
mx_usb_get_mx_interface( MX_USB_DEVICE *usb_device,
			MX_USB_INTERFACE **usb_interface,
			int configuration_number,
			int interface_number )
{
	static const char fname[] = "mx_usb_get_mx_interface()";

	MX_USB_CONFIGURATION *usb_configuration;
	MX_USB_INTERFACE *test_interface;
	int i;
	mx_status_type mx_status;

	if ( usb_device == (MX_USB_DEVICE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_DEVICE pointer passed was NULL." );
	}
	if ( usb_interface == (MX_USB_INTERFACE **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_INTERFACE pointer passed was NULL." );
	}

	mx_status = mx_usb_get_mx_configuration( usb_device,
					&usb_configuration,
					configuration_number );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	for ( i = 0; i < usb_configuration->num_interfaces; i++ ) {
		test_interface = &(usb_configuration->interface_array[i]);

		if ( interface_number == test_interface->interface_number ) {
			*usb_interface = test_interface;

			return MX_SUCCESSFUL_RESULT;
		}
	}
	return mx_error( MXE_NOT_FOUND, fname,
	"Configuration %d, interface %d was not found for the specified "
	"USB device (vendor %#lx, product %#lx, order %d).",
			configuration_number, interface_number,
			usb_device->vendor_id, usb_device->product_id,
			usb_device->order_number );
}

MX_EXPORT mx_status_type
mx_usb_get_mx_endpoint( MX_USB_DEVICE *usb_device,
			MX_USB_ENDPOINT **usb_endpoint,
			int configuration_number,
			int interface_number,
			int endpoint_number )
{
	static const char fname[] = "mx_usb_get_mx_endpoint()";

	MX_USB_INTERFACE *usb_interface;
	MX_USB_ENDPOINT *test_endpoint;
	int i;
	mx_status_type mx_status;

	if ( usb_device == (MX_USB_DEVICE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_DEVICE pointer passed was NULL." );
	}
	if ( usb_endpoint == (MX_USB_ENDPOINT **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_ENDPOINT pointer passed was NULL." );
	}

	mx_status = mx_usb_get_mx_interface( usb_device,
					&usb_interface,
					configuration_number,
					interface_number );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	for ( i = 0; i < usb_interface->num_endpoints; i++ ) {
		test_endpoint = &(usb_interface->endpoint_array[i]);

		if ( endpoint_number == test_endpoint->endpoint_number ) {
			*usb_endpoint = test_endpoint;

			return MX_SUCCESSFUL_RESULT;
		}
	}
	return mx_error( MXE_NOT_FOUND, fname,
	"Configuration %d, interface %d, endpoint %d was not found "
	"for the specified USB device (vendor %#lx, product %#lx, order %d).",
			configuration_number, interface_number, endpoint_number,
			usb_device->vendor_id, usb_device->product_id,
			usb_device->order_number );
}

MX_EXPORT mx_status_type
mx_usb_claim_interface( MX_USB_INTERFACE *usb_interface )
{
	static const char fname[] = "mx_usb_claim_interface()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB_INTERFACE * );
	mx_status_type mx_status;

	if ( usb_interface == (MX_USB_INTERFACE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_INTERFACE pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb_interface->usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fptr = fl_ptr->claim_interface;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'claim_interface' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb_interface->usb->record ),
			usb_interface->usb->record->name );
	}

	mx_status = (*fptr)( usb_interface );

	if ( mx_status.code == MXE_SUCCESS ) {
		usb_interface->interface_is_claimed = TRUE;
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_release_interface( MX_USB_INTERFACE *usb_interface )
{
	static const char fname[] = "mx_usb_release_interface()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB_INTERFACE * );
	mx_status_type mx_status;

	if ( usb_interface == (MX_USB_INTERFACE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_INTERFACE pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb_interface->usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fptr = fl_ptr->release_interface;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'release_interface' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb_interface->usb->record ),
			usb_interface->usb->record->name );
	}

	mx_status = (*fptr)( usb_interface );

	if ( mx_status.code == MXE_SUCCESS ) {
		usb_interface->interface_is_claimed = FALSE;
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_get_configuration( MX_USB_DEVICE *usb_device,
			int *configuration_number )
{
	static const char fname[] = "mx_usb_get_configuration()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB_DEVICE * );
	mx_status_type mx_status;

	if ( usb_device == (MX_USB_DEVICE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_DEVICE pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb_device->usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fptr = fl_ptr->get_configuration;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'get_configuration' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb_device->usb->record ),
			usb_device->usb->record->name );
	}

	mx_status = (*fptr)( usb_device );

	if ( configuration_number != NULL ) {
		*configuration_number = usb_device->configuration_number;
	}

	if ( mx_status.code == MXE_SUCCESS ) {
		usb_device->current_configuration_number
			= usb_device->configuration_number;
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_set_configuration( MX_USB_DEVICE *usb_device,
			int configuration_number )
{
	static const char fname[] = "mx_usb_set_configuration()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB_DEVICE * );
	mx_status_type mx_status;

	if ( usb_device == (MX_USB_DEVICE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_DEVICE pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb_device->usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fptr = fl_ptr->set_configuration;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'set_configuration' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb_device->usb->record ),
			usb_device->usb->record->name );
	}

	usb_device->configuration_number = configuration_number;

	mx_status = (*fptr)( usb_device );

	if ( mx_status.code == MXE_SUCCESS ) {
		usb_device->current_configuration_number
			= usb_device->configuration_number;
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_get_alternate_interface( MX_USB_INTERFACE *usb_interface,
			int *interface_number )
{
	static const char fname[] = "mx_usb_get_alternate_interface()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB_INTERFACE * );
	mx_status_type mx_status;

	if ( usb_interface == (MX_USB_INTERFACE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_INTERFACE pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb_interface->usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fptr = fl_ptr->get_alternate_interface;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'get_alternate_interface' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb_interface->usb->record ),
			usb_interface->usb->record->name );
	}

	mx_status = (*fptr)( usb_interface );

	if ( interface_number != NULL ) {
		*interface_number = usb_interface->interface_number;
	}

	if ( mx_status.code == MXE_SUCCESS ) {
		usb_interface->current_interface_number
			= usb_interface->interface_number;
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_set_alternate_interface( MX_USB_INTERFACE *usb_interface,
			int interface_number )
{
	static const char fname[] = "mx_usb_set_alternate_interface()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB_INTERFACE * );
	mx_status_type mx_status;

	if ( usb_interface == (MX_USB_INTERFACE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_INTERFACE pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb_interface->usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	fptr = fl_ptr->set_alternate_interface;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'set_alternate_interface' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb_interface->usb->record ),
			usb_interface->usb->record->name );
	}

	usb_interface->interface_number = interface_number;

	mx_status = (*fptr)( usb_interface );

	if ( mx_status.code == MXE_SUCCESS ) {
		usb_interface->current_interface_number
			= usb_interface->interface_number;
	}

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_control_transfer( MX_USB_DEVICE *usb_device,
			int request_type,
			int request,
			int value,
			int index_or_offset,
			int transfer_data_length,
			char *transfer_data )
{
	static const char fname[] = "mx_usb_control_transfer()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB_DEVICE * );
	mx_status_type mx_status;

	if ( usb_device == (MX_USB_DEVICE *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_DEVICE pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb_device->usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( transfer_data == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The transfer_data pointer passed was NULL." );
	}

	fptr = fl_ptr->control_transfer;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'control_transfer' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb_device->usb->record ),
			usb_device->usb->record->name );
	}

	usb_device->request_type = request_type;
	usb_device->request = request;
	usb_device->value = value;
	usb_device->index_or_offset = index_or_offset;
	usb_device->transfer_data_length = transfer_data_length;
	usb_device->transfer_data = transfer_data;

	mx_status = (*fptr)( usb_device );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_bulk_read( MX_USB_ENDPOINT *usb_endpoint,
			int transfer_data_length,
			char *transfer_data )
{
	static const char fname[] = "mx_usb_bulk_read()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB_ENDPOINT * );
	mx_status_type mx_status;

	if ( usb_endpoint == (MX_USB_ENDPOINT *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_ENDPOINT pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb_endpoint->usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( transfer_data == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The transfer_data pointer passed was NULL." );
	}

	fptr = fl_ptr->bulk_read;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'bulk_read' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb_endpoint->usb->record ),
			usb_endpoint->usb->record->name );
	}

	usb_endpoint->transfer_data_length = transfer_data_length;

	usb_endpoint->transfer_data = transfer_data;

	mx_status = (*fptr)( usb_endpoint );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_usb_bulk_write( MX_USB_ENDPOINT *usb_endpoint,
			int transfer_data_length,
			char *transfer_data )
{
	static const char fname[] = "mx_usb_bulk_write()";

	MX_USB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_USB_ENDPOINT * );
	mx_status_type mx_status;

	if ( usb_endpoint == (MX_USB_ENDPOINT *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_USB_ENDPOINT pointer passed was NULL." );
	}

	mx_status = mx_usb_get_pointers( usb_endpoint->usb, &fl_ptr, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( transfer_data == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The transfer_data pointer passed was NULL." );
	}

	fptr = fl_ptr->bulk_write;

	if ( fptr == NULL ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"The 'bulk_write' function for the '%s' driver "
		"used by USB record '%s' is not yet implemented.",
			mx_get_driver_name( usb_endpoint->usb->record ),
			usb_endpoint->usb->record->name );
	}

	usb_endpoint->transfer_data_length = transfer_data_length;

	usb_endpoint->transfer_data = transfer_data;

	mx_status = (*fptr)( usb_endpoint );

	return mx_status;
}

