/*
 * Name:    mx_usb.h
 *
 * Purpose: Header file for USB interfaces.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#ifndef __MX_USB_H__
#define __MX_USB_H__

#define MXF_USB_FIND_TYPE_ORDER		1
#define MXF_USB_FIND_TYPE_SERIAL_NUMBER	2
#define MXF_USB_FIND_TYPE_BUS_ADDRESS	3

union mx_usb_find_device_union {
	struct {
		unsigned long vendor_id;
		unsigned long product_id;
		int order_number;
	} by_order;

	struct {
		unsigned long vendor_id;
		unsigned long product_id;
		char *serial_number;
	} by_serial_number;
};

typedef struct {
	MX_RECORD *record;

	unsigned long usb_flags;

	double timeout;		/* in seconds */

	int find_type;
	union mx_usb_find_device_union find;

	void *bus_data;
} MX_USB;

typedef struct {
	MX_USB	*usb;
	unsigned long vendor_id;
	unsigned long product_id;
	int order_number;

	int current_configuration_number;
	int configuration_number;

	int request_type;
	int request;
	int value;
	int index_or_offset;
	int transfer_data_length;
	char *transfer_data;

	int num_configurations;
	struct mx_usb_configuration *configuration_array;

	void *device_data;
} MX_USB_DEVICE;

typedef struct mx_usb_configuration {
	MX_USB_DEVICE *usb_device;
	unsigned long configuration_number;

	int num_interfaces;
	struct mx_usb_interface *interface_array;

	void *configuration_data;
} MX_USB_CONFIGURATION;

typedef struct mx_usb_interface {
	MX_USB *usb;
	MX_USB_DEVICE *usb_device;
	MX_USB_CONFIGURATION *usb_configuration;

	int current_interface_number;
	int interface_number;

	int alternate_setting;

	int interface_is_claimed;

	int num_endpoints;
	struct mx_usb_endpoint *endpoint_array;

	void *interface_data;
} MX_USB_INTERFACE;

typedef struct mx_usb_endpoint {
	MX_USB *usb;
	MX_USB_DEVICE *usb_device;
	MX_USB_CONFIGURATION *usb_configuration;
	MX_USB_INTERFACE *usb_interface;

	unsigned long endpoint_number;

	int transfer_data_length;
	char *transfer_data;

	void *endpoint_data;
} MX_USB_ENDPOINT;
	
#define MX_USB_STANDARD_FIELDS \
  {-1, -1, "usb_flags", MXFT_HEX, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_USB, usb_flags), \
	{0}, NULL, (MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY)}

typedef struct {
	mx_status_type ( *enumerate_devices )( MX_USB *usb );
	mx_status_type ( *find_device )( MX_USB *usb,
					MX_USB_DEVICE **usb_device );
	mx_status_type ( *delete_device )( MX_USB_DEVICE *usb_device );
	mx_status_type ( *reset_device )( MX_USB_DEVICE *usb_device );
	mx_status_type ( *claim_interface )( MX_USB_INTERFACE *usb_interface );
	mx_status_type ( *release_interface )( MX_USB_INTERFACE *usb_interface);
	mx_status_type ( *set_configuration )( MX_USB_DEVICE *usb_device );
	mx_status_type ( *get_configuration )( MX_USB_DEVICE *usb_device );
	mx_status_type ( *set_alternate_interface )(
					MX_USB_INTERFACE *usb_interface );
	mx_status_type ( *get_alternate_interface )(
					MX_USB_INTERFACE *usb_interface );
	mx_status_type ( *control_transfer )( MX_USB_DEVICE *usb_device );
	mx_status_type ( *bulk_read )( MX_USB_ENDPOINT *usb_endpoint );
	mx_status_type ( *bulk_write )( MX_USB_ENDPOINT *usb_endpoint );
} MX_USB_FUNCTION_LIST;

/* ============== Interface function prototypes. ============== */

MX_API mx_status_type mx_usb_get_pointers( MX_USB *usb,
					MX_USB_FUNCTION_LIST **fptr,
					const char *calling_fname );

MX_API mx_status_type mx_usb_enumerate_devices( MX_USB *usb );

MX_API mx_status_type mx_usb_delete_device( MX_USB_DEVICE *usb_device );

MX_API mx_status_type mx_usb_reset_device( MX_USB_DEVICE *usb_device );

/*---*/

MX_API mx_status_type mx_usb_find_device_by_order( MX_USB *usb,
					MX_USB_DEVICE **usb_device,
					unsigned long vendor_id,
					unsigned long product_id,
					int order_number );

MX_API mx_status_type mx_usb_find_device_by_serial_number( MX_USB *usb,
					MX_USB_DEVICE **usb_device,
					unsigned long vendor_id,
					unsigned long product_id,
					char *serial_number );

MX_API mx_status_type mx_usb_find_device_by_bus_address( MX_USB *usb,
					MX_USB_DEVICE **usb_device,
					int bus_number,
					int device_number );

/*---*/

MX_API mx_status_type mx_usb_find_configurations_interfaces_and_endpoints(
					MX_USB_DEVICE *usb_device );

/*---*/

MX_API mx_status_type mx_usb_get_mx_configuration( MX_USB_DEVICE *usb_device,
				       MX_USB_CONFIGURATION **usb_configuration,
				       int configuration_number );

MX_API mx_status_type mx_usb_get_mx_interface( MX_USB_DEVICE *usb_device,
					MX_USB_INTERFACE **usb_interface,
					int configuration_number,
					int interface_number );

MX_API mx_status_type mx_usb_get_mx_endpoint( MX_USB_DEVICE *usb_device,
					MX_USB_ENDPOINT **usb_endpoint,
					int configuration_number,
					int interface_number,
					int endpoint_number );

/*---*/

MX_API mx_status_type mx_usb_claim_interface( MX_USB_INTERFACE *usb_interface );

MX_API mx_status_type mx_usb_release_interface( MX_USB_INTERFACE
							*usb_interface );

MX_API mx_status_type mx_usb_get_configuration( MX_USB_DEVICE *usb_device,
						int *configuration_number );

MX_API mx_status_type mx_usb_set_configuration( MX_USB_DEVICE *usb_device,
						int configuration_number );

MX_API mx_status_type mx_usb_get_alternate_interface(
						MX_USB_INTERFACE *usb_interface,
						int *interface_number );

MX_API mx_status_type mx_usb_set_alternate_interface(
						MX_USB_INTERFACE *usb_interface,
						int interface_number );

/*---*/

MX_API mx_status_type mx_usb_control_transfer( MX_USB_DEVICE *usb_device,
					int request_type,
					int request,
					int value,
					int index_or_offset,
					int transfer_data_length,
					char *transfer_data );

MX_API mx_status_type mx_usb_bulk_read( MX_USB_ENDPOINT *usb_endpoint,
					int transfer_data_length,
					char *transfer_data );

MX_API mx_status_type mx_usb_bulk_write( MX_USB_ENDPOINT *usb_endpoint,
					int transfer_data_length,
					char *transfer_data );

#endif /* __MX_USB_H__ */
