/*
 * Name:    ms_mxserver.c
 *
 * Purpose: This file implements the server side of the MX network protocol
 *          for the generic MX server.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/* NETWORK_DEBUG displays the messages transmitted to and from the server
 * while NETWORK_DEBUG_VERBOSE shows much more information.
 */

#define NETWORK_DEBUG			FALSE

#define NETWORK_DEBUG_VERBOSE		FALSE

#define NETWORK_DEBUG_TIMING		FALSE

#define NETWORK_DEBUG_TIMING_VERBOSE	FALSE

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_types.h"
#include "mx_driver.h"
#include "mx_record.h"
#include "mx_socket.h"
#include "mx_net.h"
#include "mx_net_socket.h"
#include "mx_array.h"
#include "mx_bit.h"

#include "ms_server.h"
#include "ms_mxserver.h"
#include "ms_security.h"

#include "ms_process_record.h"
#include "ms_process_rs232.h"
#include "ms_process_analog_input.h"
#include "ms_process_analog_output.h"
#include "ms_process_digital_input.h"
#include "ms_process_digital_output.h"
#include "ms_process_motor.h"
#include "ms_process_scaler.h"
#include "ms_process_timer.h"
#include "ms_process_amplifier.h"
#include "ms_process_relay.h"
#include "ms_process_mca.h"
#include "ms_process_mce.h"
#include "ms_process_mcs.h"
#include "ms_process_pulser.h"
#include "ms_process_autoscale.h"
#include "ms_process_sca.h"
#include "ms_process_ccd.h"
#include "ms_process_sample_changer.h"
#include "ms_process_variable.h"
#include "ms_process_list_head.h"

#if NETWORK_DEBUG_TIMING
#  include "mx_hrt_debug.h"
#endif

/*
 * If NETWORK_DEBUG_VERBOSE is TRUE, force NETWORK_DEBUG to TRUE.
 */

#if NETWORK_DEBUG_VERBOSE
#  undef NETWORK_DEBUG

#  define NETWORK_DEBUG		FALSE
#endif

/*
 * If NETWORK_DEBUG_TIMING_VERBOSE is TRUE, force NETWORK_DEBUG_TIMING to TRUE.
 */

#if NETWORK_DEBUG_TIMING_VERBOSE
#  undef NETWORK_DEBUG_TIMING

#  define NETWORK_DEBUG_TIMING  FALSE
#endif

MXSRV_MX_SERVER_SOCKET mxsrv_tcp_server_socket_struct;

#if HAVE_UNIX_DOMAIN_SOCKETS
MXSRV_MX_SERVER_SOCKET mxsrv_unix_server_socket_struct;
#endif

typedef struct {
	long mx_superclass;
	long mx_class;
	mx_status_type (*setup_fn) ( MX_RECORD * );
} MXSRV_PROCESS_FUNCTION_SETUP;

static MXSRV_PROCESS_FUNCTION_SETUP mxsrv_process_function_setup_array[] = {

{ MXR_DEVICE, MXC_MOTOR, mxsrv_setup_motor_process_functions },
{ MXR_DEVICE, MXC_SCALER, mxsrv_setup_scaler_process_functions },
{ MXR_DEVICE, MXC_TIMER, mxsrv_setup_timer_process_functions },
{ MXR_DEVICE, MXC_AMPLIFIER, mxsrv_setup_amplifier_process_functions },
{ MXR_DEVICE, MXC_RELAY, mxsrv_setup_relay_process_functions },
{ MXR_DEVICE, MXC_MULTICHANNEL_ANALYZER, mxsrv_setup_mca_process_functions },
{ MXR_DEVICE, MXC_MULTICHANNEL_ENCODER, mxsrv_setup_mce_process_functions },
{ MXR_DEVICE, MXC_MULTICHANNEL_SCALER, mxsrv_setup_mcs_process_functions },
{ MXR_DEVICE, MXC_ANALOG_INPUT, mxsrv_setup_analog_input_process_functions },
{ MXR_DEVICE, MXC_ANALOG_OUTPUT, mxsrv_setup_analog_output_process_functions },
{ MXR_DEVICE, MXC_DIGITAL_INPUT, mxsrv_setup_digital_input_process_functions },
{ MXR_DEVICE, MXC_DIGITAL_OUTPUT, mxsrv_setup_digital_output_process_functions},
{ MXR_DEVICE, MXC_PULSE_GENERATOR, mxsrv_setup_pulser_process_functions},
{ MXR_DEVICE, MXC_AUTOSCALE, mxsrv_setup_autoscale_process_functions },
{ MXR_DEVICE, MXC_SINGLE_CHANNEL_ANALYZER, mxsrv_setup_sca_process_functions },
{ MXR_DEVICE, MXC_CCD, mxsrv_setup_ccd_process_functions },
{ MXR_DEVICE, MXC_SAMPLE_CHANGER, mxsrv_setup_sample_changer_process_functions},
{ MXR_INTERFACE, MXI_RS232, mxsrv_setup_rs232_process_functions },
{ MXR_VARIABLE, MXC_ANY, mxsrv_setup_variable_process_functions },
{ MXR_LIST_HEAD, MXC_ANY, mxsrv_setup_list_head_process_functions },
{ MXR_SCAN, MXC_ANY, NULL },
{ MXR_SERVER, MXC_ANY, NULL },
{ MXR_SPECIAL, MXC_ANY, NULL },
};

static int mxsrv_num_process_function_setup_elements =
			sizeof( mxsrv_process_function_setup_array )
			/ sizeof( mxsrv_process_function_setup_array[0] );

mx_uint32_type
mxsrv_get_returning_message_type( mx_uint32_type message_type )
{
	mx_uint32_type i, returning_message_type;

	static long message_pairs[][2] = {

{MX_NETMSG_GET_ARRAY_BY_NAME, mx_server_response(MX_NETMSG_GET_ARRAY_BY_NAME)},
{MX_NETMSG_PUT_ARRAY_BY_NAME, mx_server_response(MX_NETMSG_PUT_ARRAY_BY_NAME)},
{MX_NETMSG_GET_ARRAY_BY_HANDLE,
			mx_server_response(MX_NETMSG_GET_ARRAY_BY_HANDLE)},
{MX_NETMSG_PUT_ARRAY_BY_HANDLE,
			mx_server_response(MX_NETMSG_PUT_ARRAY_BY_HANDLE)},
{MX_NETMSG_GET_FIELD_TYPE,    mx_server_response(MX_NETMSG_GET_FIELD_TYPE)},
{MX_NETMSG_SET_CLIENT_INFO,   mx_server_response(MX_NETMSG_SET_CLIENT_INFO)},
{MX_NETMSG_GET_OPTION,        mx_server_response(MX_NETMSG_GET_OPTION)},
{MX_NETMSG_SET_OPTION,        mx_server_response(MX_NETMSG_SET_OPTION)},

	};

	static long num_message_pairs = sizeof( message_pairs )
					/ sizeof( message_pairs[0] );

	for ( i = 0; i < num_message_pairs; i++ ) {
		if ( message_type == message_pairs[i][0] ) {
			break;		/* Exit the for() loop. */
		}
	}
	if ( i >= num_message_pairs ) {
		returning_message_type = MX_NETMSG_UNEXPECTED_ERROR;
	} else {
		returning_message_type = message_pairs[i][1];
	}
	return returning_message_type;
}

mx_status_type
mxsrv_initialize_record_processing( MX_RECORD *mx_record_list )
{
	MX_RECORD *current_record, *list_head_record;
	MX_RECORD_FUNCTION_LIST *flist;
	mx_status_type ( *process_function_setup ) ( MX_RECORD * );
	mx_status_type ( *special_processing_setup ) ( MX_RECORD * );
	long processing_flags, this_superclass, this_class;
	int i;
	mx_status_type mx_status;

	list_head_record = mx_record_list->list_head;

	current_record = list_head_record;

	do {

	    process_function_setup = NULL;

	    /* Do we use the default processing function for this record? */

	    processing_flags = current_record->record_processing_flags;

	    if ( processing_flags & MXF_PROC_BYPASS_DEFAULT_PROCESSING ) {

		/* Do nothing here. */

	    } else {

		/* First, setup the record processing that is used for all
		 * MX drivers.
		 */

		mx_status = mxsrv_setup_record_process_functions(
							current_record );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		/* Look for a default processing function for this class
		 * of MX drivers.
		 */

		for (i = 0; i < mxsrv_num_process_function_setup_elements; i++ )
		{
		    this_superclass = 
			    mxsrv_process_function_setup_array[i].mx_superclass;

		    this_class = mxsrv_process_function_setup_array[i].mx_class;

		    if ( ( this_class == current_record->mx_class )
		      && ( this_superclass == current_record->mx_superclass ) )
		    {
			process_function_setup =
				mxsrv_process_function_setup_array[i].setup_fn;
		    } else
		    if ( ( this_class == MXC_ANY )
		      && ( this_superclass == current_record->mx_superclass ) )
		    {
			process_function_setup =
				mxsrv_process_function_setup_array[i].setup_fn;
		    }
		}

		/* If a default processing function is available, invoke it. */

		if ( process_function_setup != NULL ) {
		    mx_status = (*process_function_setup)( current_record );

		    if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
		}
	    }

	    /* If a special processing function is available, invoke it. */

	    flist = (MX_RECORD_FUNCTION_LIST *)
			current_record->record_function_list;

	    special_processing_setup = flist->special_processing_setup;

	    if ( special_processing_setup != NULL ) {
		mx_status = (*special_processing_setup)( current_record );

		if ( mx_status.code != MXE_SUCCESS )
		    return mx_status;
	    }

	    /* Go to the next record. */

	    current_record = current_record->next_record;

	} while ( current_record != list_head_record );

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_free_client_socket_handler( MXSRV_SOCKET_HANDLER *socket_handler,
			MXSRV_SOCKET_HANDLER_LIST *socket_handler_list )
{
	int i;
	mx_status_type mx_status;

	i = socket_handler->handler_array_index;

	if ( socket_handler_list != NULL ) {
		socket_handler_list->array[i] = NULL;
	}

	if ( i >= 0 ) {
		mx_info("Client %d (socket %d) disconnected.",
			i, socket_handler->synchronous_socket->socket_fd);
	}

	/* If this handler had a callback socket, close it. */

	if ( socket_handler->callback_socket != NULL ) {
		mx_status = mx_socket_close( socket_handler->callback_socket );

		if ( socket_handler_list != NULL ) {
			socket_handler_list->num_sockets_in_use--;
		}

		if ( mx_status.code != MXE_SUCCESS ) {
			free( socket_handler );
			return mx_status;
		}
	}

	/* Close our end of the synchronous socket. */

	mx_status = mx_socket_close( socket_handler->synchronous_socket );

	if ( socket_handler_list != NULL ) {
		socket_handler_list->num_sockets_in_use--;
	}

	if ( mx_status.code != MXE_SUCCESS ) {
		free( socket_handler );
		return mx_status;
	}

	/* Free the message buffer. */

	if ( socket_handler->message_buffer != NULL ) {
		free( socket_handler->message_buffer );
	}

	/* Invalidate the contents of the socket handler just in case
	 * someone has a pointer to it.
	 */

	socket_handler->synchronous_socket = NULL;
	socket_handler->callback_socket = NULL;
	socket_handler->handler_array_index = -1;
	socket_handler->event_handler = NULL;
	socket_handler->message_buffer = NULL;

	free( socket_handler );

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_mx_server_socket_init( MX_RECORD *list_head_record,
				MXSRV_SOCKET_HANDLER_LIST *socket_handler_list,
				MXSRV_EVENT_HANDLER *event_handler )
{
	static const char fname[] = "mxsrv_mx_server_socket_init()";

	MXSRV_MX_SERVER_SOCKET *server_socket_struct;
	MXSRV_SOCKET_HANDLER *socket_handler;
	MX_SOCKET *server_socket;
	int i, socket_type, max_sockets, handler_array_size;
	mx_status_type mx_status;

	MX_DEBUG(1,("%s invoked.", fname));

	if ( socket_handler_list == (MXSRV_SOCKET_HANDLER_LIST *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MXSRV_SOCKET_HANDLER_LIST pointer passed is NULL." );
	}
	if ( event_handler == (MXSRV_EVENT_HANDLER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MXSRV_EVENT_HANDLER pointer passed is NULL." );
	}

	max_sockets = socket_handler_list->max_sockets;
	handler_array_size = socket_handler_list->handler_array_size;

	/* Get the port number of the server socket we are supposed to open. */

	server_socket_struct = (MXSRV_MX_SERVER_SOCKET *)
				( event_handler->type_struct );

	if ( server_socket_struct == (MXSRV_MX_SERVER_SOCKET *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MXSRV_MX_SERVER_SOCKET pointer for server event handler '%s' is NULL.",
			event_handler->name );
	}

	/*********** Try to create the server socket. ************/

	socket_type = server_socket_struct->type;

	if ( socket_type == MXF_SRV_TCP_SERVER_TYPE ) {

		int port_number;

		port_number = server_socket_struct->u.tcp.port;

		if ( port_number < 0 ) {
			return mx_error_quiet( MXE_NOT_FOUND, fname,
		"No TCP/IP socket number was specified for the server." );
		}

		if ( port_number > 65535 ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal server port number %d for server event handler '%s'",
				port_number, event_handler->name );
		}

		mx_status = mx_tcp_socket_open_as_server(
						&server_socket, port_number, 0,
						MX_SOCKET_DEFAULT_BUFFER_SIZE );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mx_info( "TCP server socket %d was opened.", port_number );

#if HAVE_UNIX_DOMAIN_SOCKETS
	} else if ( socket_type == MXF_SRV_UNIX_SERVER_TYPE ) {

		char *pathname;

		pathname = server_socket_struct->u.unix_domain.pathname;

		if ( strlen( pathname ) == 0 ) {
			return mx_error_quiet( MXE_NOT_FOUND, fname,
		"No Unix domain socket name was specified for the server." );
		}

		mx_status = mx_unix_socket_open_as_server(
						&server_socket, pathname, 0,
						MX_SOCKET_DEFAULT_BUFFER_SIZE );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mx_info( "Unix domain server socket '%s' was opened.",
				pathname );
#endif
	} else {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal MX server socket type %d requested.",
			socket_type );
	}

#ifndef OS_WIN32
	if ( server_socket->socket_fd < 0 ||
			server_socket->socket_fd >= max_sockets )
	{
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Server socket value %d returned by mx_socket_open_as_server() is outside "
"the allowed range of 0 to %d.", server_socket->socket_fd, max_sockets - 1 );
	}
#endif /* OS_WIN32 */

	/* Add the socket to the list of sockets that we handle. */

	for ( i = 0; i < handler_array_size; i++ ) {

		/* Look for an empty slot in the array. */

		if ( socket_handler_list->array[i] == NULL ) {
			break;                     /* Exit the for() loop. */
		}
	}

	if ( i >= handler_array_size ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"Attempted to use more than the maximum number (%d) "
			"of socket handlers allowed.", handler_array_size );
	}

	socket_handler = (MXSRV_SOCKET_HANDLER *)
				malloc( sizeof(MXSRV_SOCKET_HANDLER) );

	if ( socket_handler == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
"Ran out of memory trying to allocate MXSRV_SOCKET_HANDLER for socket %d.",
			server_socket->socket_fd );
	}

	socket_handler_list->array[i] = socket_handler;

	socket_handler->synchronous_socket = server_socket;
	socket_handler->callback_socket = NULL;
	socket_handler->handler_array_index = i;
	socket_handler->event_handler = event_handler;

	strcpy( socket_handler->client_address_string, "" );
	mx_strncpy( socket_handler->program_name, "mxserver",
					MXU_PROGRAM_NAME_LENGTH );

	mx_username( socket_handler->username, MXU_USERNAME_LENGTH );

	socket_handler->process_id = mx_process_id();

	/* The server socket does not read anything from the socket,
	 * so it does not need a message buffer.
	 */

	socket_handler->message_buffer = NULL;

	socket_handler_list->num_sockets_in_use++;

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_mx_server_socket_process_event( MX_RECORD *record_list,
				MXSRV_SOCKET_HANDLER *socket_handler,
                                MXSRV_SOCKET_HANDLER_LIST *socket_handler_list,
				MXSRV_EVENT_HANDLER *event_handler )
{
	static const char fname[] = "mxsrv_mx_server_socket_process_event()";

	MX_LIST_HEAD *list_head;
	MXSRV_MX_SERVER_SOCKET *server_socket_struct;
	MXSRV_SOCKET_HANDLER *new_socket_handler;
	MX_SOCKET *client_socket;
	int i, socket_type, max_sockets, handler_array_size;
	int saved_errno, connection_allowed;
	char *error_string;
	mx_status_type mx_status;

	void *client_address_ptr;
	char *client_address_string_ptr;

	struct sockaddr_in tcp_client_address;

#if defined(OS_VMS)

	unsigned int tcp_client_address_size = sizeof( struct sockaddr_in );
	unsigned int client_address_size;

#elif defined(__BORLANDC__) || defined(__INTEL_COMPILER)

	socklen_t tcp_client_address_size = sizeof( struct sockaddr_in );
	socklen_t client_address_size;
#else
	int tcp_client_address_size = sizeof( struct sockaddr_in );
	int client_address_size;
#endif

#if HAVE_UNIX_DOMAIN_SOCKETS
	struct sockaddr_un unix_client_address;

	int unix_client_address_size = sizeof( struct sockaddr_un );
#endif

	MX_DEBUG(1,("%s invoked.", fname));

	if ( socket_handler == (MXSRV_SOCKET_HANDLER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MXSRV_SOCKET_HANDLER pointer passed was NULL." );
	}
	if ( socket_handler_list == (MXSRV_SOCKET_HANDLER_LIST *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MXSRV_SOCKET_HANDLER_LIST pointer passed was NULL." );
	}
	if ( event_handler == (MXSRV_EVENT_HANDLER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MXSRV_EVENT_HANDLER pointer passed was NULL." );
	}

	list_head = mx_get_record_list_head_struct( record_list );

	if ( list_head == NULL ) {
		mx_info( "Exiting due to corrupt list head record." );
		exit( MXE_CORRUPT_DATA_STRUCTURE );
	}

	max_sockets = socket_handler_list->max_sockets;
	handler_array_size = socket_handler_list->handler_array_size;

	if ( socket_handler_list->num_sockets_in_use >= (max_sockets - 1) ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"Maximum number (%d) of sockets allowed reached.  "
			"Cannot handle client request for a new connection.",
			max_sockets );
	}

	server_socket_struct = (MXSRV_MX_SERVER_SOCKET *)
					(event_handler->type_struct);

	if ( server_socket_struct == (MXSRV_MX_SERVER_SOCKET *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MXSRV_MX_SERVER_SOCKET pointer for event handler type '%s' is NULL.",
			event_handler->name );
	}

	if ( server_socket_struct->client_event_handler == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"client_event_handler pointer for event handler type '%s' is NULL.",
			event_handler->name );
	}

	socket_type = server_socket_struct->type;

	switch( socket_type ) {
	case MXF_SRV_TCP_SERVER_TYPE:
		client_address_ptr = &tcp_client_address;
		client_address_size = tcp_client_address_size;
		break;

#if HAVE_UNIX_DOMAIN_SOCKETS
	case MXF_SRV_UNIX_SERVER_TYPE:
		client_address_ptr = &unix_client_address;
		client_address_size = unix_client_address_size;
		break;
#endif
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The server_socket_struct passed says that is of illegal type %d.",
			socket_type );
		break;
	}

	/* Allocate memory for the client MX_SOCKET structure. */

	client_socket = (MX_SOCKET *) malloc( sizeof(MX_SOCKET) );

	if ( client_socket == (MX_SOCKET *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a client socket." );
	}

	/* The server socket should be ready for an accept() operation,
	 * so let's try it.
	 */

	client_socket->socket_fd = accept(
				socket_handler->synchronous_socket->socket_fd,
				( struct sockaddr *) client_address_ptr,
				&client_address_size );

        saved_errno = mx_socket_check_error_status( &(client_socket->socket_fd),
					MXF_SOCKCHK_INVALID, &error_string );

        if ( saved_errno != 0 ) {
                return mx_error( MXE_NETWORK_IO_ERROR, fname,
               "Attempt to perform accept() on MX server port failed.  "
                        "Errno = %d.  Error string = '%s'.",
                        saved_errno, error_string );
        }

	/* Allocate a socket handler for the new socket. */

	new_socket_handler = (MXSRV_SOCKET_HANDLER *)
				malloc( sizeof(MXSRV_SOCKET_HANDLER) );

	if ( new_socket_handler == (MXSRV_SOCKET_HANDLER *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
"Attempt to allocate a new socket handler for new client socket %d failed.",
			client_socket->socket_fd );
	}

	new_socket_handler->handler_array_index = -1;

	new_socket_handler->username[0] = '\0';

	new_socket_handler->program_name[0] = '\0';

	new_socket_handler->process_id = -1;

	new_socket_handler->data_format = list_head->default_data_format;

	new_socket_handler->authentication_type = MXF_SRVAUTH_NONE;

	/* Allocate memory for the message buffer. */

	new_socket_handler->message_buffer = (MX_NETWORK_MESSAGE_BUFFER *)
				malloc( sizeof( MX_NETWORK_MESSAGE_BUFFER ) );

	if ( new_socket_handler->message_buffer == NULL ) {

		free( new_socket_handler );

		return mx_error( MXE_OUT_OF_MEMORY, fname,
"Ran out of memory trying to allocate MX_NETWORK_MESSAGE_BUFFER for socket %d.",
			client_socket->socket_fd );
	}

	new_socket_handler->synchronous_socket = client_socket;
	new_socket_handler->callback_socket = NULL;
	new_socket_handler->event_handler
			= server_socket_struct->client_event_handler;

	/* Perform initial checks on the new socket. */

	switch( socket_type ) {
	case MXF_SRV_TCP_SERVER_TYPE:

		/* Convert the host address of the socket to standard
		 * numbers-and-dots notation.
		 */

		client_address_string_ptr =
			inet_ntoa( tcp_client_address.sin_addr );

		mx_strncpy( new_socket_handler->client_address_string,
			client_address_string_ptr,
			MXU_ADDRESS_STRING_LENGTH );

		/* Check the connection access control list to see if the
		 * remote client is permitted to connect to us.
		 */

		if ( list_head->connection_acl == NULL ) {
			connection_allowed = TRUE;
		} else {
			mx_status =
		mxsrv_check_socket_connection_acl_permissions( record_list,
				new_socket_handler->client_address_string,
				&connection_allowed );

			if ( mx_status.code != MXE_SUCCESS ) {
				(void) mxsrv_free_client_socket_handler(
						new_socket_handler, NULL );

				return mx_status;
			}
		}

		if ( connection_allowed == FALSE ) {
			(void) mxsrv_free_client_socket_handler(
						new_socket_handler, NULL );

			return mx_error( MXE_CLIENT_REQUEST_DENIED, fname,
			"Network client '%s' is not allowed to connect to us.",
				client_address_string_ptr );
		}
		break;

#if HAVE_UNIX_DOMAIN_SOCKETS
	case MXF_SRV_UNIX_SERVER_TYPE:
		mx_strncpy( new_socket_handler->client_address_string,
				server_socket_struct->u.unix_domain.pathname,
				MXU_ADDRESS_STRING_LENGTH );

		/* MX Unix domain clients transmit a single null byte when
		 * when they connect.  User credentials are passed along with
		 * this byte for some versions of Unix.
		 */

		mx_status = mxsrv_get_unix_domain_socket_credentials(
				new_socket_handler );

		if ( mx_status.code != MXE_SUCCESS ) {
			(void) mxsrv_free_client_socket_handler(
						new_socket_handler, NULL );

			return mx_status;
		}
		break;
#endif
	}

	/* Now that we seem to have a working client socket, let's put
	 * it into the socket handler list.
	 */

	for ( i = 0; i < handler_array_size; i++ ) {

		/* Look for an empty slot in the array. */

		if ( socket_handler_list->array[i] == NULL ) {
			break;                     /* Exit the for() loop. */
		}
	}

	if ( i >= handler_array_size ) {

		(void) mxsrv_free_client_socket_handler(
				new_socket_handler, NULL );

		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"Attempted to use more than the maximum number (%d) "
			"of socket handlers allowed.", handler_array_size );
	}

	new_socket_handler->handler_array_index = i;

	socket_handler_list->array[i] = new_socket_handler;

	socket_handler_list->num_sockets_in_use++;

	mx_info("Client %d (socket %d) connected from '%s'.",
		i, client_socket->socket_fd,
		new_socket_handler->client_address_string );

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_mx_client_socket_init( MX_RECORD *record_list,
				MXSRV_SOCKET_HANDLER_LIST *socket_handler_list,
				MXSRV_EVENT_HANDLER *event_handler )
{
	/* Nothing to do here.  Client sockets are actually initialized
	 * in mxsrv_mx_server_socket_process_event().
	 */

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_mx_client_socket_process_event( MX_RECORD *record_list,
				MXSRV_SOCKET_HANDLER *socket_handler,
                                MXSRV_SOCKET_HANDLER_LIST *socket_handler_list,
				MXSRV_EVENT_HANDLER *event_handler )
{
	static const char fname[] = "mxsrv_mx_client_socket_process_event()";

	mx_uint32_type *header;
	char *receive_buffer;

	int queue_a_message, record_field_name_at_message_start;
	char *record_name, *field_name;
	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MX_SOCKET *client_socket;
	void *event_data;

	char *ptr, *message;
	int saved_errno;
	long bytes_left, bytes_received, initial_recv_length;
	mx_uint32_type magic_value, header_length, message_length;
	mx_uint32_type message_type, returned_message_type;
	mx_status_type mx_status;

	char separators[] = MX_RECORD_FIELD_SEPARATORS;

#if NETWORK_DEBUG_TIMING
	MX_HRT_TIMING recv_measurement, parse_measurement;
	MX_HRT_TIMING queue_measurement, immediate_measurement;

	MX_HRT_START( recv_measurement );
#endif

        if ( socket_handler == (MXSRV_SOCKET_HANDLER *) NULL ) {
                return mx_error( MXE_NULL_ARGUMENT, fname,
                        "MXSRV_SOCKET_HANDLER pointer passed was NULL." );
        }

	record = NULL;

	client_socket = socket_handler->synchronous_socket;

	MX_DEBUG(1,("%s invoked for socket handler %d.", fname,
				socket_handler->handler_array_index));

#if NETWORK_DEBUG_VERBOSE
	MX_DEBUG(-2,("socket_handler->synchronous_socket = %d",
				socket_handler->synchronous_socket));
	MX_DEBUG(-2,("socket_handler->callback_socket = %d",
				socket_handler->callback_socket));
	MX_DEBUG(-2,("socket_handler->handler_array_index = %d",
				socket_handler->handler_array_index));
	MX_DEBUG(-2,("socket_handler->event_handler = %p",
				socket_handler->event_handler));
#endif

        if ( socket_handler_list == (MXSRV_SOCKET_HANDLER_LIST *) NULL ) {
                return mx_error( MXE_NULL_ARGUMENT, fname,
                "MXSRV_SOCKET_HANDLER_LIST pointer passed was NULL." );
        }
        if ( event_handler == (MXSRV_EVENT_HANDLER *) NULL ) {
                return mx_error( MXE_NULL_ARGUMENT, fname,
                        "MXSRV_EVENT_HANDLER pointer passed was NULL." );
        }

	/* Try to read the beginning of the header of the incoming message. */

	header = socket_handler->message_buffer->uint32_buffer;

	ptr = ( char * ) header;

	initial_recv_length = 4 * sizeof( mx_uint32_type );

	bytes_left = initial_recv_length;

	while( bytes_left > 0 ) {
		bytes_received = recv( client_socket->socket_fd,
						ptr, bytes_left, 0 );

		switch( bytes_received ) {
		case MX_SOCKET_ERROR:
			saved_errno = mx_socket_get_last_error();

			switch( saved_errno ) {
#ifdef OS_WIN32
			case WSAECONNRESET:
#else
			case ECONNRESET:
#endif
				break;
			default:
				return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Error receiving header from remote host.  Errno = %d, error text = '%s'",
			saved_errno, mx_socket_strerror( saved_errno ) );
				break;
			}
			mx_status = mxsrv_free_client_socket_handler(
					socket_handler, socket_handler_list );

			return mx_status;
			break;
		case 0:
			/* The remote socket has closed so remove the
			 * socket handler from the socket handler list.
			 */

			mx_status = mxsrv_free_client_socket_handler(
					socket_handler, socket_handler_list );

			return mx_status;
			break;
		default:
			bytes_left -= bytes_received;
			ptr += bytes_received;
			break;
		}
	}

#if NETWORK_DEBUG_VERBOSE
	{
		long i;
		unsigned char c;

		for ( i = 0; i < initial_recv_length; i++ ) {
			c = (socket_handler->message_buffer->buffer)[i];

			if ( isprint(c) ) {
				MX_DEBUG(-2,("buffer[%ld] = %02x (%c)",
						i, c, c));
			} else {
				MX_DEBUG(-2,("buffer[%ld] = %02x", i, c));
			}
		}
	}
#endif

	magic_value    = ntohl( header[ MX_NETWORK_MAGIC ] );
        header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
        message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
	message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );

#if NETWORK_DEBUG_VERBOSE
        MX_DEBUG(-2,("%s: magic_value    = %lx", fname, magic_value));
        MX_DEBUG(-2,("%s: header_length  = %ld", fname, header_length));
        MX_DEBUG(-2,("%s: message_length = %ld", fname, message_length));
#endif

        MX_DEBUG(1,("%s: message_type   = %ld", fname, message_type));

        if ( magic_value != MX_NETWORK_MAGIC_VALUE ) {
                return mx_error( MXE_NETWORK_IO_ERROR, fname,
                "Wrong magic number %lx in received message.",
		(unsigned long) magic_value );
        }

	/* Allocate a buffer to write the message to.
	 *
	 * The one extra byte is so that a '\0' byte can be placed at
	 * the end of the receive buffer for the sake of network commands
	 * that pass their message bodies as strings.
	 */

	if ( header_length + message_length
				> MX_NETWORK_MAXIMUM_MESSAGE_SIZE )
	{
		(void) mxsrv_free_client_socket_handler(
					socket_handler, socket_handler_list );

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Requested client message length of %lu bytes for socket %d is too long "
"to fit in the server's buffer of %lu bytes.",
			header_length + message_length,
			client_socket->socket_fd,
			MX_NETWORK_MAXIMUM_MESSAGE_SIZE );
	}

	receive_buffer = socket_handler->message_buffer->char_buffer;

        /* Receive the rest of the data. */

	ptr = receive_buffer + initial_recv_length;

        bytes_left = (long)
		(header_length + message_length - initial_recv_length);

        while( bytes_left > 0 ) {
                bytes_received = recv( client_socket->socket_fd,
						ptr, bytes_left, 0 );

                switch( bytes_received ) {
		case MX_SOCKET_ERROR:
			saved_errno = mx_socket_get_last_error();

			(void) mxsrv_free_client_socket_handler(
					socket_handler, socket_handler_list );

                        return mx_error( MXE_NETWORK_IO_ERROR, fname,
                        "Error receiving message body from remote host.  "
                        "Errno = %d, error text = '%s'",
			saved_errno, mx_socket_strerror( saved_errno ) );

                        break;
                case 0:
			(void) mxsrv_free_client_socket_handler(
					socket_handler, socket_handler_list );

                        return mx_error( MXE_NETWORK_IO_ERROR, fname,
				"Network connection closed unexpectedly." );
                        break;
                default:
                        bytes_left -= bytes_received;
                        ptr += bytes_received;
                        break;
                }
        }

	message = receive_buffer + header_length;

	/* Ensure that a '\0' byte is placed after the end of the
	 * message field.
	 */

	message[ message_length ] = '\0';

#if NETWORK_DEBUG_VERBOSE
	MX_DEBUG(-2,("%s: message_length = %ld", fname, message_length));
	{
		long i;

		for ( i = 0; i < message_length; i++ ) {
			MX_DEBUG(-2,("message[%ld] = %x '%c'",
				i, message[i], message[i]));
		}
	}
#endif
#if NETWORK_DEBUG
	{
		char message_type_string[40];
		char text_buffer[200];
		size_t buffer_left;

		switch( message_type ) {
		case MX_NETMSG_GET_ARRAY_BY_NAME:
			strcpy( message_type_string, "GET_ARRAY_BY_NAME" );
			break;
		case MX_NETMSG_PUT_ARRAY_BY_NAME:
			strcpy( message_type_string, "PUT_ARRAY_BY_NAME" );
			break;
		case MX_NETMSG_GET_ARRAY_BY_HANDLE:
			strcpy( message_type_string, "GET_ARRAY_BY_HANDLE" );
			break;
		case MX_NETMSG_PUT_ARRAY_BY_HANDLE:
			strcpy( message_type_string, "PUT_ARRAY_BY_HANDLE" );
			break;
		case MX_NETMSG_GET_FIELD_TYPE:
			strcpy( message_type_string, "GET_FIELD_TYPE" );
			break;
		case MX_NETMSG_SET_CLIENT_INFO:
			strcpy( message_type_string, "SET_CLIENT_INFO" );
			break;
		case MX_NETMSG_GET_OPTION:
			strcpy( message_type_string, "GET_OPTION" );
			break;
		case MX_NETMSG_SET_OPTION:
			strcpy( message_type_string, "SET_OPTION" );
			break;
		default:
			sprintf( message_type_string,
				"Unknown message type %#lx", message_type );
		}

		sprintf( text_buffer, "%s: (%s) message = '",
			fname, message_type_string );

		buffer_left = sizeof(text_buffer) - strlen(text_buffer) - 6;

		strncat( text_buffer, message, buffer_left );

		if ( message_length > buffer_left ) {
			strcat( text_buffer, "***'" );
		} else {
			strcat( text_buffer, "'" );
		}

		MX_DEBUG(-2,("%s", text_buffer));
	}
#endif

#if NETWORK_DEBUG_TIMING
	MX_HRT_END( recv_measurement );

	MX_HRT_START( parse_measurement );
#endif

	/* Set some flags depending on the message type. */

	switch( message_type ) {
	case MX_NETMSG_GET_ARRAY_BY_NAME:
	case MX_NETMSG_PUT_ARRAY_BY_NAME:
	case MX_NETMSG_GET_ARRAY_BY_HANDLE:
	case MX_NETMSG_PUT_ARRAY_BY_HANDLE:
		queue_a_message = TRUE;
		record_field_name_at_message_start = TRUE;
		break;
	case MX_NETMSG_GET_FIELD_TYPE:
		queue_a_message = FALSE;
		record_field_name_at_message_start = TRUE;
		break;
	case MX_NETMSG_SET_CLIENT_INFO:
	case MX_NETMSG_GET_OPTION:
	case MX_NETMSG_SET_OPTION:
		queue_a_message = FALSE;
		record_field_name_at_message_start = FALSE;
		break;
	default:
		mx_status = mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
			"MX message type %#lx is not yet implemented.",
			message_type );

		(void) mx_network_socket_send_error_message( client_socket,
			MX_NETMSG_UNEXPECTED_ERROR, mx_status );

		return mx_status;
		break;
	}

	/* If this message type requires it, we parse the beginning of the
	 * message to look for a record name and a field name.  If found,
	 * we then get the corresponding MX_RECORD and MX_RECORD_FIELD
	 * pointers.
	 */

	mx_status = MX_SUCCESSFUL_RESULT;

	if ( record_field_name_at_message_start == FALSE ) {
		record = NULL;
		record_field = NULL;
	} else {
		/* The do...while(0) loop below is just a trick to make
		 * it easy to jump to the end of this block of code.
		 */

		do {
			/* Which record and field does this correspond to? */

			record_name = message;

			ptr = strchr( message, '.' );

			if ( ptr == NULL ) {
				mx_status = mx_error(
				MXE_ILLEGAL_ARGUMENT, fname,
				"No period found in record field name '%s'.",
					message );

				break;	/* Exit the do...while(0) loop. */
			}

			*ptr = '\0';

			field_name = ++ptr;

		        ptr = field_name + strcspn( field_name, separators );

		        *ptr = '\0';

			record = mx_get_record( record_list, record_name );

			if ( record == (MX_RECORD *) NULL ) {
				mx_status = mx_error( MXE_NOT_FOUND, fname,
				"This server has no record named '%s'.",
					record_name );

				break;	/* Exit the do...while(0) loop. */
			}

			mx_status = mx_find_record_field( record, field_name,
							&record_field );
		} while (0);
	}

	if ( mx_status.code != MXE_SUCCESS ) {
		returned_message_type = mxsrv_get_returning_message_type(
						message_type );

		/* Send back the error message. */

		(void) mx_network_socket_send_error_message( client_socket,
					MX_NETMSG_UNEXPECTED_ERROR, mx_status );

		return mx_status;
	}

#if NETWORK_DEBUG_TIMING
	MX_HRT_END( parse_measurement );
#endif

	/* Do we finish handling this event now or do we queue an event
	 * to be processed later?
	 */

	if ( queue_a_message == TRUE ) {

#if NETWORK_DEBUG_TIMING
		MX_HRT_START( queue_measurement );
#endif

		event_data = (void *) receive_buffer;

		/* Queue the event. */

		mx_status = mxsrv_add_queued_event( socket_handler,
					record, record_field, event_data );

		if ( mx_status.code != MXE_SUCCESS ) {
			returned_message_type
			    = mxsrv_get_returning_message_type( message_type );

			/* Send back the error message. */

			(void) mx_network_socket_send_error_message(
					client_socket,
					MX_NETMSG_UNEXPECTED_ERROR, mx_status );
		}

#if NETWORK_DEBUG_TIMING
		MX_HRT_END( queue_measurement );

		MX_HRT_RESULTS( recv_measurement, fname,
			"receiving for '%s.%s'",
			record->name, record_field->name );

		MX_HRT_RESULTS( parse_measurement, fname,
			"parsing for '%s.%s'",
			record->name, record_field->name );

		MX_HRT_RESULTS( queue_measurement, fname,
			"mxsrv_add_queued_event() for '%s.%s'",
			record->name, record_field->name );
#endif

		return MX_SUCCESSFUL_RESULT;
	}

#if NETWORK_DEBUG_TIMING
	MX_HRT_START( immediate_measurement );
#endif

	/* Here we handle message types that are to be dealt
	 * with immediately.
	 */

	switch( message_type ) {
	case MX_NETMSG_GET_FIELD_TYPE:
		mx_status = mxsrv_handle_get_field_type( client_socket,
					record_field, message, message_length );
		break;
	case MX_NETMSG_SET_CLIENT_INFO:
		mx_status = mxsrv_handle_set_client_info( socket_handler,
						message, message_length );
		break;
	case MX_NETMSG_GET_OPTION:
		mx_status = mxsrv_handle_get_option( socket_handler,
						message, message_length );
		break;
	case MX_NETMSG_SET_OPTION:
		mx_status = mxsrv_handle_set_option( socket_handler,
						message, message_length );
		break;
	default:
		mx_status = mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
			"MX message type %#lx is not yet implemented!",
			message_type );

		(void) mx_network_socket_send_error_message( client_socket,
			MX_NETMSG_UNEXPECTED_ERROR, mx_status );
		break;
	}

#if NETWORK_DEBUG_VERBOSE
	MX_DEBUG(-2,("socket_handler->synchronous_socket = %d",
				socket_handler->synchronous_socket));
	MX_DEBUG(-2,("socket_handler->callback_socket = %d",
				socket_handler->callback_socket));
	MX_DEBUG(-2,("socket_handler->handler_array_index = %d",
				socket_handler->handler_array_index));
	MX_DEBUG(-2,("socket_handler->event_handler = %p",
				socket_handler->event_handler));

	MX_DEBUG(-2,("%s exiting.", fname));
#endif

#if NETWORK_DEBUG_TIMING
	MX_HRT_END( immediate_measurement );

	if ( ( record != NULL ) && ( record_field != NULL ) ) {
		MX_HRT_RESULTS( recv_measurement, fname,
			"receiving for '%s.%s'",
			record->name, record_field->name );

		MX_HRT_RESULTS( parse_measurement, fname,
			"parsing for '%s.%s'",
			record->name, record_field->name );

		MX_HRT_RESULTS( immediate_measurement, fname,
			"immediate processing for '%s.%s'",
			record->name, record_field->name );
	} else {
		MX_HRT_RESULTS( recv_measurement, fname, "receiving" );
		MX_HRT_RESULTS( parse_measurement, fname, "parsing" );
		MX_HRT_RESULTS( immediate_measurement, fname, "immediate" );
	}
#endif
	return mx_status;
}

mx_status_type
mxsrv_mx_client_socket_proc_queued_event( MXSRV_QUEUED_EVENT *queued_event )
{
	static const char fname[] = "mxsrv_mx_client_socket_proc_queued_event()";

	MX_RECORD *record_list;
	MXSRV_SOCKET_HANDLER *socket_handler;
	char *message_buffer;
	mx_uint32_type *header;
	mx_uint32_type message_type;
	mx_status_type mx_status;

	MX_DEBUG(1,("%s invoked.", fname));

	if ( queued_event == (MXSRV_QUEUED_EVENT *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MXSRV_QUEUED_EVENT structure pointer passed is NULL." );
	}

	record_list = queued_event->record->list_head;
	socket_handler = queued_event->socket_handler;

#if NETWORK_DEBUG_VERBOSE
	MX_DEBUG(-2,("socket_handler = %p", socket_handler));
	MX_DEBUG(-2,("socket_handler->synchronous_socket = %d",
				socket_handler->synchronous_socket));
	MX_DEBUG(-2,("socket_handler->callback_socket = %d",
				socket_handler->callback_socket));
	MX_DEBUG(-2,("socket_handler->handler_array_index = %d",
				socket_handler->handler_array_index));
	MX_DEBUG(-2,("socket_handler->event_handler = %p",
				socket_handler->event_handler));
#endif

	message_buffer = (char *) queued_event->event_data;

	header = (mx_uint32_type *) queued_event->event_data;

	message_type = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );

	/* Handle the request. */

	mx_status = MX_SUCCESSFUL_RESULT;

	switch ( message_type ) {
	case MX_NETMSG_GET_ARRAY_BY_NAME:
	case MX_NETMSG_GET_ARRAY_BY_HANDLE:
		mx_status = mxsrv_handle_get_array(
				socket_handler->synchronous_socket,
				queued_event->record,
				queued_event->record_field,
				socket_handler->data_format,
				message_buffer );

		break;
	case MX_NETMSG_PUT_ARRAY_BY_NAME:
	case MX_NETMSG_PUT_ARRAY_BY_HANDLE:
		mx_status = mxsrv_handle_put_array(
				socket_handler->synchronous_socket,
				queued_event->record,
				queued_event->record_field,
				socket_handler->data_format,
				message_buffer );

		break;
	default:
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Illegal client message type %ld", message_type );
		break;
	}

#if NETWORK_DEBUG_VERBOSE
	MX_DEBUG(-2,("socket_handler->synchronous_socket = %d",
				socket_handler->synchronous_socket));
	MX_DEBUG(-2,("socket_handler->callback_socket = %d",
				socket_handler->callback_socket));
	MX_DEBUG(-2,("socket_handler->handler_array_index = %d",
				socket_handler->handler_array_index));
	MX_DEBUG(-2,("socket_handler->event_handler = %p",
				socket_handler->event_handler));

	MX_DEBUG(-2,("%s exiting.", fname));
#endif

	return mx_status;
}

mx_status_type
mxsrv_handle_get_array( MX_SOCKET *mx_socket,
			MX_RECORD *record,
			MX_RECORD_FIELD *record_field,
			unsigned long data_format,
			void *received_buffer )
{
	static const char fname[] = "mxsrv_handle_get_array()";

	char location[ sizeof(fname) + 40 ];
	char *send_buffer, *send_buffer_message;
	mx_uint32_type *send_buffer_header;
	long send_buffer_header_length, send_buffer_message_length;
	long send_buffer_message_actual_length;
	size_t num_bytes_copied;
	mx_status_type (*process_fn) (void *, void *, int);
	void *pointer_to_value;
	int array_is_dynamically_allocated;
	mx_status_type ( *token_constructor )
		(void *, char *, size_t, MX_RECORD *, MX_RECORD_FIELD *);
	mx_status_type mx_status;

#if NETWORK_DEBUG_TIMING
	MX_HRT_TIMING measurement;

	MX_HRT_START( measurement );
#endif

	mx_status = MX_SUCCESSFUL_RESULT;

	MX_DEBUG(1,("***** %s invoked for socket %d *****",
				fname, mx_socket->socket_fd));

	if ( record_field->flags & MXFF_VARARGS ) {
		array_is_dynamically_allocated = TRUE;
	} else {
		array_is_dynamically_allocated = FALSE;
	}

	/* The do...while(0) loop below is just a trick to make it easy
	 * to jump to the end of this block of code.
	 */

	send_buffer = NULL;
	pointer_to_value = NULL;

	do {
		if ( record == (MX_RECORD *) NULL ) {
			mx_status =  mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed was NULL." );

			break;		/* Exit the do...while(0) loop. */
		}

		if ( record_field == (MX_RECORD_FIELD *) NULL ) {
			mx_status = mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD_FIELD pointer passed was NULL." );

			break;		/* Exit the do...while(0) loop. */
		}

#if NETWORK_DEBUG_VERBOSE
	        MX_DEBUG(-2,("%s: record_name = '%s'", fname, record->name));
		MX_DEBUG(-2,("%s: field_name = '%s'", fname,
						record_field->name));
#endif
		/* If there is one, call the process function to read data
		 * from the hardware.
		 */

		process_fn = record_field->process_function;

		if ( process_fn != NULL ) {

			MX_DEBUG(1,("%s: '%s.%s' MX_PROCESS_GET (%ld)",
				fname, record->name, record_field->name,
				record_field->label_value ));

			/* Only invoke the process function for non-negative
			 * values of record_field->label_value.  A field
			 * with a negative label value is just a placeholder
			 * for some value, so it does not need any further
			 * record processing.
			 */

#if NETWORK_DEBUG_TIMING_VERBOSE
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname,
"#1 before record processing for '%s.%s'", record->name, record_field->name );
#endif

			if ( record_field->label_value < 0 ) {
				mx_status = MX_SUCCESSFUL_RESULT;

			} else {
				/* Invoke the record processing function. */

				mx_status = ( *process_fn )
				    ( record, record_field, MX_PROCESS_GET );

				MX_DEBUG(1,(
				"%s: process_fn returned mx_status.code = %ld",
					fname, mx_status.code ));
			}

#if NETWORK_DEBUG_TIMING_VERBOSE
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname,
"#2 after record processing for '%s.%s'", record->name, record_field->name );
#endif

			/* If the process function succeeded, invoke
			 * the value changed callback.
			 */

			if ( mx_status.code == MXE_SUCCESS ) {
				mx_status = mxsrv_invoke_callback_list(
						MXSF_VALUE_CHANGED_CALLBACK,
						record_field );
			}
		}

		/* Construct a pointer to the data to be returned to
		 * the remote client.
		 */

	        if ( record_field->flags & MXFF_VARARGS ) {
			pointer_to_value
				= mx_read_void_pointer_from_memory_location(
					record_field->data_pointer );
	        } else {
	                pointer_to_value = record_field->data_pointer;
		}

#if NETWORK_DEBUG_TIMING_VERBOSE
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname,
"#3 end of do...while loop for '%s.%s'", record->name, record_field->name );
#endif

	} while (0);

	send_buffer_header_length = MX_NETWORK_HEADER_LENGTH_VALUE;

	send_buffer_message_length = MX_NETWORK_MAXIMUM_MESSAGE_SIZE
					- MX_NETWORK_HEADER_LENGTH_VALUE;

	send_buffer_message_actual_length = 0;

	send_buffer = (char *) received_buffer;

	send_buffer_header = (mx_uint32_type *) received_buffer;

	send_buffer_message = send_buffer + send_buffer_header_length;

#if NETWORK_DEBUG_TIMING_VERBOSE
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname,
	"#4 before constructing response for '%s.%s'",
		record->name, record_field->name );
#endif

	if ( mx_status.code == MXE_SUCCESS ) {

		switch( data_format ) {
		case MX_NETWORK_DATAFMT_ASCII:

		        /* Return the data in ASCII MX database format. */

		        mx_status = mx_get_token_constructor(
				record_field->datatype, &token_constructor );

			if ( mx_status.code == MXE_SUCCESS ) {
				strcpy( send_buffer_message, "" );

			        if ( (record_field->num_dimensions == 0)
			          || ((record_field->datatype == MXFT_STRING)
			             && (record_field->num_dimensions == 1)) ) {

			                mx_status = (*token_constructor)(
						pointer_to_value,
						send_buffer_message,
						send_buffer_message_length,
						record, record_field );
				} else {
					mx_status = mx_create_array_description(
						pointer_to_value,
						record_field->num_dimensions-1,
						send_buffer_message,
						send_buffer_message_length,
						record, record_field,
						token_constructor );
				}
		        }

			send_buffer_message_actual_length
				= (long) strlen( send_buffer_message ) + 1;
			break;

		case MX_NETWORK_DATAFMT_RAW:
#if 0
			MX_DEBUG(-2,("%s: pointer_to_value = %p",
					fname, pointer_to_value));
			MX_DEBUG(-2,("%s: record_field->datatype = %ld",
					fname, record_field->datatype));
			MX_DEBUG(-2,("%s: record_field->num_dimensions = %ld",
					fname, record_field->num_dimensions));
			MX_DEBUG(-2,("%s: record_field->dimension = %p",
					fname, record_field->dimension));
			MX_DEBUG(-2,("%s: record_field->data_element_size = %p",
					fname,record_field->data_element_size));
			MX_DEBUG(-2,("%s: send_buffer_message = %p",
					fname, send_buffer_message));
			MX_DEBUG(-2,("%s: send_buffer_message_length = %ld",
				fname, (long) send_buffer_message_length));
			MX_DEBUG(-2,("%s: &num_bytes_copied = %p",
					fname, &num_bytes_copied));
#endif

			mx_status = mx_copy_array_to_buffer(
					pointer_to_value,
					array_is_dynamically_allocated,
					record_field->datatype,
					record_field->num_dimensions,
					record_field->dimension,
					record_field->data_element_size,
					send_buffer_message,
					send_buffer_message_length,
					&num_bytes_copied );

			send_buffer_message_actual_length =
						(long) num_bytes_copied;
			break;

		case MX_NETWORK_DATAFMT_XDR:
#if HAVE_XDR
			mx_status = mx_xdr_data_transfer(
					MX_XDR_ENCODE,
					pointer_to_value,
					array_is_dynamically_allocated,
					record_field->datatype,
					record_field->num_dimensions,
					record_field->dimension,
					record_field->data_element_size,
					send_buffer_message,
					send_buffer_message_length,
					&num_bytes_copied );

			send_buffer_message_actual_length =
					(long) num_bytes_copied;
#else
			mx_status = mx_error( MXE_UNSUPPORTED, fname,
				"XDR network data format is not supported "
				"on this system." );
#endif
			break;

		default:
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		    "Unrecognized network data format type %ld was requested.",
		    		data_format );
			break;
		}
	}

#if NETWORK_DEBUG_TIMING_VERBOSE
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname,
	"#5 after constructing response for '%s.%s'",
		record->name, record_field->name );
#endif

	/* Fill in the header of the message to be sent back to the client. */

	send_buffer_header[ MX_NETWORK_STATUS_CODE ] = htonl(mx_status.code);

        if ( mx_status.code != MXE_SUCCESS ) {
		*send_buffer_message = '\0';

		strncat( send_buffer_message,
			mx_status.message, send_buffer_message_length );

		send_buffer_message_actual_length
			= (long) strlen( send_buffer_message ) + 1;
	}

	send_buffer_header[ MX_NETWORK_MAGIC ]
				= htonl( MX_NETWORK_MAGIC_VALUE );

	send_buffer_header[ MX_NETWORK_MESSAGE_TYPE ]
		= htonl( mx_server_response(MX_NETMSG_GET_ARRAY_BY_NAME) );

	send_buffer_header[ MX_NETWORK_HEADER_LENGTH ]
				= htonl( send_buffer_header_length );

	send_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ]
				= htonl( send_buffer_message_actual_length );

	/* Send the string description back to the client. */

#if NETWORK_DEBUG
	{
		char text_buffer[200];
		size_t buffer_left;
		char *ptr;
		long i, length;
		long max_length = 20;

		switch( data_format ) {
		case MX_NETWORK_DATAFMT_ASCII:
			sprintf( text_buffer,
				"%s: sending response = '", fname );

			buffer_left = sizeof(text_buffer)
					- strlen(text_buffer) - 6;

			strncat( text_buffer,
					send_buffer_message, buffer_left );

			if ( buffer_left < send_buffer_message_length ) {
				strcat( text_buffer, "***'" );
			} else {
				strcat( text_buffer, "'" );
			}
			break;

		case MX_NETWORK_DATAFMT_RAW:
			sprintf( text_buffer,
				"%s: sending response = ", fname );

			ptr = text_buffer + strlen( text_buffer );

			length = send_buffer_message_length;

			if ( length > max_length )
				length = max_length;

			for ( i = 0; i < length; i++ ) {
				sprintf(ptr, "%#02x ", send_buffer_message[i]);

				ptr += 5;
			}
			break;

		default:
			sprintf( text_buffer,
				"%s: sending response in data format %d.",
				fname, data_format );
			break;
		}

		MX_DEBUG(-2,("%s", text_buffer));
	}
#endif

#if NETWORK_DEBUG_TIMING_VERBOSE
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname,
"#6 before sending reply for '%s.%s'", record->name, record_field->name );
#endif

	mx_status = mx_network_socket_send_message( mx_socket, send_buffer );

	if ( mx_status.code != MXE_SUCCESS ) {
		sprintf( location, "%s to client socket %d",
				fname, mx_socket->socket_fd );

		return mx_error( mx_status.code, location, mx_status.message );
	}

	MX_DEBUG(1,("***** %s successful *****", fname));

#if NETWORK_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname,
		"final for '%s.%s'", record->name, record_field->name );
#endif

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_handle_put_array( MX_SOCKET *mx_socket,
			MX_RECORD *record,
			MX_RECORD_FIELD *record_field,
			unsigned long data_format,
			void *receive_buffer )
{
	static const char fname[] = "mxsrv_handle_put_array()";

	char location[ sizeof(fname) + 40 ];
	char *receive_buffer_message;
	mx_uint32_type *receive_buffer_header;
	mx_uint32_type receive_buffer_header_length;
	mx_uint32_type receive_buffer_message_length;

	char *send_buffer, *send_buffer_message;
	mx_uint32_type *send_buffer_header;
	mx_uint32_type send_buffer_header_length, send_buffer_message_length;

	MX_RECORD_FIELD_PARSE_STATUS parse_status;
	mx_status_type (*process_fn) (void *, void *, int);
	char token_buffer[500];
	char *value_buffer_pointer;
	void *pointer_to_value;
	size_t buffer_left;
	unsigned long i, ptr_address, remainder, gap_size;
	int array_is_dynamically_allocated;
	mx_status_type ( *token_parser )
		( void *, char *, MX_RECORD *, MX_RECORD_FIELD *,
		MX_RECORD_FIELD_PARSE_STATUS * );
	mx_status_type mx_status;

	char separators[] = MX_RECORD_FIELD_SEPARATORS;

#if NETWORK_DEBUG_TIMING
	MX_HRT_TIMING measurement;

	MX_HRT_START( measurement );
#endif

	mx_status = MX_SUCCESSFUL_RESULT;

	MX_DEBUG(1,("***** %s invoked for socket %d *****",
		fname, mx_socket->socket_fd));

	if ( record_field->flags & MXFF_VARARGS ) {
		array_is_dynamically_allocated = TRUE;
	} else {
		array_is_dynamically_allocated = FALSE;
	}

	/* The do...while(0) loop below is just a trick to make it easy
	 * to jump to the end of this block of code.
	 */

	do {
		if ( record == (MX_RECORD *) NULL ) {
			mx_status =  mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed was NULL." );

			break;		/* Exit the do...while(0) loop. */
		}

		if ( record_field == (MX_RECORD_FIELD *) NULL ) {
			mx_status =  mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD_FIELD pointer passed was NULL." );

			break;		/* Exit the do...while(0) loop. */
		}

#if NETWORK_DEBUG_VERBOSE
	        MX_DEBUG(-2,("%s: record_name = '%s'", fname, record->name));
		MX_DEBUG(-2,("%s: field_name = '%s'", fname,
						record_field->name));
#endif

	        if ( record_field->flags & MXFF_VARARGS ) {
			pointer_to_value
				= mx_read_void_pointer_from_memory_location(
					record_field->data_pointer );
	        } else {
	                pointer_to_value = record_field->data_pointer;
		}

	        mx_status = mx_get_token_parser(
                        record_field->datatype, &token_parser );

		if ( mx_status.code != MXE_SUCCESS )
			break;		/* Exit the do...while(0) loop. */

		/* Get a pointer to the start of the value string. */

		receive_buffer_header = (mx_uint32_type *) receive_buffer;

		receive_buffer_header_length
		  = ntohl( receive_buffer_header[ MX_NETWORK_HEADER_LENGTH ] );

		receive_buffer_message = ((char *)receive_buffer)
						+ receive_buffer_header_length;

		receive_buffer_message_length
		  = ntohl( receive_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ] );

		value_buffer_pointer = receive_buffer_message
					+ MXU_RECORD_FIELD_NAME_LENGTH;

		buffer_left = receive_buffer_message_length
					- MXU_RECORD_FIELD_NAME_LENGTH;

		switch( data_format ) {
		case MX_NETWORK_DATAFMT_ASCII:

			/* The data were sent in ASCII MX database format. */

#if NETWORK_DEBUG_VERBOSE
			MX_DEBUG(-2,("%s: value_string = '%s'",
						fname, value_string));
#endif

			mx_initialize_parse_status( &parse_status,
					value_buffer_pointer, separators );

			/* If this is a string field, get the maximum length of
			 * a string token for use by the parser.
			 */

			if ( record_field->datatype == MXFT_STRING ) {
				parse_status.max_string_token_length =
				 mx_get_max_string_token_length( record_field );
			} else {
				parse_status.max_string_token_length = 0L;
			}

			/* Parse the tokens. */

			if ( (record_field->num_dimensions == 0)
	                  || ((record_field->datatype == MXFT_STRING)
			   && (record_field->num_dimensions == 1)) ) {

	                        mx_status = mx_get_next_record_token(
					&parse_status,
                                        token_buffer, sizeof(token_buffer) );

	                        if ( mx_status.code == MXE_SUCCESS ) {
#if NETWORK_DEBUG_VERBOSE
					MX_DEBUG(-2,
				    ("%s: calling *token_parser() for '%s.%s'",
				    fname, record->name, record_field->name));
#endif

		                        mx_status = ( *token_parser ) (
						pointer_to_value,
						token_buffer,
						record, record_field,
						&parse_status );
				}
	                } else {

#if NETWORK_DEBUG_VERBOSE
				MX_DEBUG(-2,
			("%s: calling mx_parse_array_description for '%s.%s'",
				fname, record->name, record_field->name));
#endif
				mx_status = mx_parse_array_description(
                                        pointer_to_value,
                                        (record_field->num_dimensions - 1),
                                        record, record_field,
                                        &parse_status, token_parser );
                	}
			break;

		case MX_NETWORK_DATAFMT_RAW:
			mx_status = mx_copy_buffer_to_array(
					value_buffer_pointer,
					buffer_left,
					pointer_to_value,
					array_is_dynamically_allocated,
					record_field->datatype,
					record_field->num_dimensions,
					record_field->dimension,
					record_field->data_element_size,
					NULL );
			break;

		case MX_NETWORK_DATAFMT_XDR:
#if HAVE_XDR
			/* The XDR data pointer 'ptr' must be aligned on a
			 * 4 byte address boundary for XDR data conversion
			 * to work correctly on all architectures.  If the
			 * pointer does not point to an address that is a
			 * multiple of 4 bytes, we move it to the next address
			 * that _is_ and fill the bytes inbetween with zeros.
		 	 */

			ptr_address = (unsigned long) value_buffer_pointer;

			remainder = ptr_address % 4;

			if ( remainder != 0 ) {
				gap_size = 4 - remainder;

				for ( i = 0; i < gap_size; i++ ) {
					value_buffer_pointer[i] = '\0';
				}

				value_buffer_pointer += gap_size;

				buffer_left -= gap_size;
			}

			MX_DEBUG( 2,
			("%s: ptr_address = %#lx, value_buffer_pointer = %p",
				fname, ptr_address, value_buffer_pointer));

			/* Now we are ready to do the XDR data conversion. */

			mx_status = mx_xdr_data_transfer(
					MX_XDR_DECODE,
					pointer_to_value,
					array_is_dynamically_allocated,
					record_field->datatype,
					record_field->num_dimensions,
					record_field->dimension,
					record_field->data_element_size,
					value_buffer_pointer,
					buffer_left,
					NULL );
#else
			mx_status = mx_error( MXE_UNSUPPORTED, fname,
				"XDR network data format is not supported "
				"on this system." );
#endif
			break;

		default:
			mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		    "Unrecognized network data format type %ld was requested.",
		    		data_format );
			break;
		}

		if ( mx_status.code != MXE_SUCCESS )
			break;		/* Exit the do...while(0) loop. */

		/* If there is one, call the process function to write data
		 * to the hardware.
		 */

		process_fn = record_field->process_function;

		if ( process_fn != NULL ) {

			MX_DEBUG(1,("%s: '%s.%s' MX_PROCESS_PUT (%ld)",
				fname, record->name, record_field->name,
				record_field->label_value ));

			/* Only invoke the process function for non-negative
			 * values of record_field->label_value.  A field
			 * with a negative label value is just a placeholder
			 * for some value, so it does not need any further
			 * record processing.
			 */

			if ( record_field->label_value < 0 ) {
				mx_status = MX_SUCCESSFUL_RESULT;

			} else {
				/* Invoke the record processing function. */

				mx_status = ( *process_fn )
				    ( record, record_field, MX_PROCESS_PUT );

				MX_DEBUG(1,(
				"%s: process_fn returned mx_status.code = %ld",
					fname, mx_status.code ));
			}

			/* If the process function succeeded, invoke
			 * the value changed callback.
			 */

			if ( mx_status.code == MXE_SUCCESS ) {
				mx_status = mxsrv_invoke_callback_list(
						MXSF_VALUE_CHANGED_CALLBACK,
						record_field );
			}
		}
	} while (0);

	/* Allocate a buffer to construct the message to be sent
	 * back to the client.
	 */

	send_buffer_header_length = MX_NETWORK_HEADER_LENGTH_VALUE;

        if ( mx_status.code == MXE_SUCCESS ) {
		send_buffer_message_length = 1;
	} else {
		send_buffer_message_length = 1 + strlen(mx_status.message);
	}

	send_buffer = (char *) receive_buffer;

	send_buffer_header = (mx_uint32_type *) receive_buffer;

	send_buffer_message = send_buffer + send_buffer_header_length;

	send_buffer_header[ MX_NETWORK_STATUS_CODE ] = htonl(mx_status.code);

        if ( mx_status.code == MXE_SUCCESS ) {
		strcpy( send_buffer_message, "" );
	} else {
		strcpy( send_buffer_message, mx_status.message );
	}

	send_buffer_header[ MX_NETWORK_MAGIC ]
				= htonl( MX_NETWORK_MAGIC_VALUE );

	send_buffer_header[ MX_NETWORK_MESSAGE_TYPE ]
		= htonl( mx_server_response(MX_NETMSG_PUT_ARRAY_BY_NAME) );

	send_buffer_header[ MX_NETWORK_HEADER_LENGTH ]
				= htonl( send_buffer_header_length );

	send_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ]
				= htonl( send_buffer_message_length );

	/* Send the string description back to the client. */

#if NETWORK_DEBUG
	{
		char text_buffer[200];

		sprintf( text_buffer, "%s: sending response = '", fname );

		buffer_left = sizeof(text_buffer) - strlen(text_buffer) - 6;

		strncat( text_buffer, send_buffer_message, buffer_left );

		if ( buffer_left < send_buffer_message_length ) {
			strcat( text_buffer, "***'" );
		} else {
			strcat( text_buffer, "'" );
		}

		MX_DEBUG(-2,("%s", text_buffer));
	}
#endif

	mx_status = mx_network_socket_send_message( mx_socket, send_buffer );

	if ( mx_status.code != MXE_SUCCESS ) {
		sprintf( location, "%s to client socket %d",
				fname, mx_socket->socket_fd );

		return mx_error( mx_status.code, location, mx_status.message );
	}

	MX_DEBUG(1,("***** %s successful *****", fname));

#if NETWORK_DEBUG_TIMING
	MX_HRT_END( measurement );

	MX_HRT_RESULTS( measurement, fname,
		"for '%s.%s'", record->name, record_field->name );
#endif

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_handle_get_field_type( MX_SOCKET *mx_socket, MX_RECORD_FIELD *field,
				void *message, mx_uint32_type message_length )
{
	static const char fname[] = "mxsrv_handle_get_field_type()";

	char location[ sizeof(fname) + 40 ];
	char *send_buffer, *message_ptr;
	mx_uint32_type *send_buffer_header, *send_buffer_message;
	mx_uint32_type i, send_buffer_header_length, send_buffer_message_length;
	mx_status_type mx_status;

	message_ptr = (char *) message;

	MX_DEBUG(1,("%s for '%s', message_length = %ld",
			fname, message_ptr, message_length));
	MX_DEBUG(1,("%s: field->datatype = %ld, field->num_dimensions = %ld",
			fname, field->datatype, field->num_dimensions));
	MX_DEBUG(1,("%s: field->dimension[0] = %ld",
			fname, field->dimension[0]));

	/* Allocate a buffer to construct the message to be sent
	 * back to the client.
	 */

	send_buffer_header_length = MX_NETWORK_HEADER_LENGTH_VALUE;

	send_buffer_message_length = sizeof(mx_uint32_type)
		* ( 2 + field->num_dimensions );

	/* Ensure that we can always assign a value to the first element
	 * of the dimension array.
	 */

	if ( send_buffer_message_length <= (2 * sizeof(mx_uint32_type)) ) {
		send_buffer_message_length = 3 * sizeof(mx_uint32_type);
	}

	send_buffer = (char *) message;

	send_buffer_header = (mx_uint32_type *) message;

	send_buffer_message = send_buffer_header
		+ ( send_buffer_header_length / sizeof(mx_uint32_type) );

	/* Construct the header of the message. */

	send_buffer_header[ MX_NETWORK_MAGIC ]
				= htonl( MX_NETWORK_MAGIC_VALUE );

	send_buffer_header[ MX_NETWORK_MESSAGE_TYPE ]
		= htonl( mx_server_response(MX_NETMSG_GET_FIELD_TYPE) );

	send_buffer_header[ MX_NETWORK_HEADER_LENGTH ]
				= htonl( send_buffer_header_length );

	send_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ]
				= htonl( send_buffer_message_length );

	send_buffer_header[ MX_NETWORK_STATUS_CODE ] = htonl( MXE_SUCCESS );

	/* Construct the body of the message. */

	send_buffer_message[0] = htonl( field->datatype );
	send_buffer_message[1] = htonl( field->num_dimensions );

	if ( field->num_dimensions <= 0 ) {
		send_buffer_message[2] = htonl( 0L );
	} else {
		for ( i = 0; i < field->num_dimensions; i++ ) {
			send_buffer_message[i+2] = htonl(field->dimension[i]);
		}
	}

#if NETWORK_DEBUG_VERBOSE
	mx_network_display_message_buffer( send_buffer );
#endif

	/* Send the field type information back to the client. */

	mx_status = mx_network_socket_send_message( mx_socket, send_buffer );

	if ( mx_status.code != MXE_SUCCESS ) {
		sprintf( location, "%s to client socket %d",
				fname, mx_socket->socket_fd );

		return mx_error( mx_status.code, location, mx_status.message );
	}

	MX_DEBUG(1,("***** %s successful *****", fname));

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_handle_set_client_info( MXSRV_SOCKET_HANDLER *socket_handler, 
				void *message, mx_uint32_type message_length )
{
	static const char fname[] = "mxsrv_handle_set_client_info()";

	char location[ sizeof(fname) + 40 ];

	mx_uint32_type *send_buffer_header;
	mx_uint32_type send_buffer_header_length, send_buffer_message_length;
	char *send_buffer_message;
	char *send_buffer, *message_string;
	char *ptr, *ptr2, *endptr;
	size_t length;
	mx_status_type mx_status;

	message_string = (char *) message;

	MX_DEBUG( 1,("%s for '%s', message_length = %ld",
			fname, message_string, message_length));

	/* The do...while(0) loop below is just a trick to make it easy
	 * to jump to the end of this block of code.
	 */

	do {
		/* How many characters at the start of the string
		 * are whitespace?
		 */

		length = strspn( message_string, MX_WHITESPACE );

		ptr = message_string + length;

		/* The username should be the first string in the message
		 * buffer.  Find out how long it is.
		 */

		length = strcspn( ptr, MX_WHITESPACE );

		if ( ((long) length) <= 0 )
			break;	/* Exit the do...while(0) loop. */

		ptr2 = ptr + length;

		*ptr2 = '\0';

		if ( length > MXU_USERNAME_LENGTH )
			length = MXU_USERNAME_LENGTH;

		/* The username is only stored if it has not yet been set. */

		if ( strlen( socket_handler->username ) == 0 ) {
			mx_strncpy( socket_handler->username, ptr, length+1 );
		}

		ptr = ptr2 + 1;

		/* The next string should be the program name.  How many
		 * whitespace characters separate us from it?
		 */

		length = strspn( ptr, MX_WHITESPACE );

		ptr += length;

		/* How long is the program name? */

		length = strcspn( ptr, MX_WHITESPACE );

		if ( ((long) length) <= 0 )
			break;	/* Exit the do...while(0) loop. */

		ptr2 = ptr + length;

		*ptr2 = '\0';

		if ( length > MXU_PROGRAM_NAME_LENGTH )
			length = MXU_PROGRAM_NAME_LENGTH;

		mx_strncpy( socket_handler->program_name, ptr, length+1 );

		ptr = ptr2 + 1;

		/* The next string should be the process id.  How many
		 * whitespace characters separate us from it?
		 */

		length = strspn( ptr, MX_WHITESPACE );

		ptr += length;

		/* How long is the process id? */

		length = strcspn( ptr, MX_WHITESPACE );

		if ( ((long) length) <= 0 )
			break;	/* Exit the do...while(0) loop. */

		ptr2 = ptr + length;

		*ptr2 = '\0';

		socket_handler->process_id = strtoul( ptr, &endptr, 0 );

		if ( *endptr != '\0' ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The string '%s' sent for the process id is not a legal number.", ptr );
		}

		/* Ignore the rest of the message buffer. */

	} while (0);

	MX_DEBUG( 2,("%s: username = '%s'", fname, socket_handler->username));
	MX_DEBUG( 2,("%s: program_name = '%s'",
				fname, socket_handler->program_name));

	/* Allocate a buffer to construct the success message to be sent
	 * back to the client.
	 */

	send_buffer_header_length = MX_NETWORK_HEADER_LENGTH_VALUE;

	send_buffer_message_length = 1;

	send_buffer = (char *) message;

	send_buffer_header = (mx_uint32_type *) message;

	send_buffer_message = (char *) ( send_buffer_header
		+ ( send_buffer_header_length / sizeof(mx_uint32_type) ) );

	/* Construct the header of the message. */

	send_buffer_header[ MX_NETWORK_MAGIC ]
				= htonl( MX_NETWORK_MAGIC_VALUE );

	send_buffer_header[ MX_NETWORK_MESSAGE_TYPE ]
		= htonl( mx_server_response(MX_NETMSG_SET_CLIENT_INFO) );

	send_buffer_header[ MX_NETWORK_HEADER_LENGTH ]
				= htonl( send_buffer_header_length );

	send_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ]
				= htonl( send_buffer_message_length );

	send_buffer_header[ MX_NETWORK_STATUS_CODE ] = htonl( MXE_SUCCESS );

	/* Construct an empty body for the message. */

	strcpy( send_buffer_message, "" );

#if NETWORK_DEBUG_VERBOSE
	mx_network_display_message_buffer( send_buffer );
#endif

	/* Send the success message back to the client. */

	mx_status = mx_network_socket_send_message(
			socket_handler->synchronous_socket, send_buffer );

	if ( mx_status.code != MXE_SUCCESS ) {
		sprintf( location, "%s to client socket %d",
			fname, socket_handler->synchronous_socket->socket_fd );

		return mx_error( mx_status.code, location, mx_status.message );
	}

	MX_DEBUG(1,("***** %s successful *****", fname));

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_handle_get_option( MXSRV_SOCKET_HANDLER *socket_handler,
			void *message,
			mx_uint32_type message_length )
{
	static const char fname[] = "mxsrv_handle_get_option()";

	char location[ sizeof(fname) + 40 ];
	char *send_buffer;
	mx_uint32_type *send_buffer_header, *send_buffer_message;
	mx_uint32_type send_buffer_header_length;
	mx_uint32_type *option_array;
	mx_uint32_type option_number, option_value;
	int illegal_option_number;
	mx_status_type mx_status;

	option_array = (mx_uint32_type *) message;

	option_number = ntohl( option_array[0] );

	MX_DEBUG( 2,("%s: option_number = %#lx", fname, option_number));

	/* Get the requested option value. */

	illegal_option_number = FALSE;

	switch( option_number ) {
	case MX_NETWORK_OPTION_DATAFMT:
		option_value = socket_handler->data_format;
		break;
	case MX_NETWORK_OPTION_NATIVE_DATAFMT:
		option_value = mx_native_data_format();
		break;
	default:
		option_value = 0;
		illegal_option_number = TRUE;
		break;
	}

	/* Allocate a buffer to construct the message to be sent
	 * back to the client.
	 */

	send_buffer_header_length = MX_NETWORK_HEADER_LENGTH_VALUE;

	send_buffer = (char *) message;

	send_buffer_header = (mx_uint32_type *) message;

	send_buffer_message = send_buffer_header
		+ ( send_buffer_header_length / sizeof(mx_uint32_type) );

	/* Construct the header of the message. */

	send_buffer_header[ MX_NETWORK_MAGIC ]
				= htonl( MX_NETWORK_MAGIC_VALUE );

	send_buffer_header[ MX_NETWORK_MESSAGE_TYPE ]
		= htonl( mx_server_response(MX_NETMSG_GET_OPTION) );

	send_buffer_header[ MX_NETWORK_HEADER_LENGTH ]
				= htonl( send_buffer_header_length );

	/* Construct the body of the message. */

	if ( illegal_option_number == FALSE ) {
		send_buffer_message[0] = htonl( option_value );

		send_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ]
				= htonl( sizeof(mx_uint32_type) );

		send_buffer_header[ MX_NETWORK_STATUS_CODE ]
				= htonl( MXE_SUCCESS );
	} else {
		char *error_message;

		error_message = (char *) send_buffer_message;

		sprintf( error_message, "Illegal option number %#lx",
						option_number );

		send_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ]
				= htonl( strlen(error_message) + 1 );

		send_buffer_header[ MX_NETWORK_STATUS_CODE ]
				= htonl( MXE_ILLEGAL_ARGUMENT );
	}

#if NETWORK_DEBUG_VERBOSE
	mx_network_display_message_buffer( send_buffer );
#endif

	/* Send the option information back to the client. */

	mx_status = mx_network_socket_send_message(
			socket_handler->synchronous_socket, send_buffer );

	if ( mx_status.code != MXE_SUCCESS ) {
		sprintf( location, "%s to client socket %d",
			fname, socket_handler->synchronous_socket->socket_fd );

		return mx_error( mx_status.code, location, mx_status.message );
	}

	MX_DEBUG( 1,("***** %s successful *****", fname));

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_handle_set_option( MXSRV_SOCKET_HANDLER *socket_handler,
			void *message,
			mx_uint32_type message_length )
{
	static const char fname[] = "mxsrv_handle_set_option()";

	char location[ sizeof(fname) + 40 ];
	char *send_buffer, *send_buffer_char_message;
	mx_uint32_type *send_buffer_header;
	mx_uint32_type send_buffer_header_length;
	mx_uint32_type *option_array;
	mx_uint32_type option_number, option_value;
	int illegal_option_number, illegal_option_value;
	mx_status_type mx_status;

	option_array = (mx_uint32_type *) message;

	option_number = ntohl( option_array[0] );
	option_value = ntohl( option_array[1] );

	MX_DEBUG( 2,("%s: option_number = %#lx, option_value = %#lx",
		fname, option_number, option_value));

	/* Set the requested option value. */

	illegal_option_number = FALSE;
	illegal_option_value = FALSE;

	switch( option_number ) {
	case MX_NETWORK_OPTION_DATAFMT:
		socket_handler->data_format = option_value;
		break;
	default:
		illegal_option_number = TRUE;
		break;
	}

	/* Allocate a buffer to construct the message to be sent
	 * back to the client.
	 */

	send_buffer_header_length = MX_NETWORK_HEADER_LENGTH_VALUE;

	send_buffer = (char *) message;

	send_buffer_header = (mx_uint32_type *) message;

	send_buffer_char_message = (char *) ( send_buffer_header
		+ ( send_buffer_header_length / sizeof(mx_uint32_type) ) );

	/* Construct the header of the message. */

	send_buffer_header[ MX_NETWORK_MAGIC ]
				= htonl( MX_NETWORK_MAGIC_VALUE );

	send_buffer_header[ MX_NETWORK_MESSAGE_TYPE ]
		= htonl( mx_server_response(MX_NETMSG_SET_OPTION) );

	send_buffer_header[ MX_NETWORK_HEADER_LENGTH ]
				= htonl( send_buffer_header_length );

	/* Construct the body of the message. */

	if ( illegal_option_number ) {
		sprintf( send_buffer_char_message,
				"Illegal option number %#lx", option_number );

		send_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ]
				= htonl( strlen(send_buffer_char_message) + 1 );

		send_buffer_header[ MX_NETWORK_STATUS_CODE ]
				= htonl( MXE_ILLEGAL_ARGUMENT );
	} else
	if ( illegal_option_value ) {
		sprintf( send_buffer_char_message,
				"Illegal option value %#lx for option %#lx",
					option_value, option_number );

		send_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ]
				= htonl( strlen(send_buffer_char_message) + 1 );

		send_buffer_header[ MX_NETWORK_STATUS_CODE ]
				= htonl( MXE_ILLEGAL_ARGUMENT );
	} else {
		send_buffer_char_message[0] = '\0';

		send_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ] = htonl( 1 );

		send_buffer_header[ MX_NETWORK_STATUS_CODE ]
				= htonl( MXE_SUCCESS );
	}

#if NETWORK_DEBUG_VERBOSE
	mx_network_display_message_buffer( send_buffer );
#endif

	/* Send the option information back to the client. */

	mx_status = mx_network_socket_send_message(
			socket_handler->synchronous_socket, send_buffer );

	if ( mx_status.code != MXE_SUCCESS ) {
		sprintf( location, "%s to client socket %d",
			fname, socket_handler->synchronous_socket->socket_fd );

		return mx_error( mx_status.code, location, mx_status.message );
	}

	MX_DEBUG( 1,("***** %s successful *****", fname));

	return MX_SUCCESSFUL_RESULT;
}

#if HAVE_UNIX_DOMAIN_SOCKETS

mx_status_type
mxsrv_get_unix_domain_socket_credentials( MXSRV_SOCKET_HANDLER *socket_handler )
{
	static const char fname[] =
		"mxsrv_get_unix_domain_socket_credentials()";

	char null_byte;
	mx_status_type mx_status;

	null_byte = 1;

	if ( socket_handler == (MXSRV_SOCKET_HANDLER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MXSRV_SOCKET_HANDLER pointer passed was NULL." );
	}

#if 0
#else
	/* If we cannot get user credentials on this platform, then just
	 * read and discard the null byte.
	 */

	socket_handler->username[0] = '\0';

	mx_status = mx_socket_receive( socket_handler->synchronous_socket,
					&null_byte, 1,
					NULL, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;
#endif

	return MX_SUCCESSFUL_RESULT;
}

#endif
